<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    actionlog
 */

/**
 * Module page class.
 */
class Module_admin_actionlog
{
    /**
     * Find details of the module.
     *
     * @return ?array Map of module info (null: module is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 2;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'actionlog';
        return $info;
    }

    /**
     * Find entry-points available within this module.
     *
     * @param  boolean $check_perms Whether to check permissions
     * @param  ?MEMBER $member_id The member to check permissions as (null: current user)
     * @param  boolean $support_crosslinks Whether to allow cross links to other modules (identifiable via a full-page-link rather than a screen-name)
     * @param  boolean $be_deferential Whether to avoid any entry-point (or even return null to disable the page in the Sitemap) if we know another module, or page_group, is going to link to that entry-point. Note that "!" and "browse" entry points are automatically merged with container page nodes (likely called by page-groupings) as appropriate.
     * @return ?array A map of entry points (screen-name=>language-code/string or screen-name=>[language-code/string, icon-theme-image]) (null: disabled)
     */
    public function get_entry_points(bool $check_perms = true, ?int $member_id = null, bool $support_crosslinks = true, bool $be_deferential = false) : ?array
    {
        if (!addon_installed('actionlog')) {
            return null;
        }

        return [
            'browse' => ['VIEW_ACTIONLOGS', 'menu/adminzone/audit/actionlog'],
        ];
    }

    public $title;

    /**
     * Module pre-run function. Allows us to know metadata for <head> before we start streaming output.
     *
     * @return ?Tempcode Tempcode indicating some kind of exceptional output (null: none)
     */
    public function pre_run() : ?object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('actionlog', $error_msg)) {
            return $error_msg;
        }

        $type = get_param_string('type', 'browse');

        require_code('actionlog');

        if ($type == 'browse') {
            set_helper_panel_tutorial('tut_trace');

            breadcrumb_set_self(do_lang_tempcode('VIEW_ACTIONLOGS'));

            $this->title = get_screen_title('VIEW_ACTIONLOGS');
        }

        if ($type == 'list') {
            breadcrumb_set_parents([['_SELF:_SELF:browse', do_lang_tempcode('VIEW_ACTIONLOGS')]]);
            breadcrumb_set_self(do_lang_tempcode('RESULTS'));

            $this->title = get_screen_title('VIEW_ACTIONLOGS');
        }

        if ($type == 'view') {
            breadcrumb_set_self(do_lang_tempcode('ENTRY'));
            breadcrumb_set_parents([['_SELF:_SELF:browse', do_lang_tempcode('VIEW_ACTIONLOGS')], ['_SELF:_SELF:list', do_lang_tempcode('RESULTS')]]);

            $this->title = get_screen_title('VIEW_ACTIONLOGS');
        }

        return null;
    }

    /**
     * Execute the module.
     *
     * @return Tempcode The result of execution
     */
    public function run() : object
    {
        require_all_lang();

        require_code('global4');

        $type = get_param_string('type', 'browse');

        if ($type == 'browse') {
            return $this->search();
        }
        if ($type == 'list') {
            return $this->choose_action();
        }
        if ($type == 'view') {
            return $this->view_action();
        }

        return new Tempcode();
    }

    /**
     * The UI to choose filter parameters.
     *
     * @return Tempcode The UI
     */
    public function search() : object
    {
        require_code('form_templates');

        $fields = new Tempcode();

        // Possible selections for member filter
        $_member_choice_list = [];
        if (get_forum_type() == 'cns') {
            if ($GLOBALS['FORUM_DB']->query_select_value('f_moderator_logs', 'COUNT(DISTINCT l_by)') < 5000) {
                $members = list_to_map('l_by', $GLOBALS['FORUM_DB']->query_select('f_moderator_logs', ['l_by', 'COUNT(*) AS cnt'], [], 'GROUP BY l_by ORDER BY cnt DESC'));
                foreach ($members as $member) {
                    $username = $GLOBALS['FORUM_DRIVER']->get_username($member['l_by'], false, USERNAME_DEFAULT_ID_RAW);
                    $_member_choice_list[$member['l_by']] = [$username, $member['cnt']];
                }
            }
        }
        if ($GLOBALS['SITE_DB']->query_select_value('actionlogs', 'COUNT(DISTINCT member_id)') < 5000) {
            $_staff = list_to_map('member_id', $GLOBALS['SITE_DB']->query_select('actionlogs', ['member_id', 'COUNT(*) AS cnt'], [], 'GROUP BY member_id ORDER BY COUNT(*) DESC'));
            foreach ($_staff as $staff) {
                $username = $GLOBALS['FORUM_DRIVER']->get_username($staff['member_id'], false, USERNAME_DEFAULT_ID_RAW);
                if (!array_key_exists($staff['member_id'], $_member_choice_list)) {
                    $_member_choice_list[$staff['member_id']] = [$username, $staff['cnt']];
                } else {
                    $_member_choice_list[$staff['member_id']][1] += $staff['cnt'];
                }
            }
        }
        $member_choice_list = new Tempcode();
        $member_choice_list->attach(form_input_list_entry('', true, do_lang_tempcode('_ALL')));
        foreach ($_member_choice_list as $id => $user_actions) {
            list($username, $action_count) = $user_actions;
            $member_choice_list->attach(form_input_list_entry(strval($id), false, do_lang(($action_count == 1) ? 'ACTIONLOG_USERCOUNT_UNI' : 'ACTIONLOG_USERCOUNT', $username, integer_format($action_count, 0))));
        }
        $fields->attach(form_input_list(do_lang_tempcode('BY'), '', 'id', $member_choice_list, null, true));

        // Possible selections for action type filter
        $_action_type_list = [];
        $rows1 = (get_forum_type() == 'cns') ? $GLOBALS['FORUM_DB']->query_select('f_moderator_logs', ['DISTINCT l_the_type']) : [];
        $rows2 = $GLOBALS['SITE_DB']->query_select('actionlogs', ['DISTINCT the_type']);
        foreach ($rows1 as $row) {
            $lang = do_lang($row['l_the_type'], null, null, null, null, false);
            if ($lang !== null) {
                $_action_type_list[$row['l_the_type']] = $lang;
            }
        }
        foreach ($rows2 as $row) {
            $lang = do_lang($row['the_type'], null, null, null, null, false);
            if ($lang !== null) {
                $_action_type_list[$row['the_type']] = $lang;
            }
        }
        foreach ($_action_type_list as $lang_id => $lang) {
            $flags = get_handler_flags($lang_id);
            if (($flags & ACTIONLOG_FLAG__GDPR) != 0) {
                $_action_type_list[$lang_id] = 'GDPR: ' . $lang;
            }
        }
        cms_mb_asort($_action_type_list, SORT_NATURAL | SORT_FLAG_CASE);
        $action_type_list = new Tempcode();
        $action_type_list->attach(form_input_list_entry('', true, do_lang_tempcode('_ALL')));
        foreach ($_action_type_list as $lang_id => $lang) {
            $action_type_list->attach(form_input_list_entry($lang_id, false, protect_from_escaping($lang)));
        }
        $fields->attach(form_input_list(do_lang_tempcode('ACTION'), '', 'to_type', $action_type_list, null, false, false));

        // Filters
        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '750bb028b5741a56f5220ff4cbd96f8f', 'TITLE' => do_lang_tempcode('FILTER')]));
        $fields->attach(form_input_line(do_lang_tempcode('PARAMETER_A'), '', 'param_a', '', false));
        $fields->attach(form_input_line(do_lang_tempcode('PARAMETER_B'), '', 'param_b', '', false));

        $post_url = build_url(['page' => '_SELF', 'type' => 'list'], '_SELF', [], false, true);
        $submit_name = do_lang_tempcode('VIEW_ACTIONLOGS');

        return do_template('FORM_SCREEN', [
            '_GUID' => 'f2c6eda24e0e973aa7e253054f6683a5',
            'GET' => true,
            'SKIP_WEBSTANDARDS' => true,
            'HIDDEN' => '',
            'TITLE' => $this->title,
            'TEXT' => '',
            'URL' => $post_url,
            'FIELDS' => $fields,
            'SUBMIT_ICON' => 'buttons/proceed',
            'SUBMIT_NAME' => $submit_name,
        ]);
    }

    /**
     * The UI to show a results table of moderation actions for a moderator.
     *
     * @return Tempcode The UI
     */
    public function choose_action() : object
    {
        $id = get_param_integer('id', null);
        $start = get_param_integer('start', 0);
        $max = get_param_integer('max', 50);
        $sortables = ['date_and_time' => do_lang_tempcode('DATE_TIME'), 'the_type' => do_lang_tempcode('ACTION')];
        $test = explode(' ', get_param_string('sort', 'date_and_time DESC', INPUT_FILTER_GET_COMPLEX), 2);
        if (count($test) == 1) {
            $test[1] = 'DESC';
        }
        list($sortable, $sort_order) = $test;
        if (((cms_strtoupper_ascii($sort_order) != 'ASC') && (cms_strtoupper_ascii($sort_order) != 'DESC')) || (!array_key_exists($sortable, $sortables))) {
            log_hack_attack_and_exit('ORDERBY_HACK');
        }

        require_code('templates_results_table');
        $field_titles = [
            do_lang_tempcode('BY'),
            do_lang_tempcode('DATE_TIME'),
            do_lang_tempcode('ACTION'),
            do_lang_tempcode('DETAILS'),
            null,
        ];
        if (addon_installed('securitylogging')) {
            $field_titles[] = do_lang_tempcode('BANNED');
        }
        $header_row = results_header_row($field_titles, $sortables, 'sort', $sortable . ' ' . $sort_order);

        $filter_to_type = get_param_string('to_type', '');
        $filter_param_a = get_param_string('param_a', '', INPUT_FILTER_GET_COMPLEX);
        $filter_param_b = get_param_string('param_b', '', INPUT_FILTER_GET_COMPLEX);

        $max_rows = 0;

        // Pull up our rows: forum...

        if (get_forum_type() == 'cns') {
            // Possible filter (called up by URL)
            $where = '1=1';
            if ($filter_to_type != '') {
                $where .= ' AND (1=2';
                foreach (explode(',', $filter_to_type) as $_filter_to_type) {
                    $where .= ' OR ' . db_string_equal_to('l_the_type', $_filter_to_type);
                }
                $where .= ')';
            }
            if ($filter_param_a != '') {
                if (is_numeric($filter_param_a)) {
                    $where .= ' AND ' . db_string_equal_to('l_param_a', $filter_param_a);
                } else {
                    $where .= ' AND l_param_a LIKE \'' . db_encode_like('%' . $filter_param_a . '%') . '\'';
                }
            }
            if ($filter_param_b != '') {
                if (is_numeric($filter_param_b)) {
                    $where .= ' AND ' . db_string_equal_to('l_param_b', $filter_param_b);
                } else {
                    $where .= ' AND l_param_b LIKE \'' . db_encode_like('%' . $filter_param_b . '%') . '\'';
                }
            }
            if ($id !== null) {
                $where .= ' AND l_by=' . strval($id);
            }

            // Fetch
            $rows1 = $GLOBALS['FORUM_DB']->query('SELECT l_reason,id,l_by AS member_id,l_date_and_time AS date_and_time,l_the_type AS the_type,l_param_a AS param_a,l_param_b AS param_b FROM ' . $GLOBALS['FORUM_DB']->get_table_prefix() . 'f_moderator_logs WHERE ' . $where . ' ORDER BY ' . $sortable . ' ' . $sort_order, $max + $start, 0, false, true);
            $max_rows += $GLOBALS['FORUM_DB']->query_value_if_there('SELECT COUNT(*) FROM ' . $GLOBALS['FORUM_DB']->get_table_prefix() . 'f_moderator_logs WHERE ' . $where, false, true);
        } else {
            $rows1 = [];
        }

        // Pull up our rows: site...

        // Possible filter (called up by URL)
        $where = '1=1';
        if ($filter_to_type != '') {
            $where .= ' AND (1=2';
            foreach (explode(',', $filter_to_type) as $_filter_to_type) {
                $where .= ' OR ' . db_string_equal_to('the_type', $_filter_to_type);
            }
            $where .= ')';
        }
        if ($filter_param_a != '') {
            if (is_numeric($filter_param_a)) {
                $where .= ' AND ' . db_string_equal_to('param_a', $filter_param_a);
            } else {
                $where .= ' AND param_a LIKE \'' . db_encode_like('%' . $filter_param_a . '%') . '\'';
            }
        }
        if ($filter_param_b != '') {
            if (is_numeric($filter_param_b)) {
                $where .= ' AND ' . db_string_equal_to('param_b', $filter_param_b);
            } else {
                $where .= ' AND param_b LIKE \'' . db_encode_like('%' . $filter_param_b . '%') . '\'';
            }
        }
        if ($id !== null) {
            $where .= ' AND member_id=' . strval($id);
        }

        // Fetch
        $rows2 = $GLOBALS['SITE_DB']->query('SELECT id,member_id,date_and_time,the_type,param_a,param_b,ip FROM ' . get_table_prefix() . 'actionlogs WHERE ' . $where . ' ORDER BY ' . $sortable . ' ' . $sort_order, $max + $start, 0, false, true);
        $max_rows += $GLOBALS['SITE_DB']->query_value_if_there('SELECT COUNT(*) FROM ' . get_table_prefix() . 'actionlogs WHERE ' . $where, false, true);
        $rows = array_merge($rows1, $rows2);

        // Render...

        $result_entries = new Tempcode();
        $pos = 0;
        while ((!empty($rows)) && (($pos - $start) < $max)) {
            $_best = mixed();

            $best = 0; // Initialise type to integer
            $_best = 0; // Initialise type to integer
            $best = null;
            $_best = null;
            foreach ($rows as $x => $row) {
                if (($best === null)
                    || (($sortable == 'date_and_time') && ($sort_order == 'ASC') && ($row['date_and_time'] < $_best))
                    || (($sortable == 'date_and_time') && ($sort_order == 'DESC') && ($row['date_and_time'] > $_best))
                    || (($sortable == 'the_type') && ($sort_order == 'ASC') && (strcmp($row['the_type'], $_best) < 0))
                    || (($sortable == 'the_type') && ($sort_order == 'DESC') && (strcmp($row['the_type'], $_best) > 0))
                ) {
                    $best = $x;
                    if ($sortable == 'date_and_time') {
                        $_best = $row['date_and_time'];
                    }
                    if ($sortable == 'the_type') {
                        $_best = $row['the_type'];
                    }
                }
            }
            if ($pos >= $start) {
                $myrow = $rows[$best];

                $username = $GLOBALS['FORUM_DRIVER']->member_profile_hyperlink($myrow['member_id'], '', false);

                $mode = array_key_exists('l_reason', $myrow) ? 'cns' : 'cms';
                $url = build_url(['page' => '_SELF', 'type' => 'view', 'id' => $myrow['id'], 'mode' => $mode], '_SELF');
                $mode_nice = ($mode == 'cms') ? 'Composr' : 'Conversr';
                $date = hyperlink($url, get_timezoned_date_time($myrow['date_and_time']), false, true, '#' . strval($myrow['id']), null, null, null, '_top');

                if ($myrow['param_a'] !== null) {
                    $a = $myrow['param_a'];
                } else {
                    $a = '';
                }
                if ($myrow['param_b'] !== null) {
                    $b = $myrow['param_b'];
                } else {
                    $b = '';
                }

                require_code('templates_tooltip');
                $crop_length_a = 12;
                $crop_length_b = 15;
                $_a = generate_tooltip_by_truncation($a, ($b == '') ? ($crop_length_a + $crop_length_b + 3/*A bit of extra tolerance*/) : $crop_length_a);
                $_b = ($b == '') ? null : generate_tooltip_by_truncation($b, $crop_length_b);

                $type_str = do_lang($myrow['the_type'], $_a, $_b, null, null, false);
                if ($type_str === null) {
                    $type_str = $myrow['the_type'];
                }

                $test = actionlog_linkage($myrow, $crop_length_a, $crop_length_b);
                if ($test !== null) {
                    list($_a, $_b) = $test;
                }

                $result_entry = [$username, $date, $type_str, $_a, $_b];

                if (addon_installed('securitylogging')) {
                    $banned_test_1 = array_key_exists('ip', $myrow) ? ip_banned($myrow['ip'], true) : false;
                    $banned_test_2 = ($GLOBALS['SITE_DB']->query_select_value_if_there('usersubmitban_member', 'the_member', ['the_member' => $myrow['member_id']]) !== null);
                    $banned_test_3 = $GLOBALS['FORUM_DRIVER']->is_banned($myrow['member_id']);
                    $banned = (((!$banned_test_1)) && ((!$banned_test_2)) && (!$banned_test_3)) ? do_lang_tempcode('NO') : do_lang_tempcode('YES');

                    $result_entry[] = $banned;
                }

                $result_entries->attach(results_entry($result_entry, true));
            }

            unset($rows[$best]);
            $pos++;
        }
        $table = results_table(do_lang_tempcode('ACTIONS'), $start, 'start', $max, 'max', $max_rows, $header_row, $result_entries, $sortables, $sortable, $sort_order, 'sort');

        $tpl = do_template('ACTIONLOGS_SCREEN', ['_GUID' => '5a2e2876c8cfa442a2266d01fca5d79a', 'TABLE' => $table, 'TITLE' => $this->title]);

        require_code('templates_internalise_screen');
        return internalise_own_screen($tpl);
    }

    /**
     * The UI to view details of a specific moderation action.
     *
     * @return Tempcode The UI
     */
    public function view_action() : object
    {
        $mode = get_param_string('mode', 'cns');
        $id = get_param_integer('id');

        if ($mode == 'cns') {
            $rows = $GLOBALS['FORUM_DB']->query_select('f_moderator_logs', ['l_reason AS reason', 'id', 'l_by AS member_id', 'l_date_and_time AS date_and_time', 'l_the_type AS the_type', 'l_param_a AS param_a', 'l_param_b AS param_b', 'l_warning_id AS warning_id'], ['id' => $id], '', 1);
        } else {
            $rows = $GLOBALS['SITE_DB']->query_select('actionlogs', ['id', 'member_id', 'date_and_time', 'the_type', 'param_a', 'param_b', 'ip', 'warning_id'], ['id' => $id], '', 1);
        }

        if (!array_key_exists(0, $rows)) {
            warn_exit(do_lang_tempcode('MISSING_RESOURCE'));
        }
        $row = $rows[0];

        $username = $GLOBALS['FORUM_DRIVER']->get_username($row['member_id']);

        $type_str = do_lang($row['the_type'], $row['param_a'], $row['param_b'], null, null, false);
        if ($type_str === null) {
            $type_str = $row['the_type'];
        }

        $fields = [
            'BY' => $GLOBALS['FORUM_DRIVER']->member_profile_hyperlink($row['member_id'], '', false),
            'DATE_TIME' => get_timezoned_date_time($row['date_and_time']),
            'ACTION' => $type_str,
            'PARAMETER_A' => ($row['param_a'] === null) ? '' : $row['param_a'],
            'PARAMETER_B' => ($row['param_b'] === null) ? '' : $row['param_b'],
        ];

        $test = actionlog_linkage($row);
        if ($test !== null) {
            $fields['DETAILS'] = $test[0];
        }

        if (array_key_exists('ip', $row)) {
            $fields['IP_ADDRESS'] = escape_html($row['ip']);
        }

        if (array_key_exists('reason', $row)) {
            $fields['REASON'] = escape_html($row['reason']);
        }

        if (addon_installed('securitylogging')) {
            if (array_key_exists('ip', $row)) {
                $banned_test_1 = ip_banned($row['ip'], true);
                $fields['IP_BANNED'] = (!$banned_test_1) ? do_lang_tempcode('NO') : do_lang_tempcode('YES');
                if ($row['ip'] != get_ip_address()) {
                    $fields['IP_BANNED']->attach(do_template('ACTIONLOGS_TOGGLE_LINK', [
                        '_GUID' => 'eff2890f2193ece32df8ec8ee48b252d',
                        'URL' => build_url(['page' => 'admin_ip_ban', 'type' => 'toggle_ip_ban', 'id' => $row['ip'], 'redirect' => protect_url_parameter(SELF_REDIRECT)], get_module_zone('admin_ip_ban')),
                    ]));
                }

                if ($row['ip'] != get_ip_address()) {
                    if ((get_option('stopforumspam_api_key') != '') || (get_option('spam_use_tornevall') == '1')) {
                        $fields['SYNDICATE_TO_STOPFORUMSPAM'] = do_template('ACTIONLOGS_TOGGLE_LINK', [
                            '_GUID' => '7d10045c6b3b48f256e2f8eb5535809c',
                            'LABEL' => do_lang_tempcode('PROCEED'),
                            'LONELY' => true,
                            'URL' => build_url(['page' => 'admin_ip_ban', 'type' => 'syndicate_ip_ban', 'ip' => $row['ip'], 'member_id' => $row['member_id'], 'reason' => do_lang('BANNED_ADDRESSES'), 'redirect' => protect_url_parameter(SELF_REDIRECT)], get_module_zone('admin_ip_ban')),
                        ]);
                    }
                }
            }

            if ((!is_guest($row['member_id'])) && ($row['member_id'] != get_member())) {
                $fields['SUBMITTER_BANNED'] = do_template('ACTIONLOGS_TOGGLE_LINK', [
                    '_GUID' => 'f79fb00ef35d89381371a67bc9c4d69b',
                    'LONELY' => true,
                    'URL' => build_url(['page' => 'admin_ip_ban', 'type' => 'toggle_submitter_ban', 'id' => $row['member_id'], 'redirect' => protect_url_parameter(SELF_REDIRECT)], get_module_zone('admin_ip_ban')),
                ]);
            } else {
                $banned_test_2 = $GLOBALS['SITE_DB']->query_select_value_if_there('usersubmitban_member', 'the_member', ['the_member' => $row['member_id']]);
                $fields['SUBMITTER_BANNED'] = ($banned_test_2 === null) ? do_lang_tempcode('NO') : do_lang_tempcode('YES');
            }

            if (((get_forum_type() == 'cns') && (!is_guest($row['member_id']))) && ($row['member_id'] != get_member())) {
                $fields['MEMBER_BANNED'] = do_template('ACTIONLOGS_TOGGLE_LINK', [
                    '_GUID' => '6b192ecfad1afc67bb8c2f1e744cc3b1',
                    'LONELY' => true,
                    'URL' => build_url(['page' => 'admin_ip_ban', 'type' => 'toggle_member_ban', 'id' => $row['member_id'], 'redirect' => protect_url_parameter(SELF_REDIRECT)], get_module_zone('admin_ip_ban')),
                ]);
            } else {
                $banned_test_3 = $GLOBALS['FORUM_DRIVER']->is_banned($row['member_id']);
                $fields['MEMBER_BANNED'] = $banned_test_3 ? do_lang_tempcode('YES') : do_lang_tempcode('NO');
            }
        }
        $fields['INVESTIGATE_USER'] = hyperlink(build_url(['page' => 'admin_lookup', 'type' => 'results', 'param' => (array_key_exists('ip', $row)) ? $row['ip'] : $row['member_id']], '_SELF'), do_lang_tempcode('PROCEED'), false, false);

        if ($row['warning_id'] !== null) {
            $fields['MODULE_TRANS_NAME_warnings'] = hyperlink(build_url(['page' => 'warnings', 'type' => 'view', 'id' => $row['warning_id'], 'redirect' => protect_url_parameter(SELF_REDIRECT)]), do_lang_tempcode('WARNING_NUMBER', strval($row['warning_id'])), false, false);
        }

        // Is there a revision here?
        require_code('revisions_engine_database');
        $revision_engine = new RevisionEngineDatabase($mode == 'cns');
        $revision = $revision_engine->find_revision_for_log($id);
        if ($revision === null) {
            require_code('revisions_engine_files');
            $revision_engine = new RevisionEngineFiles();
            $revision = $revision_engine->find_revision_for_log($id);
        }
        if ($revision !== null) {
            if (isset($revision['r_resource_type'], $revision['r_resource_id'])) {
                require_code('content');
                list($content_title, , , , $content_url) = content_get_details($revision['r_resource_type'], $revision['r_resource_id']);
                if (cms_empty_safe($content_title)) {
                    $content_title = $revision['r_original_title'];
                }
                if ($content_url !== null) {
                    $fields['VIEW'] = hyperlink($content_url, $content_title, false, true);
                }
            }

            if (isset($revision['r_original_content_owner'])) {
                $fields['CONTENT_OWNER'] = $GLOBALS['FORUM_DRIVER']->member_profile_hyperlink($revision['r_original_content_owner']);
            }

            if (isset($revision['r_original_content_timestamp'])) {
                $fields['CONTENT_DATE_TIME'] = get_timezoned_date_time($revision['r_original_content_timestamp']);
            }

            if (isset($revision['r_original_title'])) {
                if ($revision['r_original_title'] != '') {
                    $fields['TITLE_PRIOR_TO_REVISION'] = $revision['r_original_title'];
                }
            }

            $fields['TEXT_PRIOR_TO_REVISION'] = with_whitespace($revision['r_original_text'], true);

            if (isset($revision['r_original_resource_fs_path'])) {
                $fields['RESOURCE_FS_PATH_PRIOR_TO_REVISION'] = $revision['r_original_resource_fs_path'];
            }

            if (isset($revision['r_original_resource_fs_record']) && strlen($revision['r_original_resource_fs_record']) < 1024 * 50/*50kb reasonable limit*/) {
                $fields['RESOURCE_FS_RECORD_PRIOR_TO_REVISION'] = with_whitespace($revision['r_original_resource_fs_record'], true);
            }

            if (has_privilege(get_member(), 'delete_revisions')) {
                $delete_url = build_url(['page' => 'admin_revisions', 'type' => 'delete', 'id' => $revision['id'], 'revision_type' => $revision['revision_type'], 'redirect' => protect_url_parameter(SELF_REDIRECT)], get_module_zone('admin_revisions'));
                $delete = hyperlink($delete_url, do_lang_tempcode('DELETE'), false, false, do_lang_tempcode('DELETE_REVISION'), null, new Tempcode());
                $fields['DELETE_REVISION'] = $delete;
            }

            /*if (has_privilege(get_member(), 'undo_revisions')) {
                $undo_url = build_url(['page' => 'admin_revisions', 'type' => 'undo', 'id' => $revision['id'], 'revision_type' => $revision['revision_type'], 'redirect' => protect_url_parameter(SELF_REDIRECT]), get_module_zone('admin_revisions'));
                $delete = hyperlink($delete_url, do_lang_tempcode('UNDO'), false, false, do_lang_tempcode('UNDO_REVISION'), null, new Tempcode());
                $fields['UNDO_REVISION'] = $delete;
            }*/
        }

        require_code('templates_map_table');
        return map_table_screen($this->title, $fields, true);
    }
}
