<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    chat
 */

require_code('crud_module');

/**
 * Module page class.
 */
class Module_admin_chat extends Standard_crud_module
{
    protected $lang_type = 'CHATROOM';
    protected $select_name = 'NAME';
    protected $author = 'Philip Withnall';
    protected $archive_entry_point = '_SEARCH:chat';
    protected $archive_label = 'CHAT_LOBBY';
    protected $view_entry_point = '_SEARCH:chat:room:_ID';
    protected $permission_module = 'chat';
    protected $menu_label = 'SECTION_CHAT';

    /**
     * Find entry-points available within this module.
     *
     * @param  boolean $check_perms Whether to check permissions
     * @param  ?MEMBER $member_id The member to check permissions as (null: current user)
     * @param  boolean $support_crosslinks Whether to allow cross links to other modules (identifiable via a full-page-link rather than a screen-name)
     * @param  boolean $be_deferential Whether to avoid any entry-point (or even return null to disable the page in the Sitemap) if we know another module, or page_group, is going to link to that entry-point. Note that "!" and "browse" entry points are automatically merged with container page nodes (likely called by page-groupings) as appropriate.
     * @return ?array A map of entry points (screen-name=>language-code/string or screen-name=>[language-code/string, icon-theme-image]) (null: disabled)
     */
    public function get_entry_points(bool $check_perms = true, ?int $member_id = null, bool $support_crosslinks = true, bool $be_deferential = false) : ?array
    {
        if (!addon_installed('chat')) {
            return null;
        }

        $ret = [
            'browse' => ['MANAGE_CHATROOMS', 'menu/social/chat/chat'],
        ];
        $ret += parent::get_entry_points();
        $ret += [
            'delete_all' => ['DELETE_ALL_CHATROOMS', 'admin/delete3'],
        ];
        return $ret;
    }

    public $title;

    /**
     * Module pre-run function. Allows us to know metadata for <head> before we start streaming output.
     *
     * @param  boolean $top_level Whether this is running at the top level, prior to having sub-objects called
     * @param  ?ID_TEXT $type The screen type to consider for metadata purposes (null: read from environment)
     * @return ?Tempcode Tempcode indicating some kind of exceptional output (null: none)
     */
    public function pre_run(bool $top_level = true, ?string $type = null) : ?object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('chat', $error_msg)) {
            return $error_msg;
        }

        $type = get_param_string('type', 'browse');

        require_lang('chat');

        set_helper_panel_tutorial('tut_chat');

        if ($type == 'browse') {
            $also_url = build_url(['page' => 'cms_chat'], get_module_zone('cms_chat'));
            attach_message(do_lang_tempcode('menus:ALSO_SEE_CMS', escape_html($also_url->evaluate())), 'inform', true);
        }

        if ($type == 'delete_all' || $type == '_delete_all') {
            $this->title = get_screen_title('DELETE_ALL_CHATROOMS');
        }

        return parent::pre_run($top_level);
    }

    /**
     * Standard crud_module run_start.
     *
     * @param  ID_TEXT $type The type of module execution
     * @return Tempcode The output of the run
     */
    public function run_start(string $type) : object
    {
        $this->extra_donext_entries = [
            ['admin/delete3', ['_SELF', ['type' => 'delete_all'], '_SELF'], do_lang('DELETE_ALL_CHATROOMS')],
        ];

        require_code('chat');
        require_code('chat2');
        require_css('chat');

        if ($type == 'browse') {
            return $this->browse();
        }
        if ($type == 'delete_all') {
            return $this->delete_all();
        }
        if ($type == '_delete_all') {
            return $this->_delete_all();
        }
        return new Tempcode();
    }

    /**
     * The do-next manager for before content management.
     *
     * @return Tempcode The UI
     */
    public function browse() : object
    {
        $this->add_one_label = do_lang_tempcode('ADD_CHATROOM');
        $this->edit_this_label = do_lang_tempcode('EDIT_THIS_CHATROOM');
        $this->edit_one_label = do_lang_tempcode('EDIT_CHATROOM');

        require_code('templates_donext');
        return do_next_manager(
            get_screen_title('MANAGE_CHATROOMS'),
            comcode_lang_string('DOC_CHAT'),
            [
                ['admin/add', ['_SELF', ['type' => 'add'], '_SELF'], do_lang('ADD_CHATROOM')],
                ['admin/edit', ['_SELF', ['type' => 'edit'], '_SELF'], do_lang('EDIT_CHATROOM')],
                ['admin/delete3', ['_SELF', ['type' => 'delete_all'], '_SELF'], do_lang('DELETE_ALL_CHATROOMS')],
            ],
            do_lang('MANAGE_CHATROOMS')
        );
    }

    /**
     * Get Tempcode for an adding form.
     *
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function get_form_fields_for_add()
    {
        return $this->get_form_fields();
    }

    /**
     * Get Tempcode for a adding/editing form.
     *
     * @return array A pair: The input fields, Hidden fields
     */
    public function get_form_fields() : array
    {
        list($fields, $hidden) = get_chatroom_fields();

        // Permissions
        $fields->attach($this->get_permission_fields(null, null, true));
        if (addon_installed('ecommerce')) {
            require_code('ecommerce_permission_products');
            $fields->attach(permission_product_form('chat'));
        }

        return [$fields, $hidden];
    }

    /**
     * Standard crud_module list function.
     *
     * @return Tempcode The selection list
     */
    public function create_selection_list_entries() : object
    {
        require_code('chat_lobby');

        $rows = $GLOBALS['SITE_DB']->query_select('chat_rooms', ['*'], ['is_im' => 0], 'ORDER BY room_name DESC', intval(get_option('general_safety_listing_limit')));
        if (count($rows) == intval(get_option('general_safety_listing_limit'))) {
            warn_exit(do_lang_tempcode('TOO_MANY_TO_CHOOSE_FROM'));
        }
        $fields = new Tempcode();
        foreach ($rows as $row) {
            if (!handle_chatroom_pruning($row)) {
                $fields->attach(form_input_list_entry(strval($row['id']), false, $row['room_name']));
            }
        }

        return $fields;
    }

    /**
     * Standard crud_module edit form filler.
     *
     * @param  ID_TEXT $id The entry being edited
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function fill_in_edit_form(string $id)
    {
        $rows = $GLOBALS['SITE_DB']->query_select('chat_rooms', ['*'], ['id' => intval($id)], '', 1);
        if (!array_key_exists(0, $rows)) {
            warn_exit(do_lang_tempcode('MISSING_RESOURCE', 'chat'));
        }
        $row = $rows[0];

        $allow2 = $row['allow_list'];
        $allow2_groups = $row['allow_list_groups'];
        $disallow2 = $row['disallow_list'];
        $disallow2_groups = $row['disallow_list_groups'];
        if ($row['room_owner'] !== null) {
            $username = $GLOBALS['FORUM_DRIVER']->get_username($row['room_owner'], false, USERNAME_DEFAULT_BLANK);
        } else {
            $username = '';
        }

        list($fields, $hidden) = get_chatroom_fields(intval($id), false, $row['room_name'], get_translated_text($row['c_welcome']), $username, $allow2, $allow2_groups, $disallow2, $disallow2_groups);

        // Permissions
        $fields->attach($this->get_permission_fields($id));
        if (addon_installed('ecommerce')) {
            require_code('ecommerce_permission_products');
            $fields->attach(permission_product_form('chat', $id));
        }

        $delete_fields = new Tempcode();
        $logs_url = build_url(['page' => 'chat', 'type' => 'download_logs', 'id' => $id], get_module_zone('chat'));
        $delete_fields->attach(form_input_tick(do_lang_tempcode('DELETE'), do_lang_tempcode('DESCRIPTION_DELETE_CHATROOM', escape_html($logs_url->evaluate())), 'delete', false));

        return [$fields, $hidden, $delete_fields, null, true];
    }

    /**
     * Standard crud_module add actualiser.
     *
     * @return array A pair: The entry added, description about usage
     */
    public function add_actualisation() : array
    {
        list($allow2, $allow2_groups, $disallow2, $disallow2_groups) = read_in_chat_perm_fields();

        $metadata = actual_metadata_get_fields('chat', null);

        $id = add_chatroom(post_param_string('c_welcome'), post_param_string('room_name'), $GLOBALS['FORUM_DRIVER']->get_member_from_username(post_param_string('room_owner')), $allow2, $allow2_groups, $disallow2, $disallow2_groups, post_param_string('room_lang', user_lang()));

        set_url_moniker('chat', strval($id));

        $this->set_permissions(strval($id));
        if (addon_installed('ecommerce')) {
            require_code('ecommerce_permission_products');
            permission_product_save('chat', strval($id));
        }

        if (addon_installed('content_reviews')) {
            content_review_set('chat', strval($id));
        }

        return [strval($id), null];
    }

    /**
     * Standard crud_module edit actualiser.
     *
     * @param  ID_TEXT $id The entry being edited
     * @return ?Tempcode Description about usage (null: none)
     */
    public function edit_actualisation(string $id) : ?object
    {
        $_room_owner = post_param_string('room_owner', STRING_MAGIC_NULL);
        $room_owner = ($_room_owner == STRING_MAGIC_NULL) ? INTEGER_MAGIC_NULL : $GLOBALS['FORUM_DRIVER']->get_member_from_username($_room_owner);
        if ($_room_owner != STRING_MAGIC_NULL) {
            list($allow2, $allow2_groups, $disallow2, $disallow2_groups) = read_in_chat_perm_fields();
        } else {
            $allow2 = STRING_MAGIC_NULL;
            $allow2_groups = STRING_MAGIC_NULL;
            $disallow2 = STRING_MAGIC_NULL;
            $disallow2_groups = STRING_MAGIC_NULL;
        }

        $metadata = actual_metadata_get_fields('chat', $id);

        edit_chatroom(intval($id), post_param_string('c_welcome', STRING_MAGIC_NULL), post_param_string('room_name'), $room_owner, $allow2, $allow2_groups, $disallow2, $disallow2_groups, post_param_string('room_lang', STRING_MAGIC_NULL));

        $this->set_permissions($id);
        if (addon_installed('ecommerce')) {
            require_code('ecommerce_permission_products');
            permission_product_save('chat', strval($id));
        }

        if (addon_installed('content_reviews')) {
            content_review_set('chat', $id);
        }

        return null;
    }

    /**
     * Standard crud_module delete actualiser.
     *
     * @param  ID_TEXT $id The entry being deleted
     */
    public function delete_actualisation(string $id)
    {
        delete_chatroom(intval($id));
    }

    /**
     * The UI to delete all chatrooms.
     *
     * @return Tempcode The UI
     */
    public function delete_all() : object
    {
        $fields = new Tempcode();
        $fields->attach(form_input_tick(do_lang_tempcode('PROCEED'), do_lang_tempcode('Q_SURE'), 'continue_delete', false));
        $posting_name = do_lang_tempcode('PROCEED');
        $posting_url = build_url(['page' => '_SELF', 'type' => '_delete_all'], '_SELF');
        $text = paragraph(do_lang_tempcode('CONFIRM_DELETE_ALL_CHATROOMS'));
        return do_template('FORM_SCREEN', [
            '_GUID' => 'fdf02f5b3a3b9ce6d1abaccf0970ed73',
            'SKIP_WEBSTANDARDS' => true,
            'HIDDEN' => '',
            'TITLE' => $this->title,
            'FIELDS' => $fields,
            'SUBMIT_ICON' => 'admin/delete3',
            'SUBMIT_NAME' => $posting_name,
            'URL' => $posting_url,
            'TEXT' => $text,
        ]);
    }

    /**
     * The actualiser to delete all chatrooms.
     *
     * @return Tempcode The UI
     */
    public function _delete_all() : object
    {
        $delete = post_param_integer('continue_delete', 0);
        if ($delete != 1) {
            $url = build_url(['page' => '_SELF', 'type' => 'browse'], '_SELF');
            return redirect_screen($this->title, $url, do_lang_tempcode('CANCELLED'));
        }

        delete_all_chatrooms();

        return $this->do_next_manager($this->title, do_lang_tempcode('SUCCESS'));
    }
}
