<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_cleanup_tools
 */

/**
 * Module page class.
 */
class Module_admin_cleanup
{
    /**
     * Find details of the module.
     *
     * @return ?array Map of module info (null: module is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 3;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'core_cleanup_tools';
        return $info;
    }

    /**
     * Find entry-points available within this module.
     *
     * @param  boolean $check_perms Whether to check permissions
     * @param  ?MEMBER $member_id The member to check permissions as (null: current user)
     * @param  boolean $support_crosslinks Whether to allow cross links to other modules (identifiable via a full-page-link rather than a screen-name)
     * @param  boolean $be_deferential Whether to avoid any entry-point (or even return null to disable the page in the Sitemap) if we know another module, or page_group, is going to link to that entry-point. Note that "!" and "browse" entry points are automatically merged with container page nodes (likely called by page-groupings) as appropriate.
     * @return ?array A map of entry points (screen-name=>language-code/string or screen-name=>[language-code/string, icon-theme-image]) (null: disabled)
     */
    public function get_entry_points(bool $check_perms = true, ?int $member_id = null, bool $support_crosslinks = true, bool $be_deferential = false) : ?array
    {
        return [
            'browse' => ['CLEANUP_TOOLS', 'menu/adminzone/tools/cleanup'],
        ];
    }

    public $title;

    /**
     * Module pre-run function. Allows us to know metadata for <head> before we start streaming output.
     *
     * @return ?Tempcode Tempcode indicating some kind of exceptional output (null: none)
     */
    public function pre_run() : ?object
    {
        require_code('form_templates'); // Needs to run high so that the anti-click-hacking header is sent

        $type = get_param_string('type', 'browse');

        require_lang('cleanup');

        set_helper_panel_tutorial('tut_website_health');

        if ($type == 'rebuild') {
            breadcrumb_set_parents([['_SELF:_SELF:browse', do_lang_tempcode('CLEANUP_TOOLS')]]);
            breadcrumb_set_self(do_lang_tempcode('DONE'));
        }

        $this->title = get_screen_title('CLEANUP_TOOLS');

        return null;
    }

    /**
     * Execute the module.
     *
     * @return Tempcode The result of execution
     */
    public function run() : object
    {
        require_code('caches3');

        // Decide what we're doing
        $type = get_param_string('type', 'browse');

        if ($type == 'browse') {
            return $this->choose_cache_type();
        }
        if ($type == 'rebuild') {
            return $this->do_rebuild();
        }

        return new Tempcode();
    }

    /**
     * The UI for choosing caches to empty.
     *
     * @return Tempcode The UI
     */
    public function choose_cache_type() : object
    {
        $url = build_url(['page' => '_SELF', 'type' => 'rebuild'], '_SELF');

        $checkbox_ids = [];

        $_fields_cache = [];
        $fields_cache_expand = true;

        $_fields_optimise = [];
        $fields_optimise_expand = false;

        $_fields_other = [];
        $fields_other_expand = false;

        $hooks = find_all_hook_obs('systems', 'cleanup', 'Hook_cleanup_');
        foreach ($hooks as $hook => $object) {
            $output = $object->info();
            if ($output !== null) {
                $is_ticked = (get_param_string('tick', null) === $hook);
                $tick = form_input_tick($output['title'], $output['description'], 'cleanup_' . $hook, $is_ticked);
                switch ($output['type']) {
                    case 'cache':
                        $_fields_cache[$output['title']->evaluate()] = $tick;
                        $checkbox_ids[] = '#cleanup_' . $hook;
                        if ($is_ticked) {
                            $fields_cache_expand = true;
                        }
                        break;

                    case 'optimise':
                        $_fields_optimise[$output['title']->evaluate()] = $tick;
                        if ($is_ticked) {
                            $fields_optimise_expand = true;
                        }
                        break;

                    case 'other':
                        $_fields_other[$output['title']->evaluate()] = $tick;
                        if ($is_ticked) {
                            $fields_other_expand = true;
                        }
                        break;

                    default:
                }
            }
        }

        $fields = new Tempcode();

        cms_mb_ksort($_fields_cache, SORT_NATURAL | SORT_FLAG_CASE);
        $fields_cache = new Tempcode();
        foreach ($_fields_cache as $tick) {
            $fields_cache->attach($tick);
        }
        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', [
            '_GUID' => '3ddb387dba8c42ac4ef7b85621052e11',
            'SECTION_HIDDEN' =>  !$fields_cache_expand,
            'TITLE' => do_lang_tempcode('CLEANUP_PAGE_EXP_CACHES'),
            'HELP' => do_lang_tempcode('CLEANUP_PAGE_CACHES', escape_html(implode(', ', $checkbox_ids))),
        ]));
        $fields->attach($fields_cache);

        cms_mb_ksort($_fields_optimise, SORT_NATURAL | SORT_FLAG_CASE);
        $fields_optimise = new Tempcode();
        foreach ($_fields_optimise as $tick) {
            $fields_optimise->attach($tick);
        }
        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', [
            '_GUID' => '4a9d6e722f246887160c444a062a9d00',
            'SECTION_HIDDEN' =>  !$fields_optimise_expand,
            'TITLE' => do_lang_tempcode('CLEANUP_PAGE_EXP_OPTIMISERS'),
            'HELP' => do_lang_tempcode('CLEANUP_PAGE_MISC_TOOLS', escape_html(find_script('upgrader'))),
        ]));
        $fields->attach($fields_optimise);

        cms_mb_ksort($_fields_other, SORT_NATURAL | SORT_FLAG_CASE);
        $fields_other = new Tempcode();
        foreach ($_fields_other as $tick) {
            $fields_other->attach($tick);
        }
        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', [
            '_GUID' => '61eb9a53529516e6de7d761a6c060c68',
            'SECTION_HIDDEN' =>  !$fields_other_expand,
            'TITLE' => do_lang_tempcode('OTHER'),
            'HELP' => do_lang_tempcode('CLEANUP_PAGE_MISC_TOOLS', escape_html(find_script('upgrader'))),
        ]));
        $fields->attach($fields_other);

        return do_template('FORM_SCREEN', [
            '_GUID' => '85bfdf171484604594a157aa8983f920',
            'SKIP_WEBSTANDARDS' => true,
            'TEXT' => '',
            'SUBMIT_ICON' => 'menu/adminzone/tools/cleanup',
            'SUBMIT_NAME' => do_lang_tempcode('PROCEED'),
            'HIDDEN' => '',
            'TITLE' => $this->title,
            'FIELDS' => $fields,
            'URL' => $url,
        ]);
    }

    /**
     * The actualiser for emptying caches.
     *
     * @return Tempcode The UI
     */
    public function do_rebuild() : object
    {
        $hooks = find_all_hooks('systems', 'cleanup');

        // Fiddle the order a bit
        if (array_key_exists('cns_topics', $hooks)) {
            unset($hooks['cns_topics']);
            $hooks['cns_topics'] = 'sources_custom';
        }
        if (array_key_exists('cns', $hooks)) {
            unset($hooks['cns']);
            $hooks['cns'] = 'sources_custom';
        }
        if (array_key_exists('cns_members', $hooks)) {
            unset($hooks['cns_members']);
            $hooks['cns_members'] = 'sources_custom';
        }

        $todo = [];
        foreach (array_keys($hooks) as $hook) {
            if (post_param_integer('cleanup_' . $hook, 0) == 1) {
                $todo[] = $hook;
            }
        }
        $messages = composr_cleanup($todo);
        $messages->attach(paragraph(do_lang_tempcode('SUCCESS')));

        log_it('CLEANUP_TOOLS', implode(', ', $todo));

        return do_template('CLEANUP_COMPLETED_SCREEN', ['_GUID' => '598510a9ad9f01f3c0806319b32b5033', 'TITLE' => $this->title, 'MESSAGES' => $messages]);
    }
}
