<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_cns
 */

require_code('crud_module');

/**
 * Module page class.
 */
class Module_admin_cns_groups extends Standard_crud_module
{
    protected $lang_type = 'GROUP';
    protected $select_name = 'NAME';
    protected $content_type = 'group';
    protected $possibly_some_kind_of_upload = true;
    protected $output_of_action_is_confirmation = true;
    protected $do_preview = null;
    protected $archive_entry_point = '_SEARCH:groups';
    protected $archive_label = 'USERGROUPS';
    protected $view_entry_point = '_SEARCH:groups:view:_ID';
    protected $menu_label = 'USERGROUPS';
    protected $orderer = 'g_name';
    protected $donext_entry_content_type = 'group';
    protected $donext_category_content_type = null;

    public $js_function_calls = ['moduleAdminCnsGroups'];

    /**
     * Find entry-points available within this module.
     *
     * @param  boolean $check_perms Whether to check permissions
     * @param  ?MEMBER $member_id The member to check permissions as (null: current user)
     * @param  boolean $support_crosslinks Whether to allow cross links to other modules (identifiable via a full-page-link rather than a screen-name)
     * @param  boolean $be_deferential Whether to avoid any entry-point (or even return null to disable the page in the Sitemap) if we know another module, or page_group, is going to link to that entry-point. Note that "!" and "browse" entry points are automatically merged with container page nodes (likely called by page-groupings) as appropriate.
     * @return ?array A map of entry points (screen-name=>language-code/string or screen-name=>[language-code/string, icon-theme-image]) (null: disabled)
     */
    public function get_entry_points(bool $check_perms = true, ?int $member_id = null, bool $support_crosslinks = true, bool $be_deferential = false) : ?array
    {
        if (get_forum_type() != 'cns') {
            return null;
        }

        $ret = [
            'browse' => ['MANAGE_USERGROUPS', 'menu/social/groups'],
        ] + parent::get_entry_points();

        if ($support_crosslinks) {
            require_code('fields');
            $ret += manage_custom_fields_entry_points('group');
        }

        return $ret;
    }

    public $title;

    /**
     * Module pre-run function. Allows us to know metadata for <head> before we start streaming output.
     *
     * @param  boolean $top_level Whether this is running at the top level, prior to having sub-objects called
     * @param  ?ID_TEXT $type The screen type to consider for metadata purposes (null: read from environment)
     * @return ?Tempcode Tempcode indicating some kind of exceptional output (null: none)
     */
    public function pre_run(bool $top_level = true, ?string $type = null) : ?object
    {
        if (get_forum_type() != 'cns') {
            warn_exit(do_lang_tempcode('NO_CNS'));
        }

        $type = get_param_string('type', 'browse');

        require_lang('cns');
        require_css('cns_admin');
        require_javascript('core_cns');

        set_helper_panel_tutorial('tut_subcom');

        return parent::pre_run($top_level);
    }

    /**
     * Standard crud_module run_start.
     *
     * @param  ID_TEXT $type The type of module execution
     * @return Tempcode The output of the run
     */
    public function run_start(string $type) : object
    {
        cns_require_all_forum_stuff();

        require_code('cns_groups_action');
        require_code('cns_groups_action2');
        require_code('cns_groups');
        require_code('cns_groups2');

        $this->add_text = do_lang_tempcode('GROUP_TEXT');
        $this->edit_text = do_lang_tempcode('GROUP_TEXT');

        if ($type == 'add') {
            require_javascript('core_cns');
            $this->js_function_calls[] = 'moduleAdminCnsGroupsRunStart';
        }

        $this->add_one_label = do_lang_tempcode('ADD_GROUP');
        $this->edit_this_label = do_lang_tempcode('EDIT_THIS_GROUP');
        $this->edit_one_label = do_lang_tempcode('EDIT_GROUP');

        if ($type == 'browse') {
            return $this->browse();
        }
        return new Tempcode();
    }

    /**
     * The do-next manager for before content management.
     *
     * @return Tempcode The UI
     */
    public function browse() : object
    {
        require_code('templates_donext');
        require_code('fields');
        return do_next_manager(
            get_screen_title('MANAGE_USERGROUPS'),
            comcode_lang_string('DOC_GROUPS'),
            array_merge([
                ['admin/add', ['_SELF', ['type' => 'add'], '_SELF'], do_lang('ADD_GROUP')],
                ['admin/edit', ['_SELF', ['type' => 'edit'], '_SELF'], do_lang('EDIT_GROUP')],
            ], manage_custom_fields_donext_link('group')),
            do_lang('MANAGE_USERGROUPS')
        );
    }

    /**
     * Get Tempcode for an adding form.
     *
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function get_form_fields_for_add()
    {
        return $this->get_form_fields();
    }

    /**
     * Get Tempcode for adding/editing form.
     *
     * @param  ?GROUP $id The usergroup being edited (null: adding, not editing)
     * @param  SHORT_TEXT $name The usergroup name
     * @param  BINARY $is_default Whether this is a default usergroup
     * @param  BINARY $is_super_admin Whether members of the usergroup are super-administrators
     * @param  BINARY $is_super_moderator Whether members of the usergroup are super-moderators
     * @param  ID_TEXT $group_leader The username of the usergroup leader
     * @param  SHORT_TEXT $title The default title for members with this as their primary usergroup
     * @param  URLPATH $rank_image The usergroup rank image
     * @param  ?GROUP $promotion_target The target for promotion from this usergroup (null: no promotion prospects)
     * @param  ?integer $promotion_threshold The point threshold upon which promotion occurs (null: no promotion prospects)
     * @param  BINARY $promotion_approval Whether promotion requires manual approval by the usergroup lead or staff
     * @param  ?integer $flood_control_submit_secs The number of seconds between submission flood controls (null: average for existing usergroups)
     * @param  ?integer $flood_control_access_secs The number of seconds between access flood controls (null: average for existing usergroups)
     * @param  ?integer $gift_points_base The number of gift points members of this usergroup get when they start (null: average for existing usergroups)
     * @param  ?integer $gift_points_per_day The number of gift points members of this usergroup get per-day (null: average for existing usergroups)
     * @param  ?integer $max_daily_upload_mb The number of megabytes members can upload per day (null: average for existing usergroups)
     * @param  ?integer $max_attachments_per_post The maximum number of attachments members of this usergroup may have per post (null: average for existing usergroups)
     * @param  ?integer $max_avatar_width The maximum avatar width members of this usergroup may have (null: average for existing usergroups)
     * @param  ?integer $max_avatar_height The maximum avatar height members of this usergroup may have (null: average for existing usergroups)
     * @param  ?integer $max_post_length_comcode The maximum post length members of this usergroup may have (null: average for existing usergroups)
     * @param  ?integer $max_sig_length_comcode The maximum signature length members of this usergroup may have (null: average for existing usergroups)
     * @param  BINARY $enquire_on_new_ips Whether to lock out unverified IP addresses until e-mail confirmation
     * @param  BINARY $is_presented_at_install Whether the usergroup is presented for joining at joining (implies anyone may be in the usergroup, but only choosable at joining)
     * @param  BINARY $group_is_hidden Whether the name and membership of the usergroup is hidden
     * @param  ?integer $order The display order this usergroup will be given, relative to other usergroups. Lower numbered usergroups display before higher numbered usergroups (null: last).
     * @param  BINARY $rank_image_pri_only Whether the rank image will not be shown for secondary membership
     * @param  BINARY $open_membership Whether members may join this usergroup without requiring any special permission
     * @param  BINARY $is_private_club Whether this usergroup is a private club. Private clubs may be managed in the CMS zone, and do not have any special permissions - except over their own associated forum.
     * @return array A pair: The input fields, Hidden fields
     */
    public function get_form_fields(?int $id = null, string $name = '', int $is_default = 0, int $is_super_admin = 0, int $is_super_moderator = 0, string $group_leader = '', string $title = '', string $rank_image = '', ?int $promotion_target = null, ?int $promotion_threshold = null, int $promotion_approval = 0, ?int $flood_control_submit_secs = null, ?int $flood_control_access_secs = null, ?int $gift_points_base = null, ?int $gift_points_per_day = 1, ?int $max_daily_upload_mb = null, ?int $max_attachments_per_post = null, ?int $max_avatar_width = null, ?int $max_avatar_height = null, ?int $max_post_length_comcode = null, ?int $max_sig_length_comcode = null, int $enquire_on_new_ips = 0, int $is_presented_at_install = 0, int $group_is_hidden = 0, ?int $order = null, int $rank_image_pri_only = 1, int $open_membership = 0, int $is_private_club = 0) : array
    {
        if (is_on_multi_site_network()) {
            attach_message(do_lang_tempcode('EDITING_ON_WRONG_MSN'), 'warn');
        }

        $flood_control_submit_secs = take_param_int_modeavg($flood_control_submit_secs, 'g_flood_control_submit_secs', 'f_groups', 0);
        $flood_control_access_secs = take_param_int_modeavg($flood_control_access_secs, 'g_flood_control_access_secs', 'f_groups', 0);
        $max_daily_upload_mb = take_param_int_modeavg($max_daily_upload_mb, 'g_max_daily_upload_mb', 'f_groups', 70);
        $max_attachments_per_post = take_param_int_modeavg($max_attachments_per_post, 'g_max_attachments_per_post', 'f_groups', 50);
        $max_avatar_width = take_param_int_modeavg($max_avatar_width, 'g_max_avatar_width', 'f_groups', 100);
        $max_avatar_height = take_param_int_modeavg($max_avatar_height, 'g_max_avatar_height', 'f_groups', 100);
        $max_post_length_comcode = take_param_int_modeavg($max_post_length_comcode, 'g_max_post_length_comcode', 'f_groups', 30000);
        $max_sig_length_comcode = take_param_int_modeavg($max_sig_length_comcode, 'g_max_sig_length_comcode', 'f_groups', 700);
        $gift_points_base = take_param_int_modeavg($gift_points_base, 'g_gift_points_base', 'f_groups', 25);
        $gift_points_per_day = take_param_int_modeavg($gift_points_per_day, 'g_gift_points_per_day', 'f_groups', 1);

        if ($group_leader === null) {
            $group_leader = '';
        }

        $fields = new Tempcode();
        $hidden = new Tempcode();

        $fields->attach(form_input_line(do_lang_tempcode('NAME'), do_lang_tempcode('DESCRIPTION_USERGROUP_TITLE'), 'usergroup_name', $name, true));

        // This will just cause trouble if the group is then edited in cms_cns_groups, with lost data. Best to just not allow this
        /*
        if ((addon_installed('cns_clubs')) && ($id !== null)) {
            $fields->attach(form_input_tick(do_lang_tempcode('IS_PRIVATE_CLUB'), do_lang_tempcode('IS_PRIVATE_CLUB_DESCRIPTION'), 'is_private_club', $is_private_club == 1));
        }
        */

        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '9a3e18560632564111aecb8f27341fb2', 'SECTION_HIDDEN' => $title == '' && $group_leader == '', 'TITLE' => do_lang_tempcode('ADVANCED')]));

        $fields->attach(form_input_line(do_lang_tempcode('TITLE'), do_lang_tempcode('DESCRIPTION_GROUP_TITLE'), 'title', $title, false));
        $fields->attach(form_input_username(do_lang_tempcode('GROUP_LEADER'), do_lang_tempcode('DESCRIPTION_GROUP_LEADER'), 'group_leader', $group_leader, false));

        $rows = $GLOBALS['FORUM_DB']->query_select('f_groups', ['id', 'g_name', 'g_is_super_admin'], ['g_is_private_club' => 0]);

        $fields->attach(get_order_field('group', null, $order));

        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '4d72d054883ede5250a3c3e03d27d18c', 'TITLE' => do_lang_tempcode('JOINING')]));
        if (($id === null) || ($id != db_get_first_id())) {
            $fields->attach(form_input_tick(do_lang_tempcode('IS_PRESENTED_AT_INSTALL'), do_lang_tempcode('DESCRIPTION_IS_PRESENTED_AT_INSTALL'), 'is_presented_at_install', $is_presented_at_install == 1));
            if (cns_get_all_default_groups(true) === [$id]) {
                $hidden->attach(form_input_hidden('is_default', '1'));
            } else {
                $fields->attach(form_input_tick(do_lang_tempcode('DEFAULT_GROUP'), do_lang_tempcode('DESCRIPTION_IS_DEFAULT_GROUP'), 'is_default', $is_default == 1));
            }
        }
        $fields->attach(form_input_tick(do_lang_tempcode('OPEN_MEMBERSHIP'), do_lang_tempcode('OPEN_MEMBERSHIP_DESCRIPTION'), 'open_membership', $open_membership == 1));
        if (addon_installed('points')) {
            $fields->attach(form_input_tick(do_lang_tempcode('PROMOTION_APPROVAL'), do_lang_tempcode('DESCRIPTION_PROMOTION_APPROVAL'), 'promotion_approval', $promotion_approval == 1));
        }

        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '4ec20b3b67c70e1d4136432ae4fd56b6', 'SECTION_HIDDEN' => true, 'TITLE' => do_lang_tempcode('RANK')]));
        if (addon_installed('points')) {
            $promotion_target_groups = form_input_list_entry('', false, do_lang_tempcode('NA_EM'));
            foreach ($rows as $group) {
                if (($group['id'] != $id) && ($group['id'] != db_get_first_id())) {
                    $promotion_target_groups->attach(form_input_list_entry(strval($group['id']), ($group['id'] == $promotion_target), get_translated_text($group['g_name'], $GLOBALS['FORUM_DB'])));
                }
            }
            $fields->attach(form_input_list(do_lang_tempcode('PROMOTION_TARGET'), do_lang_tempcode('DESCRIPTION_PROMOTION_TARGET'), 'promotion_target', $promotion_target_groups, null, false, false));
            $fields->attach(form_input_integer(do_lang_tempcode('PROMOTION_THRESHOLD'), do_lang_tempcode('DESCRIPTION_PROMOTION_THRESHOLD'), 'promotion_threshold', $promotion_threshold, false));
        }

        require_code('themes2');
        $ids = get_all_image_ids_type('cns_rank_images', false, $GLOBALS['FORUM_DB']);

        if (get_base_url() == get_forum_base_url()) {
            $set_name = 'rank_image';
            $required = false;
            $set_title = do_lang_tempcode('RANK_IMAGE');
            $field_set = (empty($ids)) ? new Tempcode() : alternate_fields_set__start($set_name);

            require_code('images');
            $field_set->attach(form_input_upload(do_lang_tempcode('UPLOAD'), '', 'file', $required, null, null, true, get_allowed_image_file_types(IMAGE_CRITERIA_WEBSAFE)));

            $image_chooser_field = form_input_theme_image(do_lang_tempcode('STOCK'), '', 'theme_img_code', $ids, null, $rank_image, null, false, $GLOBALS['FORUM_DB']);
            $field_set->attach($image_chooser_field);

            $fields->attach(alternate_fields_set__end($set_name, $set_title, do_lang_tempcode('DESCRIPTION_RANK_IMAGE'), $field_set, $required));

            handle_max_file_size($hidden, 'image');
        } else {
            if (empty($ids)) {
                warn_exit(do_lang_tempcode('NO_SELECTABLE_THEME_IMAGES_MSN', 'cns_rank_images'));
            }

            $image_chooser_field = form_input_theme_image(do_lang_tempcode('STOCK'), '', 'theme_img_code', $ids, null, $rank_image, null, true, $GLOBALS['FORUM_DB']);
            $fields->attach($image_chooser_field);
        }

        $fields->attach(form_input_tick(do_lang_tempcode('RANK_IMAGE_PRI_ONLY'), do_lang_tempcode('RANK_IMAGE_PRI_ONLY_DESCRIPTION'), 'rank_image_pri_only', $rank_image_pri_only == 1));

        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '46bcc92ed469cdee0c16e66dda6986e7', 'SECTION_HIDDEN' => true, 'TITLE' => do_lang_tempcode('BENEFITS')]));
        $sa_descrip = do_lang_tempcode('DESCRIPTION_IS_SUPER_ADMIN');
        if ($is_super_admin == 1) {
            $sa_descrip->attach(do_lang_tempcode('DESCRIPTION_IS_SUPER_ADMIN_B'));
        }
        $fields->attach(form_input_tick(do_lang_tempcode('SUPER_ADMIN'), $sa_descrip, 'is_super_admin', $is_super_admin == 1));
        $fields->attach(form_input_tick(do_lang_tempcode('SUPER_MODERATOR'), do_lang_tempcode('DESCRIPTION_IS_SUPER_MODERATOR'), 'is_super_moderator', $is_super_moderator == 1));
        if (addon_installed('points') && get_option('enable_gift_points') == '1') {
            $fields->attach(form_input_integer(do_lang_tempcode('GIFT_POINTS_BASE'), do_lang_tempcode('DESCRIPTION_GIFT_POINTS_BASE'), 'gift_points_base', $gift_points_base, true));
            $fields->attach(form_input_integer(do_lang_tempcode('GIFT_POINTS_PER_DAY'), do_lang_tempcode('DESCRIPTION_GIFT_POINTS_PER_DAY'), 'gift_points_per_day', $gift_points_per_day, true));
        }

        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '242da621a54eb63efd528e678dbceeaa', 'SECTION_HIDDEN' => true, 'TITLE' => do_lang_tempcode('SECURITY')]));
        $fields->attach(form_input_tick(do_lang_tempcode('HIDDEN_USERGROUP'), do_lang_tempcode('DESCRIPTION_GROUP_HIDDEN'), 'hidden', $group_is_hidden == 1));
        $fields->attach(form_input_tick(do_lang_tempcode('ENQUIRE_ON_NEW_IPS'), do_lang_tempcode('DESCRIPTION_ENQUIRE_ON_NEW_IPS', is_maintained_description('two_factor', do_lang_tempcode('TWO_FACTOR_AUTHENTICATION'))), 'enquire_on_new_ips', $enquire_on_new_ips == 1));

        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => 'e894918a6e1b56fdd5f1574cd68a3a91', 'SECTION_HIDDEN' => true, 'TITLE' => do_lang_tempcode('RESTRICTIONS')]));
        $fields->attach(form_input_integer(do_lang_tempcode('MAX_ATTACHMENTS_PER_POST'), do_lang_tempcode('DESCRIPTION_MAX_ATTACHMENTS_PER_POST'), 'max_attachments_per_post', $max_attachments_per_post, true));
        $fields->attach(form_input_integer(do_lang_tempcode('MAX_DAILY_UPLOAD_MB'), do_lang_tempcode('DESCRIPTION_MAX_DAILY_UPLOAD_MB'), 'max_daily_upload_mb', $max_daily_upload_mb, true));
        if (addon_installed('cns_member_avatars')) {
            $fields->attach(form_input_dimensions(do_lang_tempcode('MAX_AVATAR_DIMENSIONS'), do_lang_tempcode('DESCRIPTION_MAX_AVATAR_DIMENSIONS'), 'max_avatar_width', 'max_avatar_height', $max_avatar_width, $max_avatar_height, true));
        }
        $fields->attach(form_input_integer(do_lang_tempcode('MAX_POST_LENGTH_COMCODE'), do_lang_tempcode('DESCRIPTION_MAX_POST_LENGTH_COMCODE'), 'max_post_length_comcode', $max_post_length_comcode, true));
        if (addon_installed('cns_signatures')) {
            $fields->attach(form_input_integer(do_lang_tempcode('MAX_SIG_LENGTH_COMCODE'), do_lang_tempcode('DESCRIPTION_MAX_SIG_LENGTH_COMCODE'), 'max_sig_length_comcode', $max_sig_length_comcode, true));
        }

        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '251cbc5b4e8a3dd600535ed29b3cd9d1', 'SECTION_HIDDEN' => true, 'TITLE' => do_lang_tempcode('FLOOD_CONTROL')]));
        $fields->attach(form_input_integer(do_lang_tempcode('FLOOD_CONTROL_ACCESS_SECS'), do_lang_tempcode('DESCRIPTION_FLOOD_CONTROL_ACCESS_SECS'), 'flood_control_access_secs', $flood_control_access_secs, true));
        $fields->attach(form_input_integer(do_lang_tempcode('FLOOD_CONTROL_SUBMIT_SECS'), do_lang_tempcode('DESCRIPTION_FLOOD_CONTROL_SUBMIT_SECS'), 'flood_control_submit_secs', $flood_control_submit_secs, true));

        $fields->attach(metadata_get_fields('group', ($id === null) ? null : strval($id), false, ['submitter']));

        if (addon_installed('content_reviews')) {
            $fields->attach(content_review_get_fields('group', ($id === null) ? null : strval($id)));
        }

        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '0fd215401ffaace7f2f9f6aa73db4ce1', 'TITLE' => do_lang_tempcode('ACTIONS')]));

        $copy_members_from_groups = new Tempcode();
        foreach ($rows as $row) {
            if (($row['id'] != db_get_first_id()) && ($row['id'] != $id)) {
                $copy_members_from_groups->attach(form_input_list_entry(strval($row['id']), false, get_translated_text($row['g_name'], $GLOBALS['FORUM_DB'])));
            }
        }
        $fields->attach(form_input_multi_list(do_lang_tempcode('COPY_MEMBERS_INTO'), do_lang_tempcode('DESCRIPTION_COPY_MEMBERS_INTO'), 'copy_members_into', $copy_members_from_groups));

        // Take permissions from
        $permissions_from_groups = new Tempcode();
        $permissions_from_groups = form_input_list_entry('', false, do_lang_tempcode('NA_EM'));
        foreach ($rows as $group) {
            if ($group['id'] != $id) {
                $permissions_from_groups->attach(form_input_list_entry(strval($group['id']), false, get_translated_text($group['g_name'], $GLOBALS['FORUM_DB'])));
            }
        }
        $fields->attach(form_input_list(do_lang_tempcode('DEFAULT_PERMISSIONS_FROM'), do_lang_tempcode(($id === null) ? 'DESCRIPTION_DEFAULT_PERMISSIONS_FROM_NEW' : 'DESCRIPTION_DEFAULT_PERMISSIONS_FROM'), 'absorb', $permissions_from_groups, null, false, false));

        $this->appended_actions_already = true;

        return [$fields, $hidden];
    }

    /**
     * Standard crud_module table function.
     *
     * @param  array $url_map Details to go to build_url for link to the next screen
     * @return array A quartet: The choose table, Whether re-ordering is supported from this screen, Search URL, Archive URL
     */
    public function create_selection_list_choose_table(array $url_map) : array
    {
        require_code('templates_results_table');

        $default_order = 'g_promotion_threshold ASC,r.id ASC';
        $current_ordering = get_param_string('sort', $default_order, INPUT_FILTER_GET_COMPLEX);
        $sortables = [
            'g_name' => do_lang_tempcode('NAME'),
            'g_is_presented_at_install' => do_lang_tempcode('IS_PRESENTED_AT_INSTALL'),
            'g_is_default' => do_lang_tempcode('DEFAULT_GROUP'),
            'g_open_membership' => do_lang_tempcode('OPEN_MEMBERSHIP'),
        ];
        if (addon_installed('points')) {
            $sortables = array_merge($sortables, [
                'g_promotion_threshold ASC,id' => do_lang_tempcode('PROMOTION_TARGET'),
            ]);
        }
        $sortables = array_merge($sortables, [
            'g_is_super_admin' => do_lang_tempcode('SUPER_ADMIN'),
            'g_order' => do_lang_tempcode('ORDER'),
        ]);
        if ($current_ordering == 'g_promotion_threshold ASC,r.id ASC') {
            list($sortable, $sort_order) = ['g_promotion_threshold ASC,r.id', 'ASC'];
        } elseif (($current_ordering == 'g_promotion_threshold DESC,r.id DESC') || ($current_ordering == 'g_promotion_threshold ASC,r.id DESC')) {
            list($sortable, $sort_order) = ['g_promotion_threshold DESC,r.id', 'DESC'];
        } else {
            if (strpos($current_ordering, ' ') === false) {
                warn_exit(do_lang_tempcode('INTERNAL_ERROR'));
            }
            list($sortable, $sort_order) = explode(' ', $current_ordering, 2);
            if (((cms_strtoupper_ascii($sort_order) != 'ASC') && (cms_strtoupper_ascii($sort_order) != 'DESC')) || (!array_key_exists($sortable, $sortables))) {
                log_hack_attack_and_exit('ORDERBY_HACK');
            }
        }

        $_header_row = [
            do_lang_tempcode('NAME'),
            do_lang_tempcode('IS_PRESENTED_AT_INSTALL'),
            do_lang_tempcode('DEFAULT_GROUP'),
            //do_lang_tempcode('IS_PRIVATE_CLUB'), Save space
            //do_lang_tempcode('GROUP_LEADER'),
            do_lang_tempcode('OPEN_MEMBERSHIP'),
        ];
        if (addon_installed('points')) {
            $_header_row = array_merge($_header_row, [
                do_lang_tempcode('PROMOTION_TARGET'),
            ]);
        }
        $_header_row = array_merge($_header_row, [
            do_lang_tempcode('SUPER_ADMIN'),
            do_lang_tempcode('ORDER'),
            do_lang_tempcode('ACTIONS'),
        ]);
        $header_row = results_header_row($_header_row, $sortables, 'sort', $sortable . ' ' . $sort_order);

        $result_entries = new Tempcode();

        $group_count = $GLOBALS['FORUM_DB']->query_select_value('f_groups', 'COUNT(*)', ['g_is_private_club' => 0]);

        list($rows, $max_rows) = $this->get_entry_rows(false, $current_ordering, ['g_is_private_club' => 0]);
        $changed = false;
        foreach ($rows as $row) {
            $new_order = post_param_integer('order_' . strval($row['id']), null);
            if ($new_order !== null) { // Ah, it's been set, better save that
                $GLOBALS['FORUM_DB']->query_update('f_groups', ['g_order' => $new_order], ['id' => $row['id']], '', 1);
                $changed = true;
            }
        }
        if ($changed) {
            list($rows, $max_rows) = $this->get_entry_rows(true, $current_ordering, ($group_count > 300) ? ['g_is_private_club' => 0] : []);
        }
        foreach ($rows as $row) {
            $edit_url = build_url($url_map + ['id' => $row['id']], '_SELF');

            $fr = [
                protect_from_escaping(cns_get_group_link($row['id'])),
                ($row['g_is_presented_at_install'] == 1) ? do_lang_tempcode('YES') : do_lang_tempcode('NO'),
                ($row['g_is_default'] == 1) ? do_lang_tempcode('YES') : do_lang_tempcode('NO'),
                //($row['g_is_private_club'] == 1) ? do_lang_tempcode('YES') : do_lang_tempcode('NO'),
                //($row['g_group_leader'] === null) ? do_lang_tempcode('NA_EM') : make_string_tempcode($GLOBALS['FORUM_DRIVER']->get_username($row['g_group_leader'])),
                ($row['g_open_membership'] == 1) ? do_lang_tempcode('YES') : do_lang_tempcode('NO'),
            ];
            if (addon_installed('points')) {
                $fr = array_merge($fr, [
                    ($row['g_promotion_target'] === null) ? do_lang_tempcode('NA_EM') : (make_string_tempcode(cns_get_group_name($row['g_promotion_target']) . ' (' . strval($row['g_promotion_threshold']) . ')')),
                ]);
            }
            $fr = array_merge($fr, [
                ($row['g_is_super_admin'] == 1) ? do_lang_tempcode('YES') : do_lang_tempcode('NO'),
            ]);

            $orderlist = new Tempcode();
            $selected_one = false;
            $order = $row['g_order'];
            for ($i = 0; $i < max(count($rows), $order, 200); $i++) {
                $selected = ($i === $order);
                if ($selected) {
                    $selected_one = true;
                }
                $orderlist->attach(form_input_list_entry(strval($i), $selected, strval($i + 1)));
            }
            if (!$selected_one) {
                $orderlist->attach(form_input_list_entry(strval($order), true, ($order == ORDER_AUTOMATED_CRITERIA) ? do_lang('NA') : strval($order + 1)));
            }
            $ordererx = protect_from_escaping(do_template('COLUMNED_TABLE_ROW_CELL_SELECT', [
                '_GUID' => '034e9c56038caa68874dece1174700e3',
                'LABEL' => do_lang_tempcode('SORT'),
                'NAME' => 'order_' . strval($row['id']),
                'LIST' => $orderlist,
            ]));

            $fr[] = $ordererx;

            $fr[] = protect_from_escaping(hyperlink($edit_url, do_lang_tempcode('EDIT'), false, false, do_lang('EDIT') . ' #' . strval($row['id'])));

            $result_entries->attach(results_entry($fr, true));
        }

        $search_url = null; //build_url(['page' => 'search', 'id' => 'cns_clubs'], get_module_zone('search'));
        $archive_url = build_url(['page' => 'groups'], get_module_zone('groups'));

        return [results_table(do_lang($this->menu_label), get_param_integer('start', 0), 'start', either_param_integer('max', 20), 'max', $max_rows, $header_row, $result_entries, $sortables, $sortable, $sort_order, 'sort', null, [], null, 8, 'gdfg43tfdgdfgdrfgd', true), true, $search_url, $archive_url];
    }

    /**
     * Standard crud_module list function.
     *
     * @return Tempcode The selection list
     */
    public function create_selection_list_entries() : object
    {
        $fields = new Tempcode();
        $order = (get_param_integer('keep_id_order', 0) == 0) ? 'g_promotion_threshold,id' : 'id';
        $group_count = $GLOBALS['FORUM_DB']->query_select_value('f_groups', 'COUNT(*)');
        $rows = $GLOBALS['FORUM_DB']->query_select('f_groups', ['id', 'g_order', 'g_name', 'g_promotion_target'], ($group_count > 300) ? ['g_is_private_club' => 0] : [], 'ORDER BY ' . $order);
        require_code('cns_groups2');
        foreach ($rows as $row) {
            $num_members = cns_get_group_members_raw_count($row['id'], true, false, true, true);

            if ($row['g_promotion_target'] === null) {
                $text = do_lang_tempcode('EXTENDED_GROUP_TITLE_NORMAL', escape_html(get_translated_text($row['g_name'], $GLOBALS['FORUM_DB'])), escape_html(strval($row['id'])), [escape_html(integer_format($row['g_order'] + 1)), escape_html(integer_format($num_members, 0))]);
            } else {
                $text = do_lang_tempcode('EXTENDED_GROUP_TITLE_RANK', escape_html(get_translated_text($row['g_name'], $GLOBALS['FORUM_DB'])), escape_html(strval($row['id'])), [strval($row['g_promotion_target']), escape_html(integer_format($row['g_order'] + 1)), escape_html(integer_format($num_members, 0))]);
            }
            $fields->attach(form_input_list_entry(strval($row['id']), false, $text));
        }

        return $fields;
    }

    /**
     * Standard crud_module delete possibility checker.
     *
     * @param  ID_TEXT $id The entry being potentially deleted
     * @return boolean Whether it may be deleted
     */
    public function may_delete_this(string $id) : bool
    {
        return !in_array(intval($id), get_all_preserved_groups(true));
    }

    /**
     * Standard crud_module edit form filler.
     *
     * @param  ID_TEXT $id The entry being edited
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function fill_in_edit_form(string $id)
    {
        $rows = $GLOBALS['FORUM_DB']->query_select('f_groups', ['*'], ['id' => intval($id)], '', 1);
        if (!array_key_exists(0, $rows)) {
            warn_exit(do_lang_tempcode('MISSING_RESOURCE', 'group'));
        }
        $myrow = $rows[0];

        if ($myrow['g_group_leader'] === null) {
            $username = '';
        } else {
            $username = $GLOBALS['FORUM_DRIVER']->get_username($myrow['g_group_leader'], false, USERNAME_DEFAULT_BLANK);
        }

        list($fields, $hidden) = $this->get_form_fields(intval($id), get_translated_text($myrow['g_name'], $GLOBALS['FORUM_DB']), $myrow['g_is_default'], $myrow['g_is_super_admin'], $myrow['g_is_super_moderator'], $username, get_translated_text($myrow['g_title'], $GLOBALS['FORUM_DB']), $myrow['g_rank_image'], $myrow['g_promotion_target'], $myrow['g_promotion_threshold'], $myrow['g_promotion_approval'], $myrow['g_flood_control_submit_secs'], $myrow['g_flood_control_access_secs'], $myrow['g_gift_points_base'], $myrow['g_gift_points_per_day'], $myrow['g_max_daily_upload_mb'], $myrow['g_max_attachments_per_post'], $myrow['g_max_avatar_width'], $myrow['g_max_avatar_height'], $myrow['g_max_post_length_comcode'], $myrow['g_max_sig_length_comcode'], $myrow['g_enquire_on_new_ips'], $myrow['g_is_presented_at_install'], $myrow['g_hidden'], $myrow['g_order'], $myrow['g_rank_image_pri_only'], $myrow['g_open_membership'], $myrow['g_is_private_club']);

        $default_group = get_first_default_group();

        $groups = new Tempcode();
        $group_count = $GLOBALS['FORUM_DB']->query_select_value('f_groups', 'COUNT(*)');
        if (($myrow['g_is_private_club'] == 1) && ($group_count > 300)) {
            $delete_fields = form_input_integer(do_lang_tempcode('NEW_USERGROUP'), do_lang_tempcode('DESCRIPTION_NEW_USERGROUP'), 'new_usergroup', null, false);
        } else {
            $rows = $GLOBALS['FORUM_DB']->query_select('f_groups', ['id', 'g_name'], ($group_count > 300) ? ['g_is_private_club' => 0] : []);
            foreach ($rows as $row) {
                if (($row['id'] != db_get_first_id()) && ($row['id'] != intval($id))) {
                    $groups->attach(form_input_list_entry(strval($row['id']), $row['id'] == ($default_group), get_translated_text($row['g_name'], $GLOBALS['FORUM_DB'])));
                }
            }
            $delete_fields = form_input_list(do_lang_tempcode('NEW_USERGROUP'), do_lang_tempcode('DESCRIPTION_NEW_USERGROUP'), 'new_usergroup', $groups);
        }

        $text = $this->edit_text;
        if (addon_installed('ecommerce')) {
            $usergroup_subs = $GLOBALS['FORUM_DB']->query_select('f_usergroup_subs', ['id', 's_title'], ['s_group_id' => intval($id)]);
            if (!empty($usergroup_subs)) {
                $subs = new Tempcode();
                foreach ($usergroup_subs as $i => $sub) {
                    if ($i != 0) {
                        $subs->attach(do_lang_tempcode('LIST_SEP'));
                    }
                    $subs->attach(hyperlink(build_url(['page' => 'admin_ecommerce', 'type' => '_edit', 'id' => $sub['id']], get_module_zone('admin_ecommerce')), get_translated_text($sub['s_title'], $GLOBALS['FORUM_DB']), false, true));
                }
                require_lang('ecommerce');
                $text->attach(paragraph(do_lang_tempcode('HAS_THESE_SUBS', $subs)));
            }
        }

        return [$fields, $hidden, $delete_fields, $text];
    }

    /**
     * Handle the "copy members from" feature.
     *
     * @param  GROUP $g The usergroup to copy members from
     */
    public function copy_members_into(int $g)
    {
        if (!array_key_exists('copy_members_into', $_POST)) {
            return;
        }
        $start = 0;
        do {
            $old_limit = cms_set_time_limit(20);

            send_http_output_ping();

            $members = $GLOBALS['FORUM_DRIVER']->member_group_query(array_map('intval', $_POST['copy_members_into']), 300, $start);
            foreach (array_keys($members) as $member_id) {
                cns_add_member_to_secondary_group($member_id, $g, 1);
                cns_update_group_approvals($member_id, get_member());
            }

            $start += 300;

            cms_set_time_limit($old_limit);
        } while (array_key_exists(0, $members));
    }

    /**
     * Read in data posted by an add/edit form.
     *
     * @return array A triplet of integers: (group leader, promotion target, promotion threshold)
     */
    public function read_in_data() : array
    {
        $_group_leader = post_param_string('group_leader', '');
        if ($_group_leader != '') {
            $group_leader = $GLOBALS['FORUM_DRIVER']->get_member_from_username($_group_leader);
            if (($group_leader === null) || (is_guest($group_leader))) {
                warn_exit(do_lang_tempcode('_MEMBER_NO_EXIST', escape_html($_group_leader)), false, false, 404);
            }
        } else {
            $group_leader = null;
        }
        if (fractional_edit()) {
            $group_leader = INTEGER_MAGIC_NULL;
        }

        $promotion_target = post_param_integer('promotion_target', fractional_edit() ? INTEGER_MAGIC_NULL : null);
        $promotion_threshold = post_param_integer('promotion_threshold', fractional_edit() ? INTEGER_MAGIC_NULL : null);
        $promotion_approval = post_param_integer('promotion_approval', fractional_edit() ? INTEGER_MAGIC_NULL : 0);

        return [$group_leader, $promotion_target, $promotion_threshold, $promotion_approval];
    }

    /**
     * Standard crud_module add actualiser.
     *
     * @return array A pair: The entry added, description about usage
     */
    public function add_actualisation() : array
    {
        require_code('themes2');

        list($group_leader, $promotion_target, $promotion_threshold, $promotion_approval) = $this->read_in_data();
        $rank_img = post_param_theme_img_code('cns_rank_images', false, 'file', 'theme_img_code', $GLOBALS['FORUM_DB']);

        $metadata = actual_metadata_get_fields('group', null, ['submitter']);

        $id = cns_make_group(post_param_string('usergroup_name'), post_param_integer('is_default', 0), post_param_integer('is_super_admin', 0), post_param_integer('is_super_moderator', 0), post_param_string('title', ''), $rank_img, $promotion_target, $promotion_threshold, $promotion_approval, $group_leader, post_param_integer('flood_control_submit_secs'), post_param_integer('flood_control_access_secs'), post_param_integer('max_daily_upload_mb'), post_param_integer('max_attachments_per_post'), post_param_integer('max_avatar_width', 100), post_param_integer('max_avatar_height', 100), post_param_integer('max_post_length_comcode'), post_param_integer('max_sig_length_comcode', 10000), post_param_integer('gift_points_base', 0), post_param_integer('gift_points_per_day', 0), post_param_integer('enquire_on_new_ips', 0), post_param_integer('is_presented_at_install', 0), post_param_integer('hidden', 0), post_param_order_field(), post_param_integer('rank_image_pri_only', 0), post_param_integer('open_membership', 0), post_param_integer('is_private_club', 0));

        set_url_moniker('group', strval($id));

        $this->copy_members_into($id);

        $absorb = post_param_integer('absorb', null);
        if ($absorb !== null) {
            cns_group_absorb_privileges_of($id, $absorb);
        }

        if (post_param_integer('is_private_club', 0) == 1) {
            $GLOBALS['SITE_DB']->query_delete('group_privileges', ['group_id' => $id]);
            $GLOBALS['FORUM_DB']->query_delete('group_privileges', ['group_id' => $id]);
            $GLOBALS['SITE_DB']->query_delete('group_zone_access', ['group_id' => $id]);
            $GLOBALS['SITE_DB']->query_delete('group_category_access', ['group_id' => $id]);
            $GLOBALS['FORUM_DB']->query_delete('group_category_access', ['group_id' => $id]);
            $GLOBALS['SITE_DB']->query_delete('group_page_access', ['group_id' => $id]);
        }

        if ($group_leader !== null) {
            cns_add_member_to_secondary_group($group_leader, $id);
            cns_update_group_approvals($group_leader, get_member());
        }

        if (addon_installed('ecommerce')) {
            require_lang('ecommerce');
            $this->extra_donext_whatever = [
                ['menu/adminzone/audit/ecommerce/ecommerce', ['admin_ecommerce', ['type' => 'add', 'group_id' => $id], '_SELF'], do_lang_tempcode('ADD_USERGROUP_SUBSCRIPTION')],
            ];
            $this->extra_donext_whatever_title = do_lang_tempcode('USERGROUP_SUBSCRIPTIONS');
        }

        if (addon_installed('content_reviews')) {
            content_review_set('group', strval($id));
        }

        return [strval($id), null];
    }

    /**
     * Standard crud_module edit actualiser.
     *
     * @param  ID_TEXT $id The entry being edited
     * @return ?Tempcode Description about usage (null: none)
     */
    public function edit_actualisation(string $id) : ?object
    {
        require_code('themes2');

        list($group_leader, $promotion_target, $promotion_threshold, $promotion_approval) = $this->read_in_data();
        if (($group_leader !== null) && ($group_leader != INTEGER_MAGIC_NULL) && (post_param_integer('confirm', 0) == 0) && (!in_array(intval($id), $GLOBALS['FORUM_DRIVER']->get_members_groups($group_leader)))) {
            require_code('templates_confirm_screen');
            return confirm_screen(get_screen_title('EDIT_GROUP'), paragraph(do_lang_tempcode('MAKE_MEMBER_GROUP_LEADER', escape_html(post_param_string('group_leader')))), '__edit', '_edit', ['confirm' => 1]);
        }

        $was_club = ($GLOBALS['FORUM_DB']->query_select_value('f_groups', 'g_is_private_club', ['id' => intval($id)]) == 1);

        $rank_img = fractional_edit() ? STRING_MAGIC_NULL : post_param_theme_img_code('cns_rank_images', false, 'file', 'theme_img_code', $GLOBALS['FORUM_DB']);

        $metadata = actual_metadata_get_fields('group', $id, ['submitter']);

        $is_super_admin = post_param_integer('is_super_admin', fractional_edit() ? INTEGER_MAGIC_NULL : 0);
        if (($is_super_admin == 0) && ($GLOBALS['FORUM_DB']->query_select_value('f_groups', 'g_is_super_admin', ['id' => intval($id)]) == 1) && ($GLOBALS['FORUM_DB']->query_select_value('f_groups', 'COUNT(*)', ['g_is_super_admin' => 1]) == 1)) {
            warn_exit(do_lang_tempcode('NO_SUICIDAL_SUPER_ADMIN_REMOVAL'));
        }

        cns_edit_group(
            intval($id),
            post_param_string('usergroup_name'),
            post_param_integer('is_default', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            $is_super_admin,
            post_param_integer('is_super_moderator', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_string('title', STRING_MAGIC_NULL),
            $rank_img,
            $promotion_target,
            $promotion_threshold,
            $promotion_approval,
            $group_leader,
            post_param_integer('flood_control_submit_secs', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('flood_control_access_secs', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('max_daily_upload_mb', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('max_attachments_per_post', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('max_avatar_width', fractional_edit() ? INTEGER_MAGIC_NULL : 100),
            post_param_integer('max_avatar_height', fractional_edit() ? INTEGER_MAGIC_NULL : 100),
            post_param_integer('max_post_length_comcode', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('max_sig_length_comcode', fractional_edit() ? INTEGER_MAGIC_NULL : 10000),
            post_param_integer('gift_points_base', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('gift_points_per_day', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('enquire_on_new_ips', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('is_presented_at_install', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('hidden', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            fractional_edit() ? INTEGER_MAGIC_NULL : post_param_order_field(),
            post_param_integer('rank_image_pri_only', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('open_membership', fractional_edit() ? INTEGER_MAGIC_NULL : 0),
            post_param_integer('is_private_club', fractional_edit() ? INTEGER_MAGIC_NULL : 0)
        );

        if (addon_installed('ecommerce')) {
            require_lang('ecommerce');
            $this->extra_donext_whatever = [
                ['menu/adminzone/audit/ecommerce/ecommerce', ['admin_ecommerce', ['type' => 'add', 'group_id' => $id], '_SELF'], do_lang_tempcode('ADD_USERGROUP_SUBSCRIPTION')],
            ];
            $this->extra_donext_whatever_title = do_lang_tempcode('USERGROUP_SUBSCRIPTIONS');
        }

        if (($group_leader !== null) && ($group_leader != INTEGER_MAGIC_NULL) && (!in_array(intval($id), $GLOBALS['FORUM_DRIVER']->get_members_groups($group_leader)))) {
            cns_add_member_to_secondary_group($group_leader, intval($id));
            cns_update_group_approvals($group_leader, get_member());
        }

        if (!fractional_edit()) {
            $absorb = post_param_integer('absorb', null);
            if ($absorb !== null) {
                cns_group_absorb_privileges_of(intval($id), $absorb);
            }

            if ((post_param_integer('is_private_club', 0) == 1) && (!$was_club)) {
                $GLOBALS['SITE_DB']->query_delete('group_privileges', ['group_id' => intval($id)]);
                $GLOBALS['SITE_DB']->query_delete('group_zone_access', ['group_id' => intval($id)]);
                $GLOBALS['SITE_DB']->query_delete('group_category_access', ['group_id' => intval($id)]);
                $GLOBALS['SITE_DB']->query_delete('group_page_access', ['group_id' => intval($id)]);
            }
        }

        if (addon_installed('content_reviews')) {
            content_review_set('group', $id);
        }

        return null;
    }

    /**
     * Standard crud_module delete actualiser.
     *
     * @param  ID_TEXT $id The entry being deleted
     */
    public function delete_actualisation(string $id)
    {
        cns_delete_group(intval($id), post_param_integer('new_usergroup', null));
    }
}
