<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    news
 */

require_code('crud_module');

/**
 * Module page class.
 */
class Module_cms_blogs extends Standard_crud_module
{
    protected $lang_type = 'NEWS_BLOG';
    protected $select_name = 'TITLE';
    protected $code_require = 'news';
    protected $permissions_require = 'mid';
    protected $permissions_module_require = 'news';
    protected $permissions_cat_name = 'main_news_category';
    protected $user_facing = true;
    protected $seo_type = 'news';
    protected $content_type = 'news';
    protected $possibly_some_kind_of_upload = true;
    protected $upload = 'image';
    protected $menu_label = 'BLOGS';
    protected $table = 'news';
    protected $orderer = 'title';
    protected $privilege_page_name = 'cms_news';

    protected $donext_type = null;

    /**
     * Find entry-points available within this module.
     *
     * @param  boolean $check_perms Whether to check permissions
     * @param  ?MEMBER $member_id The member to check permissions as (null: current user)
     * @param  boolean $support_crosslinks Whether to allow cross links to other modules (identifiable via a full-page-link rather than a screen-name)
     * @param  boolean $be_deferential Whether to avoid any entry-point (or even return null to disable the page in the Sitemap) if we know another module, or page_group, is going to link to that entry-point. Note that "!" and "browse" entry points are automatically merged with container page nodes (likely called by page-groupings) as appropriate.
     * @return ?array A map of entry points (screen-name=>language-code/string or screen-name=>[language-code/string, icon-theme-image]) (null: disabled)
     */
    public function get_entry_points(bool $check_perms = true, ?int $member_id = null, bool $support_crosslinks = true, bool $be_deferential = false) : ?array
    {
        if (!addon_installed('news')) {
            return null;
        }

        if ($member_id === null) {
            $member_id = get_member();
        }

        if (get_option('separate_blogs') == '0') {
            return null;
        }

        if (!has_privilege($member_id, 'have_personal_category', 'cms_news')) {
            return null;
        }

        $ret = [
            'browse' => ['MANAGE_BLOGS', 'menu/cms/blog'],
        ];
        $ret += parent::get_entry_points();
        $ret += [
            'import_wordpress' => ['IMPORT_WORDPRESS', 'admin/import'],
        ];
        return $ret;
    }

    public $title;

    /**
     * Module pre-run function. Allows us to know metadata for <head> before we start streaming output.
     *
     * @param  boolean $top_level Whether this is running at the top level, prior to having sub-objects called
     * @param  ?ID_TEXT $type The screen type to consider for metadata purposes (null: read from environment)
     * @return ?Tempcode Tempcode indicating some kind of exceptional output (null: none)
     */
    public function pre_run(bool $top_level = true, ?string $type = null) : ?object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('news', $error_msg)) {
            return $error_msg;
        }

        if (!addon_installed('news_shared')) {
            warn_exit(do_lang_tempcode('MISSING_ADDON', escape_html('news_shared')));
        }

        $type = get_param_string('type', 'browse');

        require_lang('news');

        inform_non_canonical_parameter('validated');
        inform_non_canonical_parameter('cat');

        set_helper_panel_tutorial('tut_news');

        if ($type == 'add' || $type == '_edit') {
            set_helper_panel_text(comcode_lang_string('DOC_NEWS'));
        }

        if ($type == 'import_wordpress' || $type == '_import_wordpress') {
            $this->title = get_screen_title('IMPORT_WORDPRESS');
        }

        if ($type == '_import_wordpress') {
            breadcrumb_set_parents([['_SELF:_SELF:import_wordpress', do_lang_tempcode('IMPORT_WORDPRESS')]]);
            breadcrumb_set_self(do_lang_tempcode('DONE'));
        }

        return parent::pre_run($top_level);
    }

    /**
     * Standard crud_module run_start.
     *
     * @param  ID_TEXT $type The type of module execution
     * @return Tempcode The output of the run
     */
    public function run_start(string $type) : object
    {
        $this->posting_form_title = do_lang_tempcode('BLOG_NEWS_ARTICLE');

        if (is_guest()) {
            access_denied('NOT_AS_GUEST');
        }

        require_code('news2');
        require_css('news');

        // Decide what to do
        if ($type == 'browse') {
            return $this->browse();
        }
        if ($type == 'import_wordpress') {
            return $this->import_wordpress();
        }
        if ($type == '_import_wordpress') {
            return $this->_import_wordpress();
        }

        return new Tempcode();
    }

    /**
     * The do-next manager for before content management.
     *
     * @return Tempcode The UI
     */
    public function browse() : object
    {
        require_code('templates_donext');
        return do_next_manager(
            get_screen_title('MANAGE_BLOGS'),
            comcode_lang_string('DOC_BLOGS'),
            [
                has_privilege(get_member(), 'submit_midrange_content', 'cms_news') ? ['admin/add', ['_SELF', ['type' => 'add'], '_SELF'], do_lang('ADD_NEWS_BLOG')] : null,
                has_privilege(get_member(), 'edit_own_midrange_content', 'cms_news') ? ['admin/edit', ['_SELF', ['type' => 'edit'], '_SELF'], do_lang('EDIT_NEWS_BLOG')] : null,
                has_privilege(get_member(), 'mass_import', 'cms_news') ? ['admin/import', ['_SELF', ['type' => 'import_wordpress'], '_SELF'], do_lang('IMPORT_WORDPRESS')] : null,
            ],
            do_lang('MANAGE_BLOGS')
        );
    }

    /**
     * Standard crud_module table function.
     *
     * @param  array $url_map Details to go to build_url for link to the next screen
     * @return ?array A quartet: The choose table, Whether re-ordering is supported from this screen, Search URL, Archive URL (null: nothing to select)
     */
    public function create_selection_list_choose_table(array $url_map) : ?array
    {
        require_code('templates_results_table');

        $current_ordering = get_param_string('sort', 'date_and_time DESC', INPUT_FILTER_GET_COMPLEX);
        if (strpos($current_ordering, ' ') === false) {
            warn_exit(do_lang_tempcode('INTERNAL_ERROR'));
        }
        list($sortable, $sort_order) = explode(' ', $current_ordering, 2);
        $sortables = [
            'title' => do_lang_tempcode('TITLE'),
            'date_and_time' => do_lang_tempcode('ADDED'),
            'news_views' => do_lang_tempcode('COUNT_VIEWS'),
        ];
        if (addon_installed('validation')) {
            $sortables['validated'] = do_lang_tempcode('VALIDATED');
        }
        if (((cms_strtoupper_ascii($sort_order) != 'ASC') && (cms_strtoupper_ascii($sort_order) != 'DESC')) || (!array_key_exists($sortable, $sortables))) {
            log_hack_attack_and_exit('ORDERBY_HACK');
        }

        $fh = [];
        $fh[] = do_lang_tempcode('TITLE');
        $fh[] = do_lang_tempcode('ADDED');
        $fh[] = do_lang_tempcode('COUNT_VIEWS');
        if (addon_installed('validation')) {
            $fh[] = protect_from_escaping(do_template('COMCODE_ABBR', ['_GUID' => '204d1050402b48e5c2c9539763a3fe50', 'TITLE' => do_lang_tempcode('VALIDATED'), 'CONTENT' => do_lang_tempcode('VALIDATED_SHORT')]));
        }
        $fh[] = do_lang_tempcode('ACTIONS');
        $header_row = results_header_row($fh, $sortables, 'sort', $sortable . ' ' . $sort_order);

        $result_entries = new Tempcode();

        $only_owned = has_privilege(get_member(), 'edit_midrange_content', 'cms_news') ? null : get_member();
        list($rows, $max_rows) = $this->get_entry_rows(false, $current_ordering, ($only_owned === null) ? [] : ['submitter' => $only_owned], false, ' JOIN ' . get_table_prefix() . 'news_categories c ON c.id=r.news_category AND nc_owner IS NOT NULL');
        if (empty($rows)) {
            return null;
        }
        foreach ($rows as $row) {
            $edit_url = build_url($url_map + ['id' => $row['id']], '_SELF');

            $fr = [];
            $fr[] = protect_from_escaping(hyperlink(build_url(['page' => 'news', 'type' => 'view', 'id' => $row['id']], get_module_zone('news')), get_translated_text($row['title']), false, true));
            $fr[] = get_timezoned_date_time($row['date_and_time']);
            $fr[] = integer_format($row['news_views']);
            if (addon_installed('validation')) {
                $fr[] = ($row['validated'] == 1) ? do_lang_tempcode('YES') : do_lang_tempcode('NO');
            }
            $fr[] = protect_from_escaping(hyperlink($edit_url, do_lang_tempcode('EDIT'), false, true, do_lang('EDIT') . ' #' . strval($row['id'])));

            $result_entries->attach(results_entry($fr, true));
        }

        $search_url = build_url(['page' => 'search', 'id' => 'news'], get_module_zone('search'));
        $archive_url = build_url(['page' => 'news'], get_module_zone('news'));

        return [results_table(do_lang($this->menu_label), get_param_integer('start', 0), 'start', either_param_integer('max', 20), 'max', $max_rows, $header_row, $result_entries, $sortables, $sortable, $sort_order), false, $search_url, $archive_url];
    }

    /**
     * Standard crud_module list function.
     *
     * @return Tempcode The selection list
     */
    public function create_selection_list_entries() : object
    {
        $only_owned = has_privilege(get_member(), 'edit_midrange_content', 'cms_news') ? null : get_member();
        return create_selection_list_news(null, $only_owned, false, true);
    }

    /**
     * Get Tempcode for an adding form.
     *
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function get_form_fields_for_add()
    {
        return $this->get_form_fields();
    }

    /**
     * Get Tempcode for a news adding/editing form.
     *
     * @param  ?AUTO_LINK $id The news ID (null: new)
     * @param  ?AUTO_LINK $main_news_category The primary category for the news (null: personal)
     * @param  ?array $news_category A list of categories the news is in (null: not known)
     * @param  SHORT_TEXT $title The news title
     * @param  LONG_TEXT $news The news summary
     * @param  SHORT_TEXT $author The name of the author
     * @param  BINARY $validated Whether the news is validated
     * @param  ?BINARY $allow_rating Whether rating is allowed (null: decide statistically, based on existing choices)
     * @param  ?SHORT_INTEGER $allow_comments Whether comments are allowed (0=no, 1=yes, 2=review style) (null: decide statistically, based on existing choices)
     * @param  ?BINARY $allow_trackbacks Whether trackbacks are allowed (null: decide statistically, based on existing choices)
     * @param  BINARY $send_trackbacks Whether to show the "send trackback" field
     * @param  LONG_TEXT $notes Notes for the video
     * @param  URLPATH $image URL to the image for the news entry (blank: use cat image)
     * @param  ?array $scheduled Scheduled go-live time (null: N/A)
     * @return array A tuple: The input fields, Hidden fields, ...
     */
    public function get_form_fields(?int $id = null, ?int $main_news_category = null, ?array $news_category = null, string $title = '', string $news = '', string $author = '', int $validated = 1, ?int $allow_rating = null, ?int $allow_comments = null, ?int $allow_trackbacks = null, int $send_trackbacks = 1, string $notes = '', string $image = '', ?array $scheduled = null) : array
    {
        list($allow_rating, $allow_comments, $allow_trackbacks) = $this->choose_feedback_fields_statistically($allow_rating, $allow_comments, $allow_trackbacks);

        if ($title == '') {
            $title = get_param_string('title', $title, INPUT_FILTER_GET_COMPLEX);
            $author = get_param_string('author', $author);
            $notes = get_param_string('notes', $notes, INPUT_FILTER_GET_COMPLEX);

            if ($main_news_category === null) {
                $param_cat = get_param_string('cat', '');
                if ($param_cat == '') {
                    $main_news_category = null;
                    $news_category = [];
                } elseif (strpos($param_cat, ',') === false) {
                    $main_news_category = intval($param_cat);
                    $news_category = [];
                } else {
                    require_code('selectcode');
                    $_param_cat = explode(',', $param_cat);
                    $_main_news_category = array_shift($_param_cat);
                    $param_cat = implode(',', $_param_cat);
                    $main_news_category = ($_main_news_category == '') ? null : intval($_main_news_category);
                    $news_category = selectcode_to_idlist_using_db($param_cat, 'id', 'news_categories', 'news_categories', null, 'id', 'id');
                }

                $author = $GLOBALS['FORUM_DRIVER']->get_username(get_member());
            }
        }

        $cats1 = create_selection_list_news_categories($main_news_category, false, true, false, true);
        if ($cats1->is_empty()) {
            warn_exit(do_lang_tempcode('NO_CATEGORIES', 'news_category'));
        }
        $cats2 = create_selection_list_news_categories((($news_category === null) || (empty($news_category))) ? [get_param_integer('cat', null)] : $news_category, false, true, true, false);

        $fields = new Tempcode();
        $fields2 = new Tempcode();
        $hidden = new Tempcode();
        $fields->attach(form_input_line_comcode(do_lang_tempcode('TITLE'), do_lang_tempcode('DESCRIPTION_TITLE'), 'title', $title, true));

        $_validated = get_param_integer('validated', 0);
        if ($validated == 0) {
            if (($_validated == 1) && (addon_installed('validation'))) {
                $validated = 1;
                attach_message(do_lang_tempcode('WILL_BE_VALIDATED_WHEN_SAVING'));
            }
        } elseif (($validated == 1) && ($_validated == 1) && ($id !== null)) {
            $action_log = build_url(['page' => 'admin_actionlog', 'type' => 'list', 'to_type' => 'VALIDATE_NEWS', 'param_a' => strval($id)]);
            attach_message(do_lang_tempcode('ALREADY_VALIDATED', escape_html($action_log->evaluate())), 'notice');
        }

        if (has_some_cat_privilege(get_member(), 'bypass_validation_' . $this->permissions_require . 'range_content', 'cms_news', $this->permissions_module_require)) {
            if (addon_installed('validation')) {
                $fields2->attach(form_input_tick(do_lang_tempcode('VALIDATED'), do_lang_tempcode($GLOBALS['FORUM_DRIVER']->is_super_admin(get_member()) ? 'DESCRIPTION_VALIDATED_SIMPLE' : 'DESCRIPTION_VALIDATED', 'news'), 'validated', $validated == 1));
            }
        }
        if ($cats1->is_empty()) {
            warn_exit(do_lang_tempcode('NO_CATEGORIES', 'news_category'));
        }
        if (addon_installed('authors')) {
            $hidden->attach(form_input_hidden('author', $author));
        }
        $fields2->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => 'cff27a28d4d01f8e0255ee645ea210b3', 'SECTION_HIDDEN' => $image == '' && ($title == ''/*=new entry and selected news cats was from URL*/ || (empty($news_category))), 'TITLE' => do_lang_tempcode('ADVANCED')]));
        $fields2->attach(form_input_text_comcode(do_lang_tempcode('BLOG_NEWS_SUMMARY'), do_lang_tempcode('DESCRIPTION_NEWS_SUMMARY'), 'news', $news, false));
        if (get_option('enable_secondary_news') == '1') {
            $fields2->attach(form_input_list(do_lang_tempcode('MAIN_CATEGORY'), do_lang_tempcode('DESCRIPTION_MAIN_CATEGORY'), 'main_news_category', $cats1));
        } else {
            $fields2->attach(form_input_hidden('main_news_category', ($main_news_category === null) ? 'personal' : strval($main_news_category)));
        }
        if (get_option('enable_secondary_news') == '1') {
            $fields2->attach(form_input_multi_list(do_lang_tempcode('SECONDARY_CATEGORIES'), do_lang_tempcode('DESCRIPTION_SECONDARY_CATEGORIES', 'news'), 'news_category', $cats2));
        }

        if ((addon_installed('calendar')) && (addon_installed('commandr')) && (has_privilege(get_member(), 'scheduled_publication_times'))) {
            $fields2->attach(form_input_date__cron(do_lang_tempcode('PUBLICATION_TIME'), do_lang_tempcode('DESCRIPTION_PUBLICATION_TIME'), 'schedule', false, ($scheduled === null), true, $scheduled, intval(date('Y')) - 1970 + 2, 1970));
        }

        require_code('feedback2');
        $fields2->attach(feedback_fields($this->content_type, $allow_rating == 1, $allow_comments == 1, $allow_trackbacks == 1, $send_trackbacks == 1, $notes, $allow_comments == 2));

        require_code('content2');
        $fields2->attach(seo_get_fields($this->seo_type, ($id === null) ? null : strval($id)));

        require_code('syndication');
        $fields2->attach(get_syndication_option_fields('news', $id !== null));

        if (addon_installed('content_privacy')) {
            require_code('content_privacy2');
            if ($id === null) {
                $fields2->attach(get_privacy_form_fields('news'));
            } else {
                $fields2->attach(get_privacy_form_fields('news', strval($id)));
            }
        }

        return [$fields, $hidden, null, null, null, null, make_string_tempcode($fields2->evaluate())];
    }

    /**
     * Standard crud_module submitter getter.
     *
     * @param  ID_TEXT $id The entry for which the submitter is sought
     * @return array The submitter, and the time of submission (null submission time implies no known submission time)
     */
    public function get_submitter(string $id) : array
    {
        $rows = $GLOBALS['SITE_DB']->query_select('news', ['submitter', 'date_and_time'], ['id' => intval($id)], '', 1);
        if (!array_key_exists(0, $rows)) {
            return [null, null];
        }
        return [$rows[0]['submitter'], $rows[0]['date_and_time']];
    }

    /**
     * Standard crud_module category getter.
     *
     * @param  ID_TEXT $id The entry for which the category is sought
     * @return string The category
     */
    public function get_cat(string $id) : string
    {
        $temp = $GLOBALS['SITE_DB']->query_select_value_if_there('news', 'news_category', ['id' => intval($id)]);
        if ($temp === null) {
            warn_exit(do_lang_tempcode('MISSING_RESOURCE', 'news'));
        }
        return strval($temp);
    }

    /**
     * Standard crud_module edit form filler.
     *
     * @param  ID_TEXT $_id The entry being edited
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function fill_in_edit_form(string $_id)
    {
        $id = intval($_id);

        require_lang('zones');

        $rows = $GLOBALS['SITE_DB']->query_select('news', ['*'], ['id' => intval($id)], '', 1);
        if (!array_key_exists(0, $rows)) {
            warn_exit(do_lang_tempcode('MISSING_RESOURCE', 'news'));
        }
        $myrow = $rows[0];

        $cat = $myrow['news_category'];

        $categories = [];
        $category_query = $GLOBALS['SITE_DB']->query_select('news_category_entries', ['news_entry_category'], ['news_entry' => $id]);

        foreach ($category_query as $value) {
            $categories[] = $value['news_entry_category'];
        }

        $scheduled = null;

        if (addon_installed('calendar')) {
            require_code('calendar2');
            $scheduled = get_schedule_code_event_time('publish_news', strval($id));
        } else {
            $scheduled = null;
        }

        $ret = $this->get_form_fields($id, $cat, $categories, get_translated_text($myrow['title']), get_translated_text($myrow['news']), $myrow['author'], $myrow['validated'], $myrow['allow_rating'], $myrow['allow_comments'], $myrow['allow_trackbacks'], 0, $myrow['notes'], $myrow['news_image'], $scheduled);
        $ret[2] = new Tempcode();
        $ret[3] = '';
        $ret[4] = false;
        $ret[5] = get_translated_text($myrow['news_article']);
        $ret[7] = get_translated_tempcode('news', $myrow, 'news_article');
        return $ret;
    }

    /**
     * Standard crud_module add actualiser.
     *
     * @return array A pair: The entry added, description about usage
     */
    public function add_actualisation() : array
    {
        $author = post_param_string('author', $GLOBALS['FORUM_DRIVER']->get_username(get_member()));
        $news = post_param_string('news');
        $title = post_param_string('title');
        $validated = post_param_integer('validated', 0);
        $news_article = post_param_string('post');
        if (post_param_string('main_news_category') != 'personal') {
            $main_news_category = post_param_integer('main_news_category');
        } else {
            $main_news_category = null;
        }

        $news_category = [];
        if (array_key_exists('news_category', $_POST)) {
            foreach ($_POST['news_category'] as $val) {
                $news_category[] = ($val == 'personal') ? null : intval($val);
            }
        }

        $allow_rating = post_param_integer('allow_rating', 0);
        $allow_comments = post_param_integer('allow_comments', 0);
        $allow_trackbacks = post_param_integer('allow_trackbacks', 0);
        $notes = post_param_string('notes', '');

        $schedule = post_param_date('schedule');
        if ((addon_installed('calendar')) && (addon_installed('commandr')) && (has_privilege(get_member(), 'scheduled_publication_times')) && ($schedule !== null) && ($schedule > time())) {
            $validated = 0;
        } else {
            $schedule = null;
        }

        if ($main_news_category !== null) {
            $owner = $GLOBALS['SITE_DB']->query_select_value('news_categories', 'nc_owner', ['id' => intval($main_news_category)]);
            if (($owner !== null) && ($owner != get_member())) {
                check_privilege('can_submit_to_others_categories', ['news', $main_news_category], null, 'cms_news');
            }
        }

        $metadata = actual_metadata_get_fields('news', null);

        $id = add_news($title, $news, $author, $validated, $allow_rating, $allow_comments, $allow_trackbacks, $notes, $news_article, $main_news_category, $news_category, $metadata['add_time'], $metadata['submitter'], $metadata['views'], null, null, '');

        require_code('feedback2');
        send_trackbacks(post_param_string('send_trackbacks', ''), $title, $news);

        set_url_moniker('news', strval($id));

        $main_news_category = $GLOBALS['SITE_DB']->query_select_value('news', 'news_category', ['id' => $id]);
        $this->donext_type = $main_news_category;

        if (addon_installed('content_privacy')) {
            require_code('content_privacy2');
            list($privacy_level, $additional_access) = read_privacy_fields();
            save_privacy_form_fields('news', strval($id), $privacy_level, $additional_access);
        }

        if (($validated == 1) || (!addon_installed('validation'))) {
            require_code('users2');
            if (has_actual_page_access(get_modal_user(), 'news')) { // NB: no category permission check, as syndication choice was explicit, and news categorisation is a bit more complex
                $privacy_ok = true;
                if (addon_installed('content_privacy')) {
                    require_code('content_privacy');
                    $privacy_ok = has_privacy_access('news', strval($id), $GLOBALS['FORUM_DRIVER']->get_guest_id());
                }
                if ($privacy_ok) {
                    require_code('syndication');
                    syndicate_content('news', strval($id), [['news:ACTIVITY_ADD_NEWS_BLOG', get_member(), $title]]);
                }
            }
        }

        if ($schedule !== null) {
            require_code('calendar');
            $parameters = [];
            $start_year = intval(date('Y', $schedule));
            $start_month = intval(date('m', $schedule));
            $start_day = intval(date('d', $schedule));
            $start_hour = intval(date('H', $schedule));
            $start_minute = intval(date('i', $schedule));
            require_code('calendar2');
            schedule_code('publish_news', strval($id), $parameters, do_lang('PUBLISH_NEWS', $title), $start_year, $start_month, $start_day, $start_hour, $start_minute);
        }

        return [strval($id), null];
    }

    /**
     * Standard crud_module edit actualiser.
     *
     * @param  ID_TEXT $_id The entry being edited
     * @return ?Tempcode Description about usage (null: none)
     */
    public function edit_actualisation(string $_id) : ?object
    {
        $id = intval($_id);

        $validated = post_param_integer('validated', fractional_edit() ? INTEGER_MAGIC_NULL : 0);

        $news_article = post_param_string('post', STRING_MAGIC_NULL);
        if (post_param_string('main_news_category') != 'personal') {
            $main_news_category = post_param_integer('main_news_category', INTEGER_MAGIC_NULL);
        } else {
            warn_exit(do_lang_tempcode('INTERNAL_ERROR'));
        }

        $news_category = [];
        if (array_key_exists('news_category', $_POST)) {
            foreach ($_POST['news_category'] as $val) {
                $news_category[] = intval($val);
            }
        }

        $allow_rating = post_param_integer('allow_rating', fractional_edit() ? INTEGER_MAGIC_NULL : 0);
        $allow_comments = post_param_integer('allow_comments', fractional_edit() ? INTEGER_MAGIC_NULL : 0);
        $allow_trackbacks = post_param_integer('allow_trackbacks', fractional_edit() ? INTEGER_MAGIC_NULL : 0);
        $notes = post_param_string('notes', STRING_MAGIC_NULL);

        $this->donext_type = $main_news_category;

        $url = STRING_MAGIC_NULL;

        $owner = $GLOBALS['SITE_DB']->query_select_value_if_there('news_categories', 'nc_owner', ['id' => $main_news_category]); // if_there in case somehow category setting corrupted
        if (($owner !== null) && ($owner != get_member())) {
            check_privilege('can_submit_to_others_categories', ['news', $main_news_category], null, 'cms_news');
        }

        $schedule = post_param_date('schedule');

        if ((addon_installed('calendar')) && (addon_installed('commandr')) && (has_privilege(get_member(), 'scheduled_publication_times'))) {
            require_code('calendar2');
            unschedule_code('publish_news', strval($id));

            if (($schedule !== null) && ($schedule > time())) {
                $validated = 0;

                $parameters = [];
                $start_year = intval(date('Y', $schedule));
                $start_month = intval(date('m', $schedule));
                $start_day = intval(date('d', $schedule));
                $start_hour = intval(date('H', $schedule));
                $start_minute = intval(date('i', $schedule));
                schedule_code('publish_news', strval($id), $parameters, do_lang('PUBLISH_NEWS', post_param_string('title')), $start_year, $start_month, $start_day, $start_hour, $start_minute);
            }
        }

        $title = post_param_string('title', STRING_MAGIC_NULL);

        if (addon_installed('content_privacy')) {
            require_code('content_privacy2');
            list($privacy_level, $additional_access) = read_privacy_fields();
            save_privacy_form_fields('news', strval($id), $privacy_level, $additional_access);
        }

        if (($validated == 1) || (!addon_installed('validation'))) {
            require_code('users2');
            if (has_actual_page_access(get_modal_user(), 'news')) {
                $privacy_ok = true;
                if (addon_installed('content_privacy')) {
                    require_code('content_privacy');
                    $privacy_ok = has_privacy_access('news', strval($id), $GLOBALS['FORUM_DRIVER']->get_guest_id());
                }
                if ($privacy_ok) {
                    $just_validated = (addon_installed('validation')) && ($GLOBALS['SITE_DB']->query_select_value('news', 'validated', ['id' => $id]) == 0);
                    $submitter = $GLOBALS['SITE_DB']->query_select_value('news', 'submitter', ['id' => $id]);

                    $activities = [];
                    if ($just_validated) {
                        if ($submitter != get_member()) {
                            $activities[] = ['news:ACTIVITY_VALIDATE_NEWS_BLOG', get_member(), $title];
                        }
                        $activities[] = ['news:ACTIVITY_ADD_NEWS_BLOG', $submitter, $title];
                    }

                    require_code('syndication');
                    syndicate_content('news', strval($id), $activities);
                }
            }
        }

        $metadata = actual_metadata_get_fields('news', strval($id));

        edit_news(intval($id), $title, post_param_string('news', STRING_MAGIC_NULL), post_param_string('author', STRING_MAGIC_NULL), $validated, $allow_rating, $allow_comments, $allow_trackbacks, $notes, $news_article, $main_news_category, $news_category, post_param_string('meta_keywords', STRING_MAGIC_NULL), post_param_string('meta_description', STRING_MAGIC_NULL), $url, $metadata['add_time'], $metadata['edit_time'], $metadata['views'], $metadata['submitter'], [], true);

        return null;
    }

    /**
     * Standard crud_module delete actualiser.
     *
     * @param  ID_TEXT $_id The entry being deleted
     */
    public function delete_actualisation(string $_id)
    {
        $id = intval($_id);

        delete_news($id);

        if (addon_installed('content_privacy')) {
            require_code('content_privacy2');
            delete_privacy_form_fields('news', strval($id));
        }

        require_code('syndication');
        unsyndicate_content('news', strval($id));
    }

    /**
     * The do-next manager for after news content management.
     *
     * @param  ?Tempcode $title The title (output of get_screen_title) (null: don't use full page)
     * @param  Tempcode $description Some description to show, saying what happened
     * @param  ?ID_TEXT $id The ID of whatever we are working with (null: deleted)
     * @return Tempcode The UI
     */
    public function do_next_manager(?object $title, object $description, ?string $id = null) : object
    {
        $cat = $this->donext_type;

        require_code('templates_donext');

        return do_next_manager(
            $this->title,
            $description,
            [],
            null,
            /* TYPED-ORDERED LIST OF 'LINKS' */
            ['_SELF', ['type' => 'add', 'cat' => $cat], '_SELF'], // Add one
            (($id === null) || (!has_privilege(get_member(), 'edit_own_midrange_content', 'cms_news', ['news', $cat]))) ? null : ['_SELF', ['type' => '_edit', 'id' => $id], '_SELF'], // Edit this
            has_privilege(get_member(), 'edit_own_midrange_content', 'cms_news') ? ['_SELF', ['type' => 'edit'], '_SELF'] : null, // Edit one
            ($id === null) ? null : ['news', ['type' => 'view', 'id' => $id, 'blog' => 1], get_module_zone('news')], // View this
            ['news', ['type' => 'browse', 'blog' => 1], get_module_zone('news')], // View archive
            has_privilege(get_member(), 'submit_cat_midrange_content', 'cms_news') ? ['cms_news', ['type' => 'add_category'], '_SELF'] : null, // Add one category
            has_privilege(get_member(), 'edit_own_cat_midrange_content', 'cms_news') ? ['cms_news', ['type' => 'edit_category'], '_SELF'] : null, // Edit one category
            ($cat === null) ? null : (has_privilege(get_member(), 'edit_own_cat_midrange_content', 'cms_news') ? ['cms_news', ['type' => '_edit_category', 'id' => $cat], '_SELF'] : null), // Edit this category
            null, // View this category
            [],
            [],
            [],
            null,
            null,
            null,
            null,
            do_lang('BLOG_NEWS_ARTICLE'),
            'news_category'
        );
    }

    /**
     * The UI to import a wordpress blog.
     *
     * @return Tempcode The UI
     */
    public function import_wordpress() : object
    {
        check_privilege('mass_import', null, null, 'cms_news');

        $lang = user_lang__with__translation_override(true);

        $submit_name = do_lang_tempcode('IMPORT_WORDPRESS');

        /* RSS method */

        require_code('news2');
        $fields = import_rss_fields(true);

        $hidden = form_input_hidden('lang', $lang);

        $xml_post_url = build_url(['page' => '_SELF', 'type' => '_import_wordpress', 'source' => 'xml'], '_SELF');

        $xml_upload_form = do_template('FORM', [
            '_GUID' => 'bdcc111acf379bab6f163f2e86d20e03',
            'TABINDEX' => strval(get_form_field_tabindex()),
            'TEXT' => '',
            'HIDDEN' => $hidden,
            'FIELDS' => $fields,
            'SUBMIT_ICON' => 'admin/import',
            'SUBMIT_NAME' => $submit_name,
            'URL' => $xml_post_url,
        ]);

        /* Database method */

        $fields = new Tempcode();

        $fields->attach(form_input_line(do_lang_tempcode('WORDPRESS_HOST_NAME'), do_lang_tempcode('DESCRIPTION_WORDPRESS_HOST_NAME'), 'wp_host', 'localhost', false));
        $fields->attach(form_input_line(do_lang_tempcode('WORDPRESS_DB_NAME'), do_lang_tempcode('DESCRIPTION_WORDPRESS_DB_NAME'), 'wp_db', 'wordpress', false));
        $fields->attach(form_input_line(do_lang_tempcode('WORDPRESS_TABLE_PREFIX'), do_lang_tempcode('DESCRIPTION_WORDPRESS_TABLE_PREFIX'), 'wp_table_prefix', 'wp', false));
        $fields->attach(form_input_line(do_lang_tempcode('WORDPRESS_DB_USERNAME'), do_lang_tempcode('DESCRIPTION_WORDPRESS_DB_USERNAME'), 'wp_db_user', 'root', false));
        $fields->attach(form_input_password(do_lang_tempcode('WORDPRESS_DB_PASSWORD'), do_lang_tempcode('DESCRIPTION_WORDPRESS_DB_PASSWORD'), 'wp_db_password', false));

        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '27350b4f39d4634a405131716a1f152c', 'SECTION_HIDDEN' => false, 'TITLE' => do_lang_tempcode('ADVANCED')]));

        $fields->attach(form_input_tick(do_lang_tempcode('IMPORT_WORDPRESS_USERS'), do_lang_tempcode('DESCRIPTION_IMPORT_WORDPRESS_USER'), 'wp_import_wordpress_users', true));
        $fields->attach(form_input_tick(do_lang_tempcode('IMPORT_BLOG_COMMENTS'), do_lang_tempcode('DESCRIPTION_IMPORT_BLOG_COMMENTS'), 'wp_import_blog_comments', true));
        if (addon_installed('validation')) {
            $fields->attach(form_input_tick(do_lang_tempcode('AUTO_VALIDATE_ALL_POSTS'), do_lang_tempcode('DESCRIPTION_VALIDATE_ALL_POSTS'), 'wp_auto_validate', true));
        }
        if ($GLOBALS['FORUM_DRIVER']->is_super_admin(get_member())) {
            $fields->attach(form_input_tick(do_lang_tempcode('ADD_TO_OWN_ACCOUNT'), do_lang_tempcode('DESCRIPTION_ADD_TO_OWN_ACCOUNT'), 'wp_to_own_account', false));
        }
        $fields->attach(form_input_tick(do_lang_tempcode('IMPORT_TO_BLOG'), do_lang_tempcode('DESCRIPTION_IMPORT_TO_BLOG'), 'wp_import_to_blog', true));
        if (has_privilege(get_member(), 'draw_to_server')) {
            $fields->attach(form_input_tick(do_lang_tempcode('DOWNLOAD_IMAGES'), do_lang_tempcode('DESCRIPTION_DOWNLOAD_IMAGES'), 'wp_download_images', true));
        }

        $hidden = new Tempcode();
        $hidden->attach(form_input_hidden('lang', $lang));
        handle_max_file_size($hidden);

        $db_post_url = build_url(['page' => '_SELF', 'type' => '_import_wordpress', 'source' => 'db'], '_SELF');

        $db_import_form = do_template('FORM', [
            '_GUID' => 'df2b4285f538bf94055c75fb8b61be6e',
            'TABINDEX' => strval(get_form_field_tabindex()),
            'TEXT' => '',
            'HIDDEN' => $hidden,
            'FIELDS' => $fields,
            'SUBMIT_ICON' => 'admin/import',
            'SUBMIT_NAME' => $submit_name,
            'URL' => $db_post_url,
        ]);

        /* Render */

        return do_template('NEWS_WORDPRESS_IMPORT_SCREEN', ['_GUID' => 'b9dcdd8211c65f3d2acd16e759d451a5', 'TITLE' => $this->title, 'XML_UPLOAD_FORM' => $xml_upload_form, 'DB_IMPORT_FORM' => $db_import_form]);
    }

    /**
     * The actualiser to import a wordpress blog.
     *
     * @return Tempcode The UI
     */
    public function _import_wordpress() : object
    {
        check_privilege('mass_import', null, null, 'cms_news');

        // Wordpress posts, XML file importing method
        if ((get_param_string('source') == 'xml')) {
            $is_validated = post_param_integer('auto_validate', 0);
            if (!addon_installed('validation')) {
                $is_validated = 1;
            }
            $download_images = post_param_integer('download_images', 0);
            if (!has_privilege(get_member(), 'draw_to_server')) {
                $download_images = 0;
            }
            $to_own_account = post_param_integer('add_to_own', 0);
            if (!$GLOBALS['FORUM_DRIVER']->is_super_admin(get_member())) {
                $to_own_account = 1;
            }
            $import_blog_comments = post_param_integer('import_blog_comments', 0);
            $import_to_blog = post_param_integer('import_to_blog', 0);

            require_code('uploads');
            is_plupload(true);
            if (((array_key_exists('file_anytype', $_FILES)) && ((is_plupload()) || (is_uploaded_file($_FILES['file_anytype']['tmp_name']))))) {
                $rss_feed = get_temporary_upload_path('file_anytype');
                $is_filesystem_path = true;
            } else {
                $rss_feed = post_param_string('rss_feed_url', null, INPUT_FILTER_URL_GENERAL);
                $is_filesystem_path = false;
            }
            if ($rss_feed === null) {
                warn_exit(do_lang_tempcode('IMPROPERLY_FILLED_IN'));
            }

            require_code('rss');
            $rss = new CMS_RSS($rss_feed, $is_filesystem_path);

            // Cleanup
            if ($is_filesystem_path) { // Means it is a temp file
                @unlink($rss_feed);
            }

            log_it('IMPORT_NEWS');

            require_code('tasks');
            $ret = call_user_func_array__long_task(do_lang('IMPORT_WORDPRESS'), $this->title, 'import_rss', [$is_validated, $download_images, $to_own_account, $import_blog_comments, $import_to_blog, $rss]);

            return $ret;
        } elseif (get_param_string('source') == 'db') { // Importing directly from wordpress DB
            $is_validated = post_param_integer('wp_auto_validate', 0);
            if (!addon_installed('validation')) {
                $is_validated = 1;
            }
            $download_images = post_param_integer('wp_download_images', 0);
            if (!has_privilege(get_member(), 'draw_to_server')) {
                $download_images = 0;
            }
            $to_own_account = post_param_integer('wp_add_to_own', 0);
            if (!$GLOBALS['FORUM_DRIVER']->is_super_admin(get_member())) {
                $to_own_account = 1;
            }
            $import_blog_comments = post_param_integer('wp_import_blog_comments', 0);
            $import_to_blog = post_param_integer('wp_import_to_blog', 0);
            if (!$GLOBALS['FORUM_DRIVER']->is_super_admin(get_member())) {
                $import_to_blog = 0;
            }
            $import_wordpress_users = (post_param_integer('wp_import_wordpress_users', 0) == 1);

            log_it('IMPORT_NEWS');

            require_code('tasks');
            return call_user_func_array__long_task(do_lang('IMPORT_WORDPRESS'), $this->title, 'import_wordpress', [$is_validated, $download_images, $to_own_account, $import_blog_comments, $import_to_blog, $import_wordpress_users]);
        }

        return new Tempcode(); // Should not get here
    }
}
