<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    forum_blocks
 */

/**
 * Block class.
 */
class Block_bottom_forum_news
{
    /**
     * Find details of the block.
     *
     * @return ?array Map of block info (null: block is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 2;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'forum_blocks';
        $info['parameters'] = ['date_key', 'param', 'forum', 'check'];
        return $info;
    }

    /**
     * Find caching details for the block.
     *
     * @return ?array Map of cache details (cache_on and ttl) (null: block is disabled)
     */
    public function caching_environment() : ?array
    {
        $info = [];
        $info['cache_on'] = <<<'PHP'
        [
            array_key_exists('param', $map) ? $map['param'] : 6,
            array_key_exists('forum', $map) ? $map['forum'] : 'Announcements',
            array_key_exists('date_key', $map) ? $map['date_key'] : 'firsttime',
            array_key_exists('check', $map) ? ($map['check'] == '1') : true,
        ]
PHP;
        $info['special_cache_flags'] = CACHE_AGAINST_DEFAULT | CACHE_AGAINST_PERMISSIVE_GROUPS;
        $info['ttl'] = 15;
        return $info;
    }

    /**
     * Execute the block.
     *
     * @param  array $map A map of parameters
     * @return Tempcode The result of execution
     */
    public function run(array $map) : object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('forum_blocks', $error_msg)) {
            return $error_msg;
        }

        if (!addon_installed('news_shared')) {
            return do_template('RED_ALERT', ['_GUID' => 'jhmazz9b50kv3de1drxujlw7p5haozvo', 'TEXT' => do_lang_tempcode('MISSING_ADDON', escape_html('news_shared'))]);
        }

        if (has_no_forum()) {
            return do_template('RED_ALERT', ['_GUID' => '5oomi6o1lipy2tch6rkgd7k9p3a6h2c8', 'TEXT' => do_lang_tempcode('NO_FORUM_INSTALLED')]);
        }

        $block_id = get_block_id($map);

        $check_perms = array_key_exists('check', $map) ? ($map['check'] == '1') : true;

        $limit = empty($map['param']) ? 6 : intval($map['param']);
        $forum_name = @cms_empty_safe($map['forum']) ? do_lang('NEWS') : $map['forum'];

        $date_key = array_key_exists('date_key', $map) ? $map['date_key'] : 'firsttime';

        $forum_ids = [];
        if ((get_forum_type() == 'cns') && ((strpos($forum_name, ',') !== false) || (preg_match('#\d[-\*\+]#', $forum_name) != 0) || (is_numeric($forum_name)))) {
            require_code('selectcode');
            $forum_names = array_map('strval', selectcode_to_idlist_using_db($forum_name, 'id', 'f_forums', 'f_forums', 'f_parent_forum', 'f_parent_forum', 'id', true, true, $GLOBALS['FORUM_DB']));
        } else {
            $forum_names = explode(',', $forum_name);
        }
        if (($forum_name == do_lang('NEWS')) && (get_forum_type() == 'cns') && ($GLOBALS['FORUM_DRIVER']->forum_id_from_name($forum_name) === null)) {
            $forum_names = [do_lang('DEFAULT_FORUM_TITLE')];
        }
        foreach ($forum_names as $forum_name) {
            $forum_name = trim($forum_name);

            $forum_id = is_numeric($forum_name) ? intval($forum_name) : $GLOBALS['FORUM_DRIVER']->forum_id_from_name($forum_name);
            if ($forum_name == '<announce>') {
                $forum_id = null;
                $forum_ids[$forum_id] = $forum_name;
            } else {
                $forum_id = is_numeric($forum_name) ? intval($forum_name) : $GLOBALS['FORUM_DRIVER']->forum_id_from_name($forum_name);
            }

            if ($forum_id !== null) {
                if ((get_forum_type() == 'cns') && ($check_perms)) {
                    if (!has_category_access(get_member(), 'forums', strval($forum_id))) {
                        continue;
                    }
                }

                $forum_ids[$forum_id] = $forum_name;
            }
        }

        $max_rows = 0;
        $topics = $GLOBALS['FORUM_DRIVER']->show_forum_topics($forum_ids, $limit, 0, $max_rows, '', '', false, $date_key);

        $out = new Tempcode();
        $_postdetailss = [];
        if ($topics !== null) {
            sort_maps_by($topics, $date_key);
            $topics = array_reverse($topics, false);

            foreach ($topics as $topic) {
                $topic_url = $GLOBALS['FORUM_DRIVER']->topic_url($topic['id'], $forum_name, true);
                $title = $topic['title'];
                $date = get_timezoned_date_tempcode($topic[$date_key]);

                $_postdetailss[] = ['DATE' => $date, 'FULL_URL' => $topic_url, 'NEWS_TITLE' => escape_html($title)];
            }
        }

        return do_template('BLOCK_BOTTOM_NEWS', [
            '_GUID' => '04d5390309dcba1f17391e9928da0d56',
            'BLOCK_ID' => $block_id,
            'POSTS' => $_postdetailss,
            'BLOG' => false,
        ]);
    }
}
