<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    syndication_blocks
 */

/**
 * Block class.
 */
class Block_bottom_rss
{
    /**
     * Find details of the block.
     *
     * @return ?array Map of block info (null: block is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 2;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'syndication_blocks';
        $info['parameters'] = ['param', 'max_entries'];
        return $info;
    }

    /**
     * Find caching details for the block.
     *
     * @return ?array Map of cache details (cache_on and ttl) (null: block is disabled)
     */
    public function caching_environment() : ?array
    {
        $info = [];
        $info['cache_on'] = ['block_bottom_rss__cache_on'];
        $info['ttl'] = intval(get_option('rss_update_time'));
        return $info;
    }

    /**
     * Execute the block.
     *
     * @param  array $map A map of parameters
     * @return Tempcode The result of execution
     */
    public function run(array $map) : object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('syndication_blocks', $error_msg)) {
            return $error_msg;
        }

        if (!addon_installed('news')) {
            return do_template('RED_ALERT', ['_GUID' => 'v8iz3vqc9lpgf554antlpja9ewhunyvm', 'TEXT' => do_lang_tempcode('MISSING_ADDON', escape_html('news'))]);
        }

        $block_id = get_block_id($map);

        $url = empty($map['param']) ? (get_brand_base_url() . '/backend.php?type=rss&mode=news') : $map['param'];

        require_code('rss');
        $rss = new CMS_RSS($url);
        if ($rss->error !== null) {
            return do_template('WARNING_BOX', ['_GUID' => '7ae6a91db7c7ac7d607b9e29ddafc344', 'WARNING' => $rss->error]);
        }

        $_postdetailss = [];

        // Now for the actual stream contents
        $max = array_key_exists('max_entries', $map) ? intval($map['max_entries']) : 10;
        $content = new Tempcode();
        foreach ($rss->gleamed_items as $i => $item) {
            if ($i >= $max) {
                break;
            }

            if (array_key_exists('full_url', $item)) {
                $full_url = $item['full_url'];
            } elseif (array_key_exists('guid', $item)) {
                $full_url = $item['guid'];
            } elseif (array_key_exists('comment_url', $item)) {
                $full_url = $item['comment_url'];
            } else {
                $full_url = '';
            }

            $_title = array_key_exists('title', $item) ? $item['title'] : do_lang('UNKNOWN');
            $date = array_key_exists('clean_add_date', $item) ? get_timezoned_date_time_tempcode($item['clean_add_date']) : (array_key_exists('add_date', $item) ? make_string_tempcode($item['add_date']) : new Tempcode());

            $_postdetailss[] = ['DATE' => $date, 'FULL_URL' => $full_url, 'NEWS_TITLE' => $_title];
        }

        return do_template('BLOCK_BOTTOM_NEWS', [
            '_GUID' => '0fc123199c4d4b7af5a26706271b1f4f',
            'BLOCK_ID' => $block_id,
            'POSTS' => $_postdetailss,
            'BLOG' => false,
        ]);
    }
}

/**
 * Find the cache signature for the block.
 *
 * @param  array $map The block parameters
 * @return array The cache signature
 */
function block_bottom_rss__cache_on(array $map) : array
{
    return [array_key_exists('param', $map) ? $map['param'] : '', array_key_exists('max_entries', $map) ? intval($map['max_entries']) : 10];
}
