<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    galleries
 */

/**
 * Block class.
 */
class Block_main_hero_slider
{
    /**
     * Find details of the block.
     *
     * @return ?array Map of block info (null: block is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Salman Abbas';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 1;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'galleries';
        $info['parameters'] = [
            'gallery_name',
            'effect',
            'fullscreen',
            'show_indicators',
            'show_scroll_down',
            'interval',
            'check',
            'blank_if_empty',
        ];

        return $info;
    }

    /**
     * Find caching details for the block.
     *
     * @return ?array Map of cache details (cache_on and ttl) (null: block is disabled)
     */
    public function caching_environment() : ?array
    {
        $info = [];
        $info['cache_on'] = <<<'PHP'
        [
            isset($map['blank_if_empty']) ? !empty($map['blank_if_empty']) : false,
            !empty($map['effect']) ? $map['effect'] : 'slide',
            !empty($map['fullscreen']),
            !empty($map['show_indicators']),
            !empty($map['show_scroll_down']),
            isset($map['interval']) ? $map['interval'] : '',
            isset($map['check']) ? !empty($map['check']) : true,
            isset($map['gallery_name']) ? $map['gallery_name'] : '',
        ]
PHP;
        $info['ttl'] = 60;
        return $info;
    }

    /**
     * Execute the block.
     *
     * @param  array $map A map of parameters
     * @return Tempcode The result of execution
     */
    public function run(array $map) : object
    {
        i_solemnly_declare(I_UNDERSTAND_SQL_INJECTION | I_UNDERSTAND_XSS | I_UNDERSTAND_PATH_INJECTION);

        require_code('galleries');
        require_code('content');
        require_css('galleries');

        $blank_if_empty = isset($map['blank_if_empty']) ? !empty($map['blank_if_empty']) : false;

        if (!addon_installed('galleries')) {
            return $blank_if_empty ? new Tempcode() : do_template('RED_ALERT', ['_GUID' => '8692692a208449e3862d6ff482dce94b', 'TEXT' => do_lang_tempcode('MISSING_ADDON', escape_html('galleries'))]);
        }

        $block_id = get_block_id($map);

        $gallery_name = @cms_empty_safe($map['gallery_name']) ? 'homepage_hero_slider' : $map['gallery_name'];
        $effect = !empty($map['effect']) ? $map['effect'] : 'slide'; // Valid values: 'fade' or 'slide'
        $fullscreen = !empty($map['fullscreen']);
        $show_indicators = !empty($map['show_indicators']);
        $show_scroll_down = !empty($map['show_scroll_down']);
        $interval = isset($map['interval']) ? strval(intval($map['interval']) * 1000) : null;
        $check_perms = isset($map['check']) ? !empty($map['check']) : true;

        // Check if the gallery exists
        $gallery_name = $GLOBALS['SITE_DB']->query_select_value_if_there('galleries', 'name', ['name' => $gallery_name]);
        if ($gallery_name === null) {
            // Maybe the user specified a gallery's full name instead of the codename?
            $gallery_name = $GLOBALS['SITE_DB']->query_select_value_if_there('galleries', 'name', ['fullname' => $gallery_name]);
        }
        if ($gallery_name === null) {
            if ($blank_if_empty) {
                return new Tempcode();
            } else {
                return do_template('RED_ALERT', ['_GUID' => '19737ba0c9c84c36b92690b7c896040d', 'TEXT' => do_lang_tempcode('MISSING_RESOURCE', 'gallery')]);
            }
        }

        if ($gallery_name === 'root') {
            $cat_select = db_string_equal_to('cat', $gallery_name);
        } else {
            require_code('selectcode');
            $cat_select = selectcode_to_sqlfragment($gallery_name, 'cat', 'galleries', 'parent_id', 'cat', 'name', false, false);
        }

        $extra_where = ' AND ' . $cat_select;
        list($rows, $max_rows) = content_rows_for_multi_type(
            ['image', 'video'],
            null,
            $extra_where,
            '',
            'recent ASC',
            0,
            100, /*reasonable limit*/
            '',
            '',
            '',
            $check_perms,
            []
        );

        require_code('images');

        if (empty($rows)) {
            if ($blank_if_empty) {
                return new Tempcode();
            }

            $submit_url = null;
            if ((has_actual_page_access(null, 'cms_galleries', null, null)) && (has_submit_permission('mid', get_member(), get_ip_address(), 'cms_galleries', ['galleries', $gallery_name])) && (can_submit_to_gallery($gallery_name))) {
                $submit_url = build_url(['page' => 'cms_galleries', 'type' => 'add', 'cat' => $gallery_name, 'redirect' => protect_url_parameter(SELF_REDIRECT_RIP)], get_module_zone('cms_galleries'));
            }
            return do_template('BLOCK_NO_ENTRIES', [
                '_GUID' => '27fb27331dc5471f977781b1c505b307',
                'TITLE' => do_lang_tempcode('GALLERY'),
                'MESSAGE' => do_lang_tempcode('NO_ENTRIES', 'image'),
                'ADD_NAME' => do_lang_tempcode('ADD_IMAGE'),
                'SUBMIT_URL' => $submit_url,
            ]);
        }

        $items = [];
        foreach ($rows as $row) {
            $full_url = $row['url'];
            if (url_is_local($full_url)) {
                $full_url = get_custom_base_url() . '/' . $full_url;
            }

            $image_url = $row[($row['content_type'] == 'image') ? 'url' : 'thumb_url'];
            if (url_is_local($image_url)) {
                $image_url = get_custom_base_url() . '/' . $image_url;
            }

            $just_media_row = db_map_restrict($row, ['id', 'the_description']);

            $description = get_translated_tempcode($row['content_type'] . 's', $just_media_row, 'the_description');

            $edit_url = null;
            if ((has_actual_page_access(get_member(), 'cms_galleries')) && (has_submit_permission('mid', get_member(), get_ip_address(), 'cms_galleries', ['galleries', $gallery_name]))) {
                $url_type = '_edit';
                if ($row['content_type'] === 'video') {
                    $url_type = '_edit_other';
                }
                $edit_url = build_url(['page' => 'cms_galleries', 'type' => $url_type, 'id' => $row['id']], get_module_zone('cms_galleries'));
            }

            $items[] = [
                'BACKGROUND_ID' => strval($row['id']),
                'BACKGROUND_TYPE' => $row['content_type'],
                'BACKGROUND_URL' => $full_url,
                'BACKGROUND_IMAGE_URL' => $image_url,
                'CONTENT_HTML' => $description,
                'EDIT_URL' => $edit_url,
            ];
        }

        return do_template('BLOCK_MAIN_HERO_SLIDER', [
            '_GUID' => '3afebc2955314f1fbc1b2d4935e998e4',
            'BLOCK_ID' => $block_id,
            'EFFECT' => $effect,
            'FULLSCREEN' => $fullscreen,
            'SHOW_INDICATORS' => $show_indicators,
            'SHOW_SCROLL_DOWN' => $show_scroll_down,
            'INTERVAL' => $interval,
            'ITEMS' => $items,
            'HAS_MULTIPLE_ITEMS' => (count($items) > 1) ? '1' : '0',
        ]);
    }
}
