<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    polls
 */

/**
 * Block class.
 */
class Block_main_poll
{
    /**
     * Find details of the block.
     *
     * @return ?array Map of block info (null: block is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 2;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'polls';
        $info['parameters'] = ['param', 'zone'];
        return $info;
    }

    /**
     * Execute the block.
     *
     * @param  array $map A map of parameters
     * @return Tempcode The result of execution
     */
    public function run(array $map) : object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('polls', $error_msg)) {
            return $error_msg;
        }

        $zone = array_key_exists('zone', $map) ? $map['zone'] : get_module_zone('polls');

        require_code('polls');
        require_css('polls');
        require_lang('polls');

        $block_id = get_block_id($map);

        // Lookup poll row
        $poll_id = null;
        if (array_key_exists('param', $map)) {
            $poll_id = intval($map['param']);
        }
        if ($poll_id === null) {
            $rows = persistent_cache_get('POLL');
            if ($rows === null) {
                $rows = $GLOBALS['SITE_DB']->query_select('poll', ['*'], ['is_current' => 1], 'ORDER BY add_time DESC', 1);
                persistent_cache_set('POLL', $rows);
            }
        } else {
            $rows = $GLOBALS['SITE_DB']->query_select('poll', ['*'], ['id' => $poll_id], '', 1);
        }
        if (!array_key_exists(0, $rows)) {
            // Action links
            if ((has_actual_page_access(null, 'cms_polls', null, null)) && (has_submit_permission('mid', get_member(), get_ip_address(), 'cms_polls'))) {
                $submit_url = build_url(['page' => 'cms_polls', 'type' => 'add', 'redirect' => protect_url_parameter(SELF_REDIRECT_RIP)], get_module_zone('cms_polls'));
            } else {
                $submit_url = new Tempcode();
            }

            return do_template('BLOCK_NO_ENTRIES', [
                '_GUID' => 'fdc85bb2e14bdf00830347e52f25cdac',
                'BLOCK_ID' => $block_id,
                'HIGH' => true,
                'TITLE' => do_lang_tempcode('POLL'),
                'MESSAGE' => do_lang_tempcode('NO_ENTRIES', 'poll'),
                'ADD_NAME' => do_lang_tempcode('ADD_POLL'),
                'SUBMIT_URL' => $submit_url,
            ]);
        }
        $myrow = $rows[0];
        $poll_id = $myrow['id'];

        // Show the poll
        $show_poll_results = get_param_integer('show_poll_results_' . strval($poll_id), 0);
        if ($show_poll_results == 0) {
            $content = render_poll_box(false, $myrow, $zone, true, false);
        } else {
            // Vote
            $cast = post_param_integer('cast_' . strval($poll_id), null);
            $myrow = vote_in_poll($poll_id, $cast, $myrow); // Either an active vote, or a forfeited vote (viewing results)

            // Show poll, with results
            $content = render_poll_box(true, $myrow, $zone, true, false);
        }

        // Render block wrapper template around poll
        return do_template('BLOCK_MAIN_POLL', [
            '_GUID' => '06a5b384015504a6a57fc4ddedbe91a7',
            'BLOCK_ID' => $block_id,
            'BLOCK_PARAMS' => comma_list_arr_to_str($map),
            'CONTENT' => $content,
        ]);
    }
}
