<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_adminzone_dashboard
 */

/**
 * Block class.
 */
class Block_main_staff_tips
{
    /**
     * Find details of the block.
     *
     * @return ?array Map of block info (null: block is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 2;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'core_adminzone_dashboard';
        $info['parameters'] = [];
        return $info;
    }

    /**
     * Uninstall the block.
     */
    public function uninstall()
    {
        $GLOBALS['SITE_DB']->drop_table_if_exists('staff_tips_dismissed');
    }

    /**
     * Install the block.
     *
     * @param  ?integer $upgrade_from What version we're upgrading from (null: new install)
     * @param  ?integer $upgrade_from_hack What hack version we're upgrading from (null: new-install/not-upgrading-from-a-hacked-version)
     */
    public function install(?int $upgrade_from = null, ?int $upgrade_from_hack = null)
    {
        $GLOBALS['SITE_DB']->create_table('staff_tips_dismissed', [
            't_member' => '*MEMBER',
            't_tip' => '*ID_TEXT',
        ]);
    }

    /**
     * Execute the block.
     *
     * @param  array $map A map of parameters
     * @return Tempcode The result of execution
     */
    public function run(array $map) : object
    {
        require_css('adminzone_dashboard');
        require_lang('tips');

        $block_id = get_block_id($map);

        // Anything to dismiss?
        $dismiss = get_param_string('staff_tips_dismiss', '');
        if ($dismiss != '') {
            $GLOBALS['SITE_DB']->query_delete('staff_tips_dismissed', ['t_tip' => $dismiss, 't_member' => get_member()], '', 1);
            $GLOBALS['SITE_DB']->query_insert('staff_tips_dismissed', ['t_tip' => $dismiss, 't_member' => get_member()]);
        }

        // What tips have been permanently dismissed by the current member?
        $read = collapse_1d_complexity('t_tip', $GLOBALS['SITE_DB']->query_select('staff_tips_dismissed', ['t_tip'], ['t_member' => get_member()]));

        $esc_free_support = escape_html(get_brand_page_url(['page' => 'contact', 'type' => 'free'], ''));
        $esc_brand_name = escape_html(brand_name());
        $remaining_tip_params = [
            escape_html(get_brand_page_url(['page' => 'professional_support'], ''/*is site, except runs with single public zone*/)),
            escape_html(get_brand_page_url(['page' => ''], 'forum')),
            escape_html(get_tutorial_url('tutorials')),
            escape_html(get_tutorial_url(null)),
            escape_html(get_tutorial_url('tut_do')),
        ];

        // Load up tips by searching for the correctly named language files; also choose level
        $tips = [];
        $level = 0;
        $letters = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z'];
        for ($i = 0; $i < 5; $i++) {
            $tips[$i] = [];
            foreach ($letters as $j) {
                $tip_id = strval($i) . $j;
                if (!in_array($tip_id, $read)) {
                    $lang2 = do_lang('TIP_' . $tip_id, null, null, null, null, false);
                    if ($lang2 !== null) {
                        $lang = do_lang_tempcode('TIP_' . $tip_id, $esc_free_support, $esc_brand_name, $remaining_tip_params);
                        $tips[$i][$tip_id] = $lang;
                    }
                }
            }
            if (empty($tips[$level])) {
                $level = $i + 1;
            }
        }

        // Choose a tip from the level we're on
        if (!array_key_exists($level, $tips)) {
            $tip = do_lang_tempcode('ALL_TIPS_READ', $esc_free_support, $esc_brand_name, [$remaining_tip_params, escape_html(get_tutorial_url('tut_do'))]);
            $level = 5;
            $tip_code = '';
            $count = 0;
        } else {
            $tip_pool = array_values($tips[$level]);
            $count = count($tip_pool);
            $choose_id = mt_rand(0, $count - 1);
            $tip = $tip_pool[$choose_id];
            $tip_keys = array_keys($tips[$level]);
            $tip_code = $tip_keys[$choose_id];
        }

        return do_template('BLOCK_MAIN_STAFF_TIPS', [
            '_GUID' => 'c2cffc480b7bd9beef7f78a8ee7b7359',
            'BLOCK_ID' => $block_id,
            'BLOCK_PARAMS' => comma_list_arr_to_str($map),
            'TIP' => $tip,
            'TIP_CODE' => $tip_code,
            'LEVEL' => integer_format($level),
            'COUNT' => integer_format($count),
        ]);
    }
}
