<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_cns
 */

/* This file exists to alleviate PHP memory usage. It shaves over 100KB of memory need for any Conversr request. */

/**
 * Add the specified custom field to the forum (some forums implemented this using proper Custom Profile Fields, others through adding a new field).
 *
 * @param  object $this_ref Link to the real forum driver
 * @param  string $name The name of the new custom field
 * @param  integer $length The length of the new custom field
 * @param  BINARY $locked Whether the field is locked
 * @param  BINARY $viewable Whether the field is for viewing
 * @param  BINARY $settable Whether the field is for setting
 * @param  BINARY $required Whether the field is required
 * @param  string $description Description
 * @param  string $type The field type (it's the same as the Composr field types, although we only expect forum drivers to specifically support short_text/long_text/integer/float and for the rest to be mapped to long_text)
 * @param  BINARY $encrypted Whether the field is encrypted
 * @param  ?string $default Default field value (null: standard for field type)
 * @param  SHORT_TEXT $options Field options
 * @param  BINARY $include_in_main_search Whether to include in main keyword search
 * @param  BINARY $allow_template_search Whether to allow template search
 * @param  ID_TEXT $icon An icon to show the CPF with on the member profiles
 * @param  ID_TEXT $section A section to show with on the member-links part of member profiles
 * @param  LONG_TEXT $tempcode This is Tempcode that is used for displaying the field. See the DESCRIPTION_CPF_CODE language string.
 * @param  ID_TEXT $autofill_type Autofill field name from https://html.spec.whatwg.org/multipage/form-control-infrastructure.html#autofill-field
 * @param  ID_TEXT $autofill_hint Autofill hint: '' or 'shipping' or 'billing'
 * @return boolean Whether the custom field was created successfully
 *
 * @ignore
 */
function _helper_install_create_custom_field(object $this_ref, string $name, int $length, int $locked = 1, int $viewable = 0, int $settable = 0, int $required = 0, string $description = '', string $type = 'long_text', int $encrypted = 0, ?string $default = null, string $options = '', int $include_in_main_search = 0, int $allow_template_search = 0, string $icon = '', string $section = '', string $tempcode = '', string $autofill_type = '', string $autofill_hint = '') : bool
{
    cns_require_all_forum_stuff();
    require_code('cns_members_action');

    $name = 'cms_' . $name;
    $id = $this_ref->db->query_select_value_if_there('f_custom_fields', 'id', [$this_ref->db->translate_field_ref('cf_name') => $name]);
    if ($id === null) {
        if ($default === null) {
            $default = (strpos($name, 'points') !== false) ? '0' : '';
        }
        $id = cns_make_custom_field($name, $locked, $description, $default, $viewable, $viewable, $settable, $encrypted, $type, $required, 0, 0, null, '', 0, $options, $include_in_main_search, $allow_template_search, $icon, $section, $tempcode, true, $autofill_type, $autofill_hint);
    }
    return $id !== null;
}

/**
 * Edit the specified custom profile field (some forums implemented this using proper Custom Profile Fields, others through adding a new field).
 *
 * @param  object $this_ref Link to the real forum driver
 * @param  ID_TEXT $old_name The name of the custom field to modify
 * @param  string $name The new name of the new custom field
 * @param  integer $length The length of the new custom field
 * @param  BINARY $locked Whether the field is locked
 * @param  BINARY $viewable Whether the field is for viewing
 * @param  BINARY $settable Whether the field is for setting
 * @param  BINARY $required Whether the field is required
 * @param  string $description Description
 * @param  string $type The field type (it's the same as the Composr field types, although we only expect forum drivers to specifically support short_text/long_text/integer/float and for the rest to be mapped to long_text)
 * @param  BINARY $encrypted Whether the field is encrypted
 * @param  ?string $default Default field value (null: standard for field type)
 * @param  SHORT_TEXT $options Field options
 * @param  BINARY $include_in_main_search Whether to include in main keyword search
 * @param  BINARY $allow_template_search Whether to allow template search
 * @param  ID_TEXT $icon An icon to show the CPF with on the member profiles
 * @param  ID_TEXT $section A section to show with on the member-links part of member profiles
 * @param  LONG_TEXT $tempcode This is Tempcode that is used for displaying the field. See the DESCRIPTION_CPF_CODE language string.
 * @param  ID_TEXT $autofill_type Autofill field name from https://html.spec.whatwg.org/multipage/form-control-infrastructure.html#autofill-field
 * @param  ID_TEXT $autofill_hint Autofill hint: '' or 'shipping' or 'billing'
 * @return boolean Whether the custom field was edited successfully
 *
 * @ignore
 */
function _helper_install_edit_custom_field(object $this_ref, string $old_name, string $name, int $length, int $locked = 1, int $viewable = 0, int $settable = 0, int $required = 0, string $description = '', string $type = 'long_text', int $encrypted = 0, ?string $default = null, string $options = '', int $include_in_main_search = 0, int $allow_template_search = 0, string $icon = '', string $section = '', string $tempcode = '', string $autofill_type = '', string $autofill_hint = '') : bool
{
    cns_require_all_forum_stuff();
    require_code('cns_members_action');

    $old_name = 'cms_' . $old_name;
    $name = 'cms_' . $name;

    $id = $this_ref->db->query_select_value_if_there('f_custom_fields', 'id', [$this_ref->db->translate_field_ref('cf_name') => $old_name]);
    if ($id === null) {
        return false;
    }

    if ($default === null) {
        $default = (strpos($name, 'points') !== false) ? '0' : '';
    }
    cns_edit_custom_field($id, $name, $description, $default, $viewable, $viewable, $settable, $encrypted, $required, 0, 0, 0, '', $type, 0, $options, $include_in_main_search, $allow_template_search, $icon, $section, $tempcode, $autofill_type, $autofill_hint);
    return true;
}

/**
 * Get an array of attributes to take in from the installer. Almost all forums require a table prefix, which the requirement there-of is defined through this function.
 * The attributes have 4 values in an array:
 * - name, the name of the attribute for _config.php
 * - default, the default value (perhaps obtained through autodetection from forum config)
 * - description, a textual description of the attributes
 * - title, a textual title of the attribute
 *
 * @return array The attributes for the forum
 *
 * @ignore
 */
function _helper_install_specifics() : array
{
    $a = [];
    $a['name'] = 'cns_table_prefix';
    $a['default'] = function_exists('get_default_table_prefix') ? get_default_table_prefix() : 'cms_';
    $a['description'] = do_lang('MOST_DEFAULT');
    $a['title'] = do_lang('TABLE_PREFIX');

    $b = [];
    $b['name'] = 'clear_existing_forums_on_install';
    $b['default'] = 'no';
    $b['description'] = do_lang_tempcode('DESCRIPTION_CLEAR_EXISTING_FORUMS_ON_INSTALL');
    $b['title'] = do_lang_tempcode('CLEAR_EXISTING_FORUMS_ON_INSTALL');

    $c = [];
    $c['name'] = 'admin_username';
    $c['default'] = 'admin';
    $c['description'] = do_lang_tempcode('DESCRIPTION_ADMIN_USERNAME');
    $c['title'] = do_lang_tempcode('ADMIN_USERNAME');
    $c['required'] = true;

    $d = [];
    $d['name'] = 'cns_admin_password';
    $d['default'] = '';
    $d['description'] = do_lang_tempcode('DESCRIPTION_ADMIN_USERS_PASSWORD');
    $d['title'] = do_lang_tempcode('ADMIN_USERS_PASSWORD');

    return [$a, $b, $c, $d];
}

/**
 * Searches for forum auto-config at this path.
 *
 * @param  PATH $path The path in which to search
 * @return boolean Whether the forum auto-config could be found
 *
 * @ignore
 */
function _helper_install_test_load_from(string $path) : bool
{
    global $PROBED_FORUM_CONFIG;

    $PROBED_FORUM_CONFIG['sql_host'] = '';
    $PROBED_FORUM_CONFIG['sql_database'] = 'cms';
    $PROBED_FORUM_CONFIG['sql_user'] = $GLOBALS['DB_DRIVER']->default_user();
    $PROBED_FORUM_CONFIG['sql_pass_exists'] = ($GLOBALS['DB_DRIVER']->default_password() != '');

    $base_url = post_param_string('base_url', get_base_url(), INPUT_FILTER_URL_GENERAL);
    $PROBED_FORUM_CONFIG['board_url'] = $base_url . '/forum';

    $PROBED_FORUM_CONFIG['cookie_member_id'] = 'cms_member_id';
    $PROBED_FORUM_CONFIG['cookie_member_hash'] = 'cms_member_hash';

    return true;
}
