<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    ecommerce
 */

/**
 * Add a usergroup subscription.
 *
 * @param  SHORT_TEXT $title The title
 * @param  LONG_TEXT $description The description
 * @param  REAL $price The price
 * @param  ID_TEXT $tax_code The tax code
 * @param  integer $length The length
 * @param  SHORT_TEXT $length_units The units for the length
 * @set y m d w
 * @param  BINARY $auto_recur Auto-recur
 * @param  GROUP $group_id The usergroup that purchasing gains membership to
 * @param  BINARY $uses_primary Whether this is applied to primary usergroup membership
 * @param  BINARY $enabled Whether this is currently enabled
 * @param  ?LONG_TEXT $mail_start The text of the e-mail to send out when a subscription is start (null: default)
 * @param  ?LONG_TEXT $mail_end The text of the e-mail to send out when a subscription is ended (null: default)
 * @param  ?LONG_TEXT $mail_uhoh The text of the e-mail to send out when a subscription cannot be renewed because the subproduct is gone (null: default)
 * @param  array $mails Other e-mails to send
 * @return AUTO_LINK The ID
 */
function add_usergroup_subscription(string $title, string $description, float $price, string $tax_code, int $length, string $length_units, int $auto_recur, int $group_id, int $uses_primary, int $enabled, ?string $mail_start, ?string $mail_end, ?string $mail_uhoh, array $mails = []) : int
{
    require_code('global4');
    prevent_double_submit('ADD_USERGROUP_SUBSCRIPTION', null, $title);

    $db = get_db_for('f_usergroup_subs');

    $map = [
        's_price' => $price,
        's_tax_code' => $tax_code,
        's_length' => $length,
        's_length_units' => $length_units,
        's_auto_recur' => $auto_recur,
        's_group_id' => $group_id,
        's_uses_primary' => $uses_primary,
        's_enabled' => $enabled,
    ];
    $map += insert_lang('s_title', $title, 2, $db);
    $map += insert_lang_comcode('s_description', $description, 2, $db);
    $map += insert_lang('s_mail_start', $mail_start, 2, $db);
    $map += insert_lang('s_mail_end', $mail_end, 2, $db);
    $map += insert_lang('s_mail_uhoh', $mail_uhoh, 2, $db);
    $id = $db->query_insert('f_usergroup_subs', $map, true);

    foreach ($mails as $mail) {
        $map = [
            'm_usergroup_sub_id' => $id,
            'm_ref_point' => $mail['ref_point'],
            'm_ref_point_offset' => $mail['ref_point_offset'],
        ];
        $map += insert_lang('m_subject', $mail['subject'], 2, $db);
        $map += insert_lang('m_body', $mail['body'], 2, $db);
        $db->query_insert('f_usergroup_sub_mails', $map);
    }

    require_lang('ecommerce');
    log_it('ADD_USERGROUP_SUBSCRIPTION', strval($id), $title);

    if ((addon_installed('commandr')) && (!running_script('install')) && (!get_mass_import_mode())) {
        require_code('resource_fs');
        generate_resource_fs_moniker('usergroup_subscription', strval($id), null, null, true);
    }

    return $id;
}

/**
 * Edit a usergroup subscription.
 *
 * @param  AUTO_LINK $id The ID
 * @param  SHORT_TEXT $title The title
 * @param  LONG_TEXT $description The description
 * @param  REAL $price The price
 * @param  ID_TEXT $tax_code The tax code
 * @param  integer $length The length
 * @param  SHORT_TEXT $length_units The units for the length
 * @set y m d w
 * @param  BINARY $auto_recur Auto-recur
 * @param  GROUP $group_id The usergroup that purchasing gains membership to
 * @param  BINARY $uses_primary Whether this is applied to primary usergroup membership
 * @param  BINARY $enabled Whether this is currently enabled
 * @param  ?LONG_TEXT $mail_start The text of the e-mail to send out when a subscription is start (null: default)
 * @param  ?LONG_TEXT $mail_end The text of the e-mail to send out when a subscription is ended (null: default)
 * @param  ?LONG_TEXT $mail_uhoh The text of the e-mail to send out when a subscription cannot be renewed because the subproduct is gone (null: default)
 * @param  ?array $mails Other e-mails to send (null: do not change)
 */
function edit_usergroup_subscription(int $id, string $title, string $description, float $price, string $tax_code, int $length, string $length_units, int $auto_recur, int $group_id, int $uses_primary, int $enabled, ?string $mail_start, ?string $mail_end, ?string $mail_uhoh, ?array $mails = null)
{
    $db = get_db_for('f_usergroup_subs');

    $rows = $db->query_select('f_usergroup_subs', ['*'], ['id' => $id], '', 1);
    if (!array_key_exists(0, $rows)) {
        warn_exit(do_lang_tempcode('MISSING_RESOURCE', 'usergroup_subscription'));
    }
    $myrow = $rows[0];

    // If usergroup has changed, do a move
    if ($myrow['s_group_id'] != $group_id) {
        require_code('cns_groups_action');
        require_code('cns_groups_action2');
        $type_code = 'USERGROUP' . strval($id);
        $subscriptions = $GLOBALS['SITE_DB']->query_select('ecom_subscriptions', ['*'], ['s_type_code' => $type_code]);
        foreach ($subscriptions as $sub) {
            $member_id = $sub['s_member_id'];
            if ((get_value('unofficial_ecommerce') === '1') && (get_forum_type() != 'cns')) {
                if ((method_exists($GLOBALS['FORUM_DRIVER'], 'remove_member_from_group')) && (method_exists($GLOBALS['FORUM_DRIVER'], 'add_member_to_group'))) {
                    $GLOBALS['FORUM_DRIVER']->remove_member_from_group($member_id, $group_id);
                    $GLOBALS['FORUM_DRIVER']->add_member_to_group($member_id, $group_id);
                }
            } else {
                $db->query_delete('f_group_members', ['gm_group_id' => $group_id, 'gm_member_id' => $member_id], '', 1);
                cns_add_member_to_secondary_group($member_id, $group_id);
                cns_update_group_approvals($member_id, get_member(), [$group_id]);
            }
        }
    }

    $_title = $myrow['s_title'];
    $_description = $myrow['s_description'];
    $_mail_start = $myrow['s_mail_start'];
    $_mail_end = $myrow['s_mail_end'];
    $_mail_uhoh = $myrow['s_mail_uhoh'];

    $map = [
        's_price' => $price,
        's_tax_code' => $tax_code,
        's_length' => $length,
        's_length_units' => $length_units,
        's_auto_recur' => $auto_recur,
        's_group_id' => $group_id,
        's_uses_primary' => $uses_primary,
        's_enabled' => $enabled,
    ];
    $map += lang_remap('s_title', $_title, $title, $db);
    $map += lang_remap_comcode('s_description', $_description, $description, $db);
    $map += lang_remap('s_mail_start', $_mail_start, $mail_start, $db);
    $map += lang_remap('s_mail_end', $_mail_end, $mail_end, $db);
    $map += lang_remap('s_mail_uhoh', $_mail_uhoh, $mail_uhoh, $db);
    $db->query_update('f_usergroup_subs', $map, ['id' => $id], '', 1);

    // Handle extra mails. Add/edit/delete as required
    if ($mails !== null) {
        $existing_mails = [];

        $start = 0;
        $max = 100;
        do {
            $_mails = $db->query_select('f_usergroup_sub_mails', ['*'], ['m_usergroup_sub_id' => $id], 'ORDER BY id', $max, $start);
            foreach ($_mails as $_mail) {
                $existing_mails[] = [$_mail['id'], $_mail['m_subject'], $_mail['m_body']];
            }

            $start += $max;
        } while (!empty($_mails));

        foreach ($mails as $i => $mail) {
            if (isset($existing_mails[$i])) {
                $map = [
                    'm_usergroup_sub_id' => $id,
                    'm_ref_point' => $mail['ref_point'],
                    'm_ref_point_offset' => $mail['ref_point_offset'],
                ];
                $map += lang_remap('m_subject', $existing_mails[$i][1], $mail['subject'], $db);
                $map += lang_remap('m_body', $existing_mails[$i][2], $mail['body'], $db);
                $db->query_update('f_usergroup_sub_mails', $map, ['id' => $existing_mails[$i][0]], '', 1);
            } else {
                $map = [
                    'm_usergroup_sub_id' => $id,
                    'm_ref_point' => $mail['ref_point'],
                    'm_ref_point_offset' => $mail['ref_point_offset'],
                ];
                $map += insert_lang('m_subject', $mail['subject'], 2, $db);
                $map += insert_lang('m_body', $mail['body'], 2, $db);
                $db->query_insert('f_usergroup_sub_mails', $map);
            }
        }
        for ($i = count($mails); $i < count($existing_mails); $i++) {
            $db->query_delete('f_usergroup_sub_mails', ['id' => $existing_mails[$i][0]], '', 1);
            delete_lang($existing_mails[$i][1], $db);
            delete_lang($existing_mails[$i][2], $db);
        }
    }

    require_lang('ecommerce');
    log_it('EDIT_USERGROUP_SUBSCRIPTION', strval($id), $title);

    if ((addon_installed('commandr')) && (!running_script('install')) && (!get_mass_import_mode())) {
        require_code('resource_fs');
        generate_resource_fs_moniker('usergroup_subscription', strval($id));
    }
}

/**
 * Delete a usergroup subscription.
 *
 * @param  AUTO_LINK $id The ID
 * @param  LONG_TEXT $uhoh_mail The cancellation mail to send out (blank: none)
 */
function delete_usergroup_subscription(int $id, string $uhoh_mail = '')
{
    $db = get_db_for('f_usergroup_subs');

    $rows = $db->query_select('f_usergroup_subs', ['*'], ['id' => $id], '', 1);
    if (!array_key_exists(0, $rows)) {
        warn_exit(do_lang_tempcode('MISSING_RESOURCE', 'usergroup_subscription'));
    }
    $myrow = $rows[0];
    $new_group = $myrow['s_group_id'];

    // Remove benefits
    $type_code = 'USERGROUP' . strval($id);
    $subscriptions = $GLOBALS['SITE_DB']->query_select('ecom_subscriptions', ['*'], ['s_type_code' => $type_code]);
    $to_members = [];
    foreach ($subscriptions as $sub) {
        $member_id = $sub['s_member_id'];

        $test = in_array($new_group, $GLOBALS['FORUM_DRIVER']->get_members_groups($member_id));
        if ($test) {
            if ($db->query_select_value_if_there('f_group_member_timeouts', 'member_id', ['member_id' => $member_id, 'group_id' => $new_group]) === null) {
                // Remove them from the group

                if ((method_exists($GLOBALS['FORUM_DRIVER'], 'remove_member_from_group')) && (get_value('unofficial_ecommerce') === '1') && (get_forum_type() != 'cns')) {
                    $GLOBALS['FORUM_DRIVER']->remove_member_from_group($member_id, $new_group);
                } else {
                    $db->query_delete('f_group_members', ['gm_group_id' => $new_group, 'gm_member_id' => $member_id], '', 1);
                }
                $to_members[] = $member_id;
            }
        }
    }
    if ($uhoh_mail != '') {
        require_code('notifications');
        dispatch_notification('paid_subscription_messages', null, do_lang('PAID_SUBSCRIPTION_ENDED', null, null, null, get_site_default_lang()), $uhoh_mail, $to_members);
    }

    $_title = $myrow['s_title'];
    $_description = $myrow['s_description'];
    $title = get_translated_text($_title, $db);
    $_mail_start = $myrow['s_mail_start'];
    $_mail_end = $myrow['s_mail_end'];
    $_mail_uhoh = $myrow['s_mail_uhoh'];

    $db->query_delete('f_usergroup_subs', ['id' => $id], '', 1);
    delete_lang($_title, $db);
    delete_lang($_description, $db);
    delete_lang($_mail_start, $db);
    delete_lang($_mail_end, $db);
    delete_lang($_mail_uhoh, $db);

    $_mails = $db->query_select('f_usergroup_sub_mails', ['*'], ['m_usergroup_sub_id' => $id]);
    foreach ($_mails as $_mail) {
        delete_lang($_mail['m_subject'], $db);
        delete_lang($_mail['m_body'], $db);
    }

    require_lang('ecommerce');
    log_it('DELETE_USERGROUP_SUBSCRIPTION', strval($id), $title);

    if ((addon_installed('commandr')) && (!running_script('install')) && (!get_mass_import_mode())) {
        require_code('resource_fs');
        expunge_resource_fs_moniker('usergroup_subscription', strval($id));
    }
}
