<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core
 */

/*
This code is not in failure.php due to the dynamic class declaration, triggering: https://bugs.php.net/bug.php?id=35634
*/

/**
 * Syndicate a spammer report out to wherever we can.
 *
 * @param  IP $ip_addr IP address to report
 * @param  ID_TEXT $username Username address to report
 * @param  EMAIL $email Email address to report
 * @param  string $reason The reason for the report (blank: none)
 * @param  boolean $trigger_error Whether to emit a Composr error, on error. Should not be 'true' for automatic spammer reports, as the spammer should not see the submission process in action!
 * @return boolean Whether it was successful
 */
function syndicate_spammer_report(string $ip_addr, string $username, string $email, string $reason, bool $trigger_error = false) : bool
{
    $did_something = false;

    // Syndicate to dnsbl.tornevall.org
    // See https://docs.tornevall.net/display/TORNEVALL/Endpoint%3A+dnsbl+-+DNSBL+v5+with+API+v3
    // ================================

    $can_do_torn = (get_option('spam_use_tornevall') != '');

    if ($can_do_torn) {
        $torn_url = 'https://api.tornevall.net/3.0/dnsbl/';

        $payload = json_encode(['ip' => [$ip_addr]]);
        require_code('character_sets');
        $payload = convert_to_internal_encoding($payload, get_charset(), 'utf-8');

        $_result = cms_http_request($torn_url, ['convert_to_internal_encoding' => true, 'http_verb' => 'PUT', 'post_params' => $payload, 'trigger_error' => false, 'ignore_http_status' => true, 'raw_content_type' => 'application/json']);

        if ($trigger_error) {
            $result = @json_decode($_result->data);
            if (is_array($result)) {
                if (isset($result['errors'])) {
                    attach_message('dnsbl.tornevall.org: ' . $result['errors']['faultstring'], 'warn', false, true);
                }
            } else {
                attach_message('dnsbl.tornevall.org: ' . $_result->message, 'warn', false, true);
            }
        }

        $did_something = true;
    }

    // Syndicate to Stop Forum Spam
    // ============================

    $stopforumspam_key = get_option('stopforumspam_api_key');
    $can_do_stopforumspam = ($stopforumspam_key != '') && ($username != '') && ($email != '');

    if ($can_do_stopforumspam) {
        require_code('files');
        require_code('character_sets');
        $url = 'https://www.stopforumspam.com/add.php?api_key=' . urlencode($stopforumspam_key) . '&ip_addr=' . urlencode($ip_addr);
        if ($username != '') {
            $url .= '&username=' . urlencode(convert_to_internal_encoding($username, get_charset(), 'utf-8'));
        }
        if ($email != '') {
            $url .= '&email=' . urlencode(convert_to_internal_encoding($email, get_charset(), 'utf-8'));
        }
        if ($reason != '') {
            $url .= '&evidence=' . urlencode(convert_to_internal_encoding($reason, get_charset(), 'utf-8'));
        }
        $result = http_get_contents($url, ['convert_to_internal_encoding' => true, 'trigger_error' => $trigger_error]);
        if (($trigger_error) && ($result != '') && (strpos($result, 'data submitted successfull') === false)) {
            attach_message($result . ' [ ' . $url . ' ]', 'warn', false, true);
        }

        $did_something = true;
    }

    // ---

    // Did we get anything done?
    if (($trigger_error) && (!$did_something)) {
        attach_message(do_lang('SPAM_REPORT_NO_EMAIL_OR_USERNAME'), 'warn');
    }

    return $did_something;
}
