<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    downloads
 */

/**
 * Hook class.
 */
class Hook_whatsnew_downloads
{
    /**
     * Find selectable (filterable) categories.
     *
     * @param  TIME $updated_since The time that there must be entries found newer than
     * @return ?array Tuple of result details: HTML list of all types that can be chosen, title for selection list (null: disabled)
     */
    public function choose_categories(int $updated_since) : ?array
    {
        if (!addon_installed('downloads')) {
            return null;
        }

        require_lang('downloads');

        require_code('downloads');
        $cats = create_selection_list_download_category_tree(null, false, false, $updated_since);

        return [$cats, do_lang('SECTION_DOWNLOADS')];
    }

    /**
     * Run function for newsletter hooks.
     *
     * @param  TIME $cutoff_time The time that the entries found must be newer than
     * @param  LANGUAGE_NAME $lang The language the entries found must be in
     * @param  string $filter Category filter to apply
     * @return array Tuple of result details
     */
    public function run(int $cutoff_time, string $lang, string $filter) : array
    {
        if (!addon_installed('downloads')) {
            return [];
        }

        require_lang('downloads');

        $max = intval(get_option('max_newsletter_whatsnew'));

        $new = new Tempcode();

        require_code('selectcode');
        $or_list = selectcode_to_sqlfragment($filter, 'category_id');

        $privacy_join = '';
        $privacy_where = '';
        if (addon_installed('content_privacy')) {
            require_code('content_privacy');
            list($privacy_join, $privacy_where) = get_privacy_where_clause('download', 'r', $GLOBALS['FORUM_DRIVER']->get_guest_id());
        }

        $rows = $GLOBALS['SITE_DB']->query('SELECT name,the_description,id,add_date,submitter FROM ' . get_table_prefix() . 'download_downloads r' . $privacy_join . ' WHERE validated=1 AND add_date>' . strval($cutoff_time) . ' AND (' . $or_list . ')' . $privacy_where . ' ORDER BY add_date DESC', $max);

        if (count($rows) == $max) {
            return [];
        }

        foreach ($rows as $row) {
            $id = $row['id'];
            $_url = build_url(['page' => 'downloads', 'type' => 'entry', 'id' => $row['id']], get_module_zone('downloads'), [], false, false, true);
            $url = $_url->evaluate();
            $name = get_translated_text($row['name'], null, $lang);
            $description = get_translated_text($row['the_description'], null, $lang);
            $member_id = (is_guest($row['submitter'])) ? null : strval($row['submitter']);
            $image_url = null;
            if (addon_installed('galleries')) {
                $gallery_images = $GLOBALS['SITE_DB']->query_select('images', ['url', 'add_date'], ['cat' => 'download_' . strval($row['id'])], 'ORDER BY add_date ASC', 1);
                if (array_key_exists(0, $gallery_images)) {
                    $image_url = $gallery_images[0]['url'];
                    if ($image_url != '') {
                        if (url_is_local($image_url)) {
                            $image_url = get_custom_base_url() . '/' . $image_url;
                        }
                        break;
                    }
                }
            }
            $new->attach(do_template('NEWSLETTER_WHATSNEW_RESOURCE_FCOMCODE', ['_GUID' => 'bbd85ed54500b9d6df998e3c835b45e9', 'MEMBER_ID' => $member_id, 'URL' => $url, 'NAME' => $name, 'DESCRIPTION' => $description, 'IMAGE_URL' => $image_url, 'CONTENT_TYPE' => 'download', 'CONTENT_ID' => strval($id)], null, false, null, '.txt', 'text'));

            require_code('urls2');
            mark_if_url_exists($url); // We know it works, so mark it valid so as to not waste CPU checking within the generated Comcode
        }

        return [$new, do_lang('SECTION_DOWNLOADS', '', '', '', $lang)];
    }
}
