<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    news
 */

/**
 * Hook class.
 */
class Hook_whatsnew_news
{
    /**
     * Find selectable (filterable) categories.
     *
     * @param  TIME $updated_since The time that there must be entries found newer than
     * @return ?array Tuple of result details: HTML list of all types that can be chosen, title for selection list (null: disabled)
     */
    public function choose_categories(int $updated_since) : ?array
    {
        if (!addon_installed('news')) {
            return null;
        }

        require_lang('news');

        require_code('news');
        $cats = create_selection_list_news_categories(null, false, false, true, null, false, $updated_since);
        return [$cats, do_lang('NEWS')];
    }

    /**
     * Run function for newsletter hooks.
     *
     * @param  TIME $cutoff_time The time that the entries found must be newer than
     * @param  LANGUAGE_NAME $lang The language the entries found must be in
     * @param  string $filter Category filter to apply
     * @param  BINARY $in_full Whether to use full article instead of summary
     * @return array Tuple of result details
     */
    public function run(int $cutoff_time, string $lang, string $filter, int $in_full = 1) : array
    {
        if (!addon_installed('news')) {
            return [];
        }

        require_lang('news');

        $max = intval(get_option('max_newsletter_whatsnew'));

        $new = new Tempcode();

        require_code('selectcode');
        $or_list = selectcode_to_sqlfragment($filter, 'news_category');
        $or_list_2 = selectcode_to_sqlfragment($filter, 'news_entry_category');

        $extra_join = '';
        $extra_where = '';
        if (addon_installed('content_privacy')) {
            require_code('content_privacy');
            list($extra_join, $extra_where) = get_privacy_where_clause('news', 'r', $GLOBALS['FORUM_DRIVER']->get_guest_id());
        }

        if (get_option('filter_regions') == '1') {
            require_code('locations');
            $extra_where .= sql_region_filter('news', 'r.id');
        }

        $rows = $GLOBALS['SITE_DB']->query('SELECT DISTINCT title,news,news_article,id,date_and_time,submitter,news_image FROM ' . get_table_prefix() . 'news r LEFT JOIN ' . get_table_prefix() . 'news_category_entries ON news_entry=id' . $extra_join . ' WHERE validated=1 AND date_and_time>' . strval($cutoff_time) . ' AND ((' . $or_list . ') OR (' . $or_list_2 . '))' . $extra_where . ' ORDER BY date_and_time DESC', $max);

        if (count($rows) == $max) {
            return [];
        }

        foreach ($rows as $row) {
            $id = $row['id'];
            $_url = build_url(['page' => 'news', 'type' => 'view', 'id' => $row['id']], get_module_zone('news'), [], false, false, true);
            $url = $_url->evaluate();
            $name = get_translated_text($row['title'], null, $lang);
            $description = get_translated_text($row[($in_full == 1) ? 'news_article' : 'news'], null, $lang);
            if ($description == '') {
                $description = get_translated_text($row[($in_full == 1) ? 'news' : 'news_article'], null, $lang);
            }
            $member_id = (is_guest($row['submitter'])) ? null : strval($row['submitter']);
            $image_url = $row['news_image'];
            if ($image_url != '') {
                if (url_is_local($image_url)) {
                    $image_url = get_custom_base_url() . '/' . $image_url;
                }
            } else {
                $image_url = null;
            }
            $new->attach(do_template('NEWSLETTER_WHATSNEW_RESOURCE_FCOMCODE', ['_GUID' => '4eaf5ec00db1f0b89cef5120c2486521', 'MEMBER_ID' => $member_id, 'URL' => $url, 'NAME' => $name, 'DESCRIPTION' => $description, 'IMAGE_URL' => $image_url, 'CONTENT_TYPE' => 'news', 'CONTENT_ID' => strval($id)], null, false, null, '.txt', 'text'));

            require_code('urls2');
            mark_if_url_exists($url); // We know it works, so mark it valid so as to not waste CPU checking within the generated Comcode
        }

        return [$new, do_lang('NEWS', '', '', '', $lang)];
    }
}
