<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    setupwizard
 */

/**
 * Hook class.
 */
class Hook_sw_core
{
    /**
     * Run function for features in the setup wizard.
     *
     * @return array Current settings
     */
    public function get_current_settings() : array
    {
        $settings = [];

        $settings['show_content_tagging'] = (get_theme_option('show_content_tagging', null, post_param_string('source_theme', 'default')) == '1') ? '1' : '0';
        $settings['show_content_tagging_inline'] = (get_theme_option('show_content_tagging_inline', null, post_param_string('source_theme', 'default')) == '1') ? '1' : '0';
        $settings['show_screen_actions'] = (get_theme_option('show_screen_actions', null, post_param_string('source_theme', 'default')) == '1') ? '1' : '0';
        $settings['single_public_zone'] = (get_option('single_public_zone') == '1') ? '1' : '0';
        $settings['guest_zone_access'] = has_zone_access($GLOBALS['FORUM_DRIVER']->get_guest_id(), 'site') ? '1' : '0';
        $settings['accept_user_submissions'] = has_zone_access($GLOBALS['FORUM_DRIVER']->get_guest_id(), 'cms') ? '1' : '0';

        return $settings;
    }

    /**
     * Run function for features in the setup wizard.
     *
     * @param  array $field_defaults Default values for the fields, from the install-profile
     * @return array A pair: Input fields, Hidden fields
     */
    public function get_fields(array $field_defaults) : array
    {
        $fields = new Tempcode();
        $hidden = new Tempcode();

        $field_defaults += $this->get_current_settings(); // $field_defaults will take precedence, due to how "+" operator works in PHP

        if (get_theme_option('setupwizard__lock_show_content_tagging', null, post_param_string('source_theme', 'default')) == '1') {
            $fields->attach(form_input_tick(do_lang_tempcode('SHOW_CONTENT_TAGGING'), do_lang_tempcode('CONFIG_OPTION_show_content_tagging'), 'show_content_tagging', $field_defaults['show_content_tagging'] == '1'));
        } else {
            $hidden->attach(form_input_hidden('show_content_tagging', get_theme_option('show_content_tagging', null, post_param_string('source_theme', 'default'))));
        }
        if (get_theme_option('setupwizard__lock_show_content_tagging_inline', null, post_param_string('source_theme', 'default')) == '1') {
            $fields->attach(form_input_tick(do_lang_tempcode('SHOW_CONTENT_TAGGING_INLINE'), do_lang_tempcode('CONFIG_OPTION_show_content_tagging_inline'), 'show_content_tagging_inline', $field_defaults['show_content_tagging_inline'] == '1'));
        } else {
            $hidden->attach(form_input_hidden('show_content_tagging_inline', get_theme_option('show_content_tagging_inline', null, post_param_string('source_theme', 'default'))));
        }
        if (get_theme_option('setupwizard__lock_show_screen_actions', null, post_param_string('source_theme', 'default')) == '1') {
            $fields->attach(form_input_tick(do_lang_tempcode('SHOW_SCREEN_ACTIONS'), do_lang_tempcode('CONFIG_OPTION_show_screen_actions'), 'show_screen_actions', $field_defaults['show_screen_actions'] == '1'));
        } else {
            $hidden->attach(form_input_hidden('show_screen_actions', get_theme_option('show_screen_actions', null, post_param_string('source_theme', 'default'))));
        }

        $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '1f8970c551c886532158e16596f9c9b8', 'TITLE' => do_lang_tempcode('menus:STRUCTURE'), 'HELP' => do_lang_tempcode('SETUPWIZARD_5x_DESCRIBE')]));

        if (get_theme_option('setupwizard__lock_single_public_zone', null, post_param_string('source_theme', 'default')) == '1') {
            $fields->attach(form_input_tick(do_lang_tempcode('SINGLE_PUBLIC_ZONE'), do_lang_tempcode('CONFIG_OPTION_single_public_zone'), 'single_public_zone', $field_defaults['single_public_zone'] == '1'));
        } else {
            $hidden->attach(form_input_hidden('single_public_zone', get_option('single_public_zone')));
        }
        $fields->attach(form_input_tick(do_lang_tempcode('GUEST_ZONE_ACCESS'), do_lang_tempcode('DESCRIPTION_GUEST_ZONE_ACCESS'), 'guest_zone_access', $field_defaults['guest_zone_access'] == '1'));
        $fields->attach(form_input_tick(do_lang_tempcode('ACCEPT_USER_SUBMISSIONS'), do_lang_tempcode('DESCRIPTION_ACCEPT_USER_SUBMISSIONS'), 'accept_user_submissions', $field_defaults['accept_user_submissions'] == '1'));

        return [$fields, new Tempcode()];
    }

    /**
     * Run function for setting features from the setup wizard.
     */
    public function set_fields()
    {
        set_option('show_content_tagging', post_param_string('show_content_tagging', '0'));
        set_option('show_content_tagging_inline', post_param_string('show_content_tagging_inline', '0'));
        set_option('show_screen_actions', post_param_string('show_screen_actions', '0'));

        // Zone structure...

        $collapse_zones = post_param_integer('single_public_zone', 0) == 1;
        set_option('single_public_zone', $collapse_zones ? '1' : '0');

        $guest_groups = $GLOBALS['FORUM_DRIVER']->get_members_groups($GLOBALS['FORUM_DRIVER']->get_guest_id());
        if (post_param_integer('guest_zone_access', 0) == 1) {
            $test = $GLOBALS['SITE_DB']->query_select_value_if_there('group_zone_access', 'zone_name', ['zone_name' => 'site', 'group_id' => $guest_groups[0]]);
            if ($test === null) {
                $GLOBALS['SITE_DB']->query_insert('group_zone_access', ['zone_name' => 'site', 'group_id' => $guest_groups[0]]);
            }
        } else {
            $GLOBALS['SITE_DB']->query_delete('group_zone_access', ['zone_name' => 'site', 'group_id' => $guest_groups[0]], '', 1);
        }

        $usergroups = $GLOBALS['FORUM_DRIVER']->get_usergroup_list();
        $admin_groups = array_unique(array_merge($GLOBALS['FORUM_DRIVER']->get_super_admin_groups(), $GLOBALS['FORUM_DRIVER']->get_moderator_groups()));
        if (post_param_integer('accept_user_submissions', 0) == 1) {
            foreach (array_keys($usergroups) as $id) {
                $test = $GLOBALS['SITE_DB']->query_select_value_if_there('group_zone_access', 'zone_name', ['zone_name' => 'cms', 'group_id' => $id]);
                if ($test === null) {
                    $GLOBALS['SITE_DB']->query_insert('group_zone_access', ['zone_name' => 'cms', 'group_id' => $id]);
                }
            }
        } else {
            foreach (array_keys($usergroups) as $id) {
                if (!in_array($id, $admin_groups)) {
                    $GLOBALS['SITE_DB']->query_delete('group_zone_access', ['zone_name' => 'cms', 'group_id' => $id]);
                }
            }
        }
    }

    /**
     * Run function for blocks in the setup wizard.
     *
     * @return array A map between block names and pairs (BLOCK_POSITION_* constants for what is supported, then a BLOCK_POSITION_* constant for what is the default)
     */
    public function get_blocks() : array
    {
        return ['main_multi_content' => [BLOCK_POSITION_MAIN, BLOCK_POSITION_MAIN]];
    }
}
