<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    stats
 */

/**
 * Hook class.
 */
class Hook_admin_stats_sitemap extends CMSStatsProvider
{
    /**
     * Find metadata about stats graphs that are provided by this stats hook.
     *
     * @param  boolean $for_kpi Whether this is for setting up a KPI
     * @return ?array Map of metadata (null: hook is disabled)
     */
    public function info(bool $for_kpi = false) : ?array
    {
        require_lang('zones');

        return [
            'sitemap_growth' => [
                'label' => do_lang_tempcode('SITEMAP'),
                'category' => 'content_growth',
                'filters' => [
                    'sitemap_growth__month_range' => new CMSStatsDateMonthRangeFilter('sitemap_growth__month_range', do_lang_tempcode('DATE_RANGE'), null, $for_kpi),
                    'sitemap_growth__page_link' => new CMSStatsTextFilter('sitemap_growth__page_link', do_lang_tempcode('PAGE_LINK')),
                ],
                'pivot' => new CMSStatsDatePivot('sitemap_growth__pivot', $this->get_date_pivots(!$for_kpi)),
                'support_kpis' => self::KPI_HIGH_IS_GOOD,
            ],
        ];
    }

    /**
     * Preprocess raw data in the database into something we can efficiently draw graphs/conclusions from.
     *
     * @param  TIME $start_time Start timestamp
     * @param  TIME $end_time End timestamp
     * @param  array $data_buckets Map of data buckets; a map of bucket name to nested maps with the following maps in sequence: 'month', 'pivot', 'value' (then further map data) ; extended and returned by reference
     */
    public function preprocess_raw_data(int $start_time, int $end_time, array &$data_buckets)
    {
        $server_timezone = get_server_timezone();

        $max = 1000;
        $start = 0;

        $date_pivots = $this->get_date_pivots();

        $query = 'SELECT * FROM ' . get_table_prefix() . 'sitemap_cache WHERE ';
        $query .= 'add_date>=' . strval($start_time) . ' AND ';
        $query .= 'add_date<=' . strval($end_time);
        $query .= ' ORDER BY add_date';
        do {
            $rows = $GLOBALS['SITE_DB']->query($query, $max, $start);
            foreach ($rows as $row) {
                $timestamp = $row['add_date'];
                $timestamp = tz_time($timestamp, $server_timezone);

                $month = get_stats_month_for_timestamp($timestamp);

                list($zone, $attributes) = page_link_decode($row['page_link']);
                $page = isset($attributes['page']) ? $attributes['page'] : DEFAULT_ZONE_PAGE_NAME;
                $page_link = $zone . ':' . $page;

                foreach (array_keys($date_pivots) as $pivot) {
                    $pivot_value = $this->calculate_date_pivot_value($pivot, $timestamp);

                    if (!isset($data_buckets['sitemap_growth'][$month][$pivot][$pivot_value][$page_link])) {
                        $data_buckets['sitemap_growth'][$month][$pivot][$pivot_value][$page_link] = 0;
                    }
                    $data_buckets['sitemap_growth'][$month][$pivot][$pivot_value][$page_link]++;
                }
            }

            $start += $max;
        } while (!empty($rows));
    }

    /**
     * Generate final data from preprocessed data.
     *
     * @param  string $bucket Data bucket we want data for
     * @param  string $pivot Pivot value
     * @param  array $filters Map of filters (including pivot if applicable)
     * @return ?array Final data in standardised map format (null: could not generate)
     */
    public function generate_final_data(string $bucket, string $pivot, array $filters) : ?array
    {
        $range = $this->convert_month_range_filter_to_pair($filters[$bucket . '__month_range']);

        $data = $this->fill_data_by_date_pivots($pivot, $range[0], $range[1]);

        $where = [
            'p_bucket' => $bucket,
            'p_pivot' => $pivot,
        ];
        $extra = '';
        $extra .= ' AND p_month>=' . strval($range[0]);
        $extra .= ' AND p_month<=' . strval($range[1]);
        $data_rows = $GLOBALS['SITE_DB']->query_select('stats_preprocessed', ['p_data'], $where, $extra);
        foreach ($data_rows as $data_row) {
            $_data = @unserialize($data_row['p_data']);
            foreach ($_data as $pivot_value => $__) {
                $pivot_value = $this->make_date_pivot_value_nice($pivot, $pivot_value);

                foreach ($__ as $page_link => $num_new_nodes) {
                    if (!empty($filters[$bucket . '__page_link'])) {
                        list($current_zone_name, $attributes) = page_link_decode($page_link);
                        $current_page_name = isset($attributes['page']) ? $attributes['page'] : DEFAULT_ZONE_PAGE_NAME;
                        if (!match_key_match($filters[$bucket . '__page_link'], false, null, $current_zone_name, $current_page_name)) {
                            continue;
                        }
                    }

                    if (!isset($data[$pivot_value])) {
                        $data[$pivot_value] = 0;
                    }
                    $data[$pivot_value] += $num_new_nodes;
                }
            }
        }

        return [
            'type' => null,
            'data' => $data,
            'x_axis_label' => do_lang_tempcode('TIME_IN_TIMEZONE', escape_html(make_nice_timezone_name(get_site_timezone()))),
            'y_axis_label' => do_lang_tempcode('COUNT_TOTAL'),
        ];
    }
}
