<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    filedump
 */

/**
 * Hook class.
 */
class Hook_search_filedump extends FieldsSearchHook
{
    /**
     * Find details for this search hook.
     *
     * @param  boolean $check_permissions Whether to check permissions
     * @param  ?MEMBER $member_id The member ID to check with (null: current member)
     * @return ~?array Map of search hook details (null: hook is disabled) (false: access denied)
     */
    public function info(bool $check_permissions = true, ?int $member_id = null)
    {
        if (!addon_installed('filedump')) {
            return null;
        }

        if ($member_id === null) {
            $member_id = get_member();
        }

        if ($check_permissions) {
            if (!has_actual_page_access($member_id, 'filedump')) {
                return false;
            }
        }

        require_code('files2');
        if (empty(get_directory_contents(get_custom_file_base() . '/uploads/filedump'))) {
            return null;
        }

        require_lang('filedump');

        $info = [];
        $info['lang'] = do_lang_tempcode('FILEDUMP');
        $info['default'] = (get_option('search_filedump') == '1');
        $info['extra_sort_fields'] = ['file_size' => do_lang_tempcode('FILE_SIZE')];

        $info['permissions'] = [
            [
                'type' => 'zone',
                'zone_name' => get_module_zone('filedump'),
            ],
            [
                'type' => 'page',
                'zone_name' => get_module_zone('filedump'),
                'page_name' => 'filedump',
            ],
        ];

        return $info;
    }

    /**
     * Get details for an ajax-tree-list of entries for the content covered by this search hook.
     *
     * @return array A pair: the hook, and the options
     */
    public function ajax_tree() : array
    {
        return ['choose_filedump_file', ['compound_list' => false, 'folder' => true]];
    }

    /**
     * Run function for search results.
     *
     * @param  string $search_query Search query
     * @param  string $content_where WHERE clause that selects the content according to the search query; passed in addition to $search_query to avoid unnecessary reparsing.  ? refers to the yet-unknown field name (blank: full-text search)
     * @param  string $where_clause Initial WHERE clause that already takes $search_under into account (should be nothing else unless it is guaranteed hook will use the global get_search_rows function)
     * @param  string $search_under Comma-separated list of categories to search under
     * @param  boolean $only_search_meta Whether to only do a META (tags) search
     * @param  boolean $only_titles Whether only to search titles (as opposed to both titles and content)
     * @param  integer $max Start position in total results
     * @param  integer $start Maximum results to return in total
     * @param  string $sort The sort type (gets remapped to a field in this function)
     * @param  ID_TEXT $direction Order direction
     * @param  SHORT_TEXT $author Username/Author to match for
     * @param  ?MEMBER $author_id Member-ID to match for (null: unknown)
     * @param  mixed $cutoff Cutoff date (TIME or a pair representing the range or null)
     * @return array List of maps (template, orderer)
     */
    public function run(string $search_query, string $content_where, string $where_clause, string $search_under, bool $only_search_meta, bool $only_titles, int $max, int $start, string $sort, string $direction, string $author, ?int $author_id, $cutoff) : array
    {
        require_lang('zones');

        // Calculate our where clause (search)
        if ($author != '') {
            return [];
        }

        require_code('files2');

        if ($search_under == '!') {
            $search_under = '';
        }

        $files = get_directory_contents(get_custom_file_base() . '/uploads/filedump' . (($search_under == '') ? '' : ('/' . $search_under)));
        $_rows = $GLOBALS['SITE_DB']->query_select('filedump');
        $rows = [];
        foreach ($_rows as $row) {
            $rows[$row['subpath']] = $row;
        }
        $i = 0;
        $out = [];
        foreach ($files as $_path) {
            if ($search_under != '') {
                $_path = $search_under . '/' . $_path;
            }

            $path = get_custom_file_base() . '/uploads/filedump/' . $_path;
            if (!$this->_handle_date_check_runtime($cutoff, filemtime($path))) {
                continue;
            }
            if (in_memory_search_match(['content' => $search_query], $path)) {
                $caption = array_key_exists($_path, $rows) ? $rows[$_path] : $_path;
                $dirs = explode('/', dirname($_path));

                $pre = '';
                $file_breadcrumbs = [];
                $breadcrumbs_page_link = build_page_link(['page' => 'filedump', 'subpath' => $pre . '/'], get_module_zone('filedump'));
                $file_breadcrumbs[] = [$breadcrumbs_page_link, do_lang_tempcode('ROOT')];
                foreach ($dirs as $dir) {
                    $breadcrumbs_page_link = build_page_link(['page' => 'filedump', 'subpath' => $pre . $dir . '/'], get_module_zone('filedump'));
                    $file_breadcrumbs[] = [$breadcrumbs_page_link, $dir];

                    $pre .= $dir . '/';
                }

                $url = get_custom_base_url() . '/uploads/filedump/' . str_replace('%2F', '/', rawurlencode($_path));

                require_code('images');
                if (!is_image($url, IMAGE_CRITERIA_WEBSAFE, true)) {
                    $tpl = paragraph(hyperlink($url, $caption, true, true), 'dfdsfu09wl;f');
                    if (!empty($file_breadcrumbs)) {
                        $tpl->attach(paragraph(do_lang_tempcode('LOCATED_IN', breadcrumb_segments_to_tempcode($file_breadcrumbs)), '', 'breadcrumbs'));
                    }

                    $out[$i]['template'] = do_template('SIMPLE_PREVIEW_BOX', [
                        '_GUID' => '51bc0cf751f4ccbd0b7f1a247b092368',
                        'ID' => $_path,
                        'TITLE' => basename($_path),
                        'SUMMARY' => $tpl,
                        'RESOURCE_TYPE' => '_filedump_file',
                    ]);
                } else {
                    $tpl = do_image_thumb($url, $caption, true, false, null, null, true);

                    $out[$i]['template'] = do_template('SIMPLE_PREVIEW_BOX', [
                        '_GUID' => '61bc0cf751f4ccbd0b7f1a247b092368',
                        'TITLE' => basename($_path),
                        'SUMMARY' => $tpl,
                        'BREADCRUMBS' => $file_breadcrumbs,
                        'URL' => $url,
                        'RESOURCE_TYPE' => '_filedump_file_image',
                    ]);
                }

                if ($sort == 'title') {
                    $out[$i]['orderer'] = $path;
                } elseif ($sort == 'add_date') {
                    $out[$i]['orderer'] = filectime($path);
                } elseif ($sort == 'file_size') {
                    $out[$i]['orderer'] = filesize($path);
                }

                $i++;
            }
        }

        return $out;
    }
}
