<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    news
 */

/**
 * Hook class.
 */
class Hook_search_news extends FieldsSearchHook
{
    /**
     * Find details for this search hook.
     *
     * @param  boolean $check_permissions Whether to check permissions
     * @param  ?MEMBER $member_id The member ID to check with (null: current member)
     * @return ~?array Map of search hook details (null: hook is disabled) (false: access denied)
     */
    public function info(bool $check_permissions = true, ?int $member_id = null)
    {
        if (!addon_installed('news')) {
            return null;
        }

        if ($member_id === null) {
            $member_id = get_member();
        }

        if ($check_permissions) {
            if (!has_actual_page_access($member_id, 'news')) {
                return false;
            }
        }

        if ($GLOBALS['SITE_DB']->query_select_value('news', 'COUNT(*)') == 0) {
            return null;
        }

        require_lang('news');

        $info = [];
        $info['lang'] = do_lang_tempcode('NEWS');
        $info['default'] = (get_option('search_news') == '1');
        $info['category'] = 'news_category';
        $info['integer_category'] = true;
        $info['extra_sort_fields'] = $this->_get_extra_sort_fields('_news');

        $info['permissions'] = [
            [
                'type' => 'zone',
                'zone_name' => get_module_zone('news'),
            ],
            [
                'type' => 'page',
                'zone_name' => get_module_zone('news'),
                'page_name' => 'news',
            ],
        ];

        return $info;
    }

    /**
     * Get a list of entries for the content covered by this search hook. In hierarchical list selection format.
     *
     * @param  string $_selected The default selected item
     * @return Tempcode Tree structure
     */
    public function get_tree(string $_selected) : object
    {
        $selected = ($_selected == '' || $_selected == '!') ? [] : [intval($_selected)];

        require_code('news');

        $tree = create_selection_list_news_categories($selected);
        return $tree;
    }

    /**
     * Get a list of extra fields to ask for.
     *
     * @return ?array A list of maps specifying extra fields (null: no tree)
     */
    public function get_fields() : ?array
    {
        return $this->_get_fields('_news');
    }

    /**
     * Run function for search results.
     *
     * @param  string $search_query Search query
     * @param  string $content_where WHERE clause that selects the content according to the search query; passed in addition to $search_query to avoid unnecessary reparsing.  ? refers to the yet-unknown field name (blank: full-text search)
     * @param  string $where_clause Initial WHERE clause that already takes $search_under into account (should be nothing else unless it is guaranteed hook will use the global get_search_rows function)
     * @param  string $search_under Comma-separated list of categories to search under
     * @param  boolean $only_search_meta Whether to only do a META (tags) search
     * @param  boolean $only_titles Whether only to search titles (as opposed to both titles and content)
     * @param  integer $max Start position in total results
     * @param  integer $start Maximum results to return in total
     * @param  string $sort The sort type (gets remapped to a field in this function)
     * @param  ID_TEXT $direction Order direction
     * @param  SHORT_TEXT $author Username/Author to match for
     * @param  ?MEMBER $author_id Member-ID to match for (null: unknown)
     * @param  mixed $cutoff Cutoff date (TIME or a pair representing the range or null)
     * @return array List of maps (template, orderer)
     */
    public function run(string $search_query, string $content_where, string $where_clause, string $search_under, bool $only_search_meta, bool $only_titles, int $max, int $start, string $sort, string $direction, string $author, ?int $author_id, $cutoff) : array
    {
        $remapped_orderer = '';
        switch ($sort) {
            case 'average_rating':
            case 'compound_rating':
                $remapped_orderer = $sort . ':news:id';
                break;

            case 'title':
                $remapped_orderer = 'title';
                break;

            case 'add_date':
                $remapped_orderer = 'date_and_time';
                break;
        }

        require_code('news');
        require_lang('news');
        require_css('news');

        // Calculate our where clause (search)
        $sq = build_search_submitter_clauses('submitter', $author_id, $author, 'author');
        if ($sq === null) {
            return [];
        } else {
            $where_clause .= $sq;
        }
        $this->_handle_date_check($cutoff, 'date_and_time', $where_clause);

        if ((!has_privilege(get_member(), 'see_not_validated')) && (addon_installed('validation'))) {
            $where_clause .= ' AND ';
            $where_clause .= 'validated=1';
        }

        $privacy_join = '';
        if (addon_installed('content_privacy')) {
            require_code('content_privacy');
            list($privacy_join, $privacy_where) = get_privacy_where_clause('news', 'r');
            $where_clause .= $privacy_where;
        }

        if (get_option('filter_regions') == '1') {
            require_code('locations');
            $where_clause .= sql_region_filter('news', 'r.id');
        }

        $table = 'news r';
        $trans_fields = ['r.title' => 'SHORT_TRANS__COMCODE', 'r.news' => 'LONG_TRANS__COMCODE', 'r.news_article' => 'LONG_TRANS__COMCODE'];
        $nontrans_fields = [];
        $this->_get_search_parameterisation_advanced_for_content_type('_news', $table, $where_clause, $trans_fields, $nontrans_fields);

        // Calculate and perform query
        $rows = get_search_rows('news', 'id', $search_query, $content_where, $where_clause, $only_search_meta, $only_titles, $max, $start, $remapped_orderer, $direction, $table . $privacy_join, 'r.*', $trans_fields, $nontrans_fields, 'news', 'news_category');

        $out = [];
        foreach ($rows as $i => $row) {
            $out[$i]['data'] = $row;
            unset($rows[$i]);
            if (($remapped_orderer != '') && (array_key_exists($remapped_orderer, $row))) {
                $out[$i]['orderer'] = $row[$remapped_orderer];
            } elseif (strpos($remapped_orderer, '_rating:') !== false) {
                $out[$i]['orderer'] = $row[$remapped_orderer];
            }
        }

        return $out;
    }

    /**
     * Run function for rendering a search result.
     *
     * @param  array $myrow The data row stored when we retrieved the result
     * @return Tempcode The output
     */
    public function render(array $myrow) : object
    {
        require_code('news');
        load_news_cat_rows('nc_owner IS NULL');
        return render_news_box($myrow);
    }
}
