<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_permission_management
 */

/**
 * Hook class.
 */
class Hook_addon_registry_core_permission_management
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of Composr this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return null;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Architecture';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Manage permissions.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_permissions',
            'tut_adv_permissions',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/adminzone/security/permissions/privileges.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_permissions.php',
            'sources/hooks/systems/addon_registry/core_permission_management.php',
            'sources/hooks/systems/sitemap/privilege_category.php',
            'themes/default/css/permissions_editor.css',
            'themes/default/images/icons/menu/adminzone/security/permissions/index.html',
            'themes/default/images/icons/menu/adminzone/security/permissions/permission_tree_editor.svg',
            'themes/default/images/icons/menu/adminzone/security/permissions/privileges.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/security/permissions/index.html',
            'themes/default/images/icons_monochrome/menu/adminzone/security/permissions/permission_tree_editor.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/security/permissions/privileges.svg',
            'themes/default/images/perm_levels/0.svg',
            'themes/default/images/perm_levels/1.svg',
            'themes/default/images/perm_levels/2.svg',
            'themes/default/images/perm_levels/3.svg',
            'themes/default/images/perm_levels/index.html',
            'themes/default/images/perm_levels/inherit.svg',
            'themes/default/images/pte_view_help.svg',
            'themes/default/javascript/core_permission_management.js',
            'themes/default/templates/PERMISSIONS_CONTENT_ACCESS_LIST.tpl',
            'themes/default/templates/PERMISSIONS_CONTENT_ACCESS_SCREEN.tpl',
            'themes/default/templates/PERMISSIONS_CONTENT_ACCESS_TICK.tpl',
            'themes/default/templates/PERMISSIONS_TREE_EDITOR_SCREEN.tpl',
            'themes/default/templates/PERMISSION_CELL.tpl',
            'themes/default/templates/PERMISSION_COLUMN_SIZER.tpl',
            'themes/default/templates/PERMISSION_HEADER_CELL.tpl',
            'themes/default/templates/PERMISSION_KEYS_MESSAGE_ROW.tpl',
            'themes/default/templates/PERMISSION_KEYS_PERMISSIONS_SCREEN.tpl',
            'themes/default/templates/PERMISSION_KEYS_PERMISSION_ROW.tpl',
            'themes/default/templates/PERMISSION_PRIVILEGES_SCREEN.tpl',
            'themes/default/templates/PERMISSION_ROW.tpl',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/PERMISSIONS_TREE_EDITOR_SCREEN.tpl' => 'administrative__permissions_tree_editor_screen',
            'templates/PERMISSION_HEADER_CELL.tpl' => 'administrative__permission_keys_permissions_screen',
            'templates/PERMISSION_CELL.tpl' => 'administrative__permission_keys_permissions_screen',
            'templates/PERMISSION_KEYS_PERMISSION_ROW.tpl' => 'administrative__permission_keys_permissions_screen',
            'templates/PERMISSION_KEYS_MESSAGE_ROW.tpl' => 'administrative__permission_keys_permissions_screen',
            'templates/PERMISSION_KEYS_PERMISSIONS_SCREEN.tpl' => 'administrative__permission_keys_permissions_screen',
            'templates/PERMISSION_COLUMN_SIZER.tpl' => 'administrative__permission_s_permissions_screen',
            'templates/PERMISSION_ROW.tpl' => 'administrative__permission_s_permissions_screen',
            'templates/PERMISSION_PRIVILEGES_SCREEN.tpl' => 'administrative__permission_s_permissions_screen',
            'templates/PERMISSIONS_CONTENT_ACCESS_SCREEN.tpl' => 'administrative__content_access_screen',
            'templates/PERMISSIONS_CONTENT_ACCESS_LIST.tpl' => 'administrative__content_access_screen',
            'templates/PERMISSIONS_CONTENT_ACCESS_TICK.tpl' => 'administrative__content_access_screen',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__permissions_tree_editor_screen() : object
    {
        require_css('sitemap_editor');

        require_javascript('tree_list');
        require_lang('permissions');

        require_css('forms');

        $groups = new Tempcode();
        foreach (placeholder_array() as $id => $group_name) {
            $groups->attach(form_input_list_entry(strval($id), false, $group_name));
        }

        return lorem_globalise(do_lorem_template('PERMISSIONS_TREE_EDITOR_SCREEN', [
            'USERGROUPS' => placeholder_array(),
            'TITLE' => lorem_screen_title(),
            'INITIAL_GROUP' => lorem_phrase(),
            'COLOR' => lorem_phrase(),
            'GROUPS' => $groups,
            'EDITOR' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__permission_keys_permissions_screen() : object
    {
        require_lang('permissions');

        require_css('forms');

        $color = 'b7b7b7';
        $header_cells = new Tempcode();
        foreach (placeholder_array() as $id => $name) {
            $header_cells->attach(do_lorem_template('PERMISSION_HEADER_CELL', [
                'COLOR' => $color,
                'GROUP' => $name,
            ]));
        }

        $header_cells->attach(do_lorem_template('PERMISSION_HEADER_CELL', [
            'COLOR' => $color,
            'GROUP' => '+/-',
        ]));

        $rows = new Tempcode();
        $p_rows = placeholder_array();
        foreach ($p_rows as $id => $page) {
            $cells = new Tempcode();
            $code = '';

            foreach (placeholder_array() as $gid => $g_name) {
                $cells->attach(do_lorem_template('PERMISSION_CELL', [
                    'CHECKED' => true,
                    'HUMAN' => lorem_phrase(),
                    'NAME' => 'p_' . strval($id) . '__' . strval($gid),
                ]));
            }

            $rows->attach(do_lorem_template('PERMISSION_KEYS_PERMISSION_ROW', [
                'ROW_MODERATOR_GROUP_CELL_IDS' => [],
                'ROW_GROUP_CELL_IDS' => [],
                'ALL_OFF' => false,
                'KEY' => lorem_word(),
                'UID' => strval($id),
                'CELLS' => $cells,
            ]));
        }

        // Match-key messages
        $m_rows = [];
        $m_rows[] = [
            'id' => 'new_1',
            'k_message' => '',
            'k_match_key' => '',
        ];

        $rows2 = new Tempcode();
        foreach ($m_rows as $row) {
            $msg = lorem_phrase();
            $rows2->attach(do_lorem_template('PERMISSION_KEYS_MESSAGE_ROW', [
                'KEY' => lorem_word(),
                'MSG' => $msg,
                'UID' => $row['id'],
            ]));
        }

        return lorem_globalise(do_lorem_template('PERMISSION_KEYS_PERMISSIONS_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'URL' => placeholder_url(),
            'HEADER_CELLS' => $header_cells,
            'ROWS' => $rows,
            'ROWS2' => $rows2,
            'COLS' => '',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__permission_s_permissions_screen() : object
    {
        require_lang('permissions');

        require_css('forms');

        $sections = new Tempcode();
        $rows = new Tempcode();
        $header_cells = new Tempcode();
        $cols = new Tempcode();
        $k = 0;
        foreach (placeholder_array() as $permission) {
            $cells = new Tempcode();

            foreach (placeholder_array() as $id => $group) {
                $k++;
                $cells->attach(do_lorem_template('PERMISSION_CELL', [
                    'CHECKED' => true,
                    'HUMAN' => lorem_phrase(),
                    'NAME' => $group . strval($k),
                ]));
            }

            $tpl_map = [
                'HAS' => lorem_word(),
                'ABBR' => $permission,
                'PERMISSION' => lorem_phrase(),
                'CELLS' => $cells,
                'CODE' => '',
                'DESCRIPTION' => lorem_phrase(),
            ];

            $rows->attach(do_lorem_template('PERMISSION_ROW', $tpl_map));

            $cols->attach(do_lorem_template('PERMISSION_COLUMN_SIZER'));

            $header_cells->attach(do_lorem_template('PERMISSION_HEADER_CELL', [
                'COLOR' => 'FF00FF',
                'GROUP' => $permission,
            ]));
        }

        $header_cells->attach(do_lorem_template('PERMISSION_HEADER_CELL', [
            'COLOR' => 'FF00FF',
            'GROUP' => '+/-',
        ]));

        return lorem_globalise(do_lorem_template('PERMISSION_PRIVILEGES_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'URL' => placeholder_url(),
            'HEADER_CELLS' => $header_cells,
            'ROWS' => $rows,
            'COLS' => '',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__content_access_screen() : object
    {
        require_lang('permissions');

        $zones = [];
        $pages = [];
        $pages[] = [
            'PAGE_NAME' => lorem_word(),
            'HAS_ACCESS' => true,
            'SAVE_ID' => lorem_word() . ':' . lorem_word(),
        ];
        $zones[] = [
            'ZONE_NAME' => lorem_word(),
            'ZONE_TITLE' => lorem_phrase(),
            'PAGES' => $pages,
            'HAS_ACCESS' => true,
            'SAVE_ID' => lorem_word(),
        ];

        $modules = [];
        $privileges = [];
        $privileges[] = [
            'PRIVILEGE_CODENAME' => lorem_word(),
            'PRIVILEGE_LABEL' => lorem_phrase(),
            'HAS_ACCESS' => true,
        ];
        $item_privileges = [];
        $item_privileges[] = [
            'ENABLED' => true,
            'PRIVILEGE_CODENAME' => lorem_word(),
            'PRIVILEGE_LABEL' => lorem_phrase(),
            'HAS_ACCESS' => true,
        ];
        $items = [];
        $items[] = [
            'ITEM_LABEL' => lorem_phrase(),
            'DEPTH' => '0',
            'ITEM_PRIVILEGES' => $item_privileges,
            'SAVE_ID' => lorem_word() . ':' . lorem_word() . ':' . lorem_word(),
            'HAS_ACCESS' => true,
        ];
        $items[] = [
            'ITEM_LABEL' => lorem_phrase(),
            'DEPTH' => '1',
            'ITEM_PRIVILEGES' => $item_privileges,
            'SAVE_ID' => lorem_word() . ':' . lorem_word() . ':' . lorem_word() . '2',
        ];
        $modules[] = [
            'CONTENT_TYPE_LABEL' => lorem_phrase(),
            'PRIVILEGES' => $privileges,
            'ITEMS' => $items,
            'SAVE_ID' => lorem_word() . ':' . lorem_word(),
        ];

        return lorem_globalise(do_lorem_template('PERMISSIONS_CONTENT_ACCESS_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'ZONES' => $zones,
            'MODULES' => $modules,
            'URL' => placeholder_url(),
            'COLOR' => 'FFFFFF',
        ]), null, '', true);
    }
}
