<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_primary_layout
 */

/**
 * Hook class.
 */
class Hook_addon_registry_core_primary_layout
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of Composr this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return null;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Architecture';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Core rendering functionality.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_themes',
            'tut_designer_themes',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/admin/component.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'sources/hooks/systems/addon_registry/core_primary_layout.php',
            'themes/default/css/helper_panel.css',
            'themes/default/css/messages.css',
            'themes/default/templates/BREADCRUMB_LINK_WRAP.tpl',
            'themes/default/templates/BREADCRUMB_LONE_WRAP.tpl',
            'themes/default/templates/BREADCRUMB_SEPARATOR.tpl',
            'themes/default/templates/CLOSED_SITE.tpl',
            'themes/default/templates/CSS_NEED.tpl',
            'themes/default/templates/CSS_NEED_FULL.tpl',
            'themes/default/templates/CSS_NEED_INLINE.tpl',
            'themes/default/templates/GLOBAL_HELPER_PANEL.tpl',
            'themes/default/templates/GLOBAL_HTML_WRAP.tpl',
            'themes/default/templates/HEADER_CLASSIC.tpl',
            'themes/default/templates/HEADER_MODERN.tpl',
            'themes/default/templates/HEADER_SIDE.tpl',
            'themes/default/templates/JAVASCRIPT_NEED.tpl',
            'themes/default/templates/JAVASCRIPT_NEED_FULL.tpl',
            'themes/default/templates/MAIL.tpl',
            'themes/default/templates/MAIL_RAW.tpl',
            'themes/default/templates/MESSAGE.tpl',
            'themes/default/templates/RSS_HEADER.tpl',
            'themes/default/templates/SCREEN_TITLE.tpl',
            'themes/default/text/MAIL.txt',
            'themes/default/text/MAIL_SUBJECT.txt',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/BREADCRUMB_SEPARATOR.tpl' => 'breadcrumb',
            'templates/BREADCRUMB_LONE_WRAP.tpl' => 'breadcrumb',
            'templates/BREADCRUMB_LINK_WRAP.tpl' => 'breadcrumb',
            'templates/CLOSED_SITE.tpl' => 'closed_site',
            'templates/MESSAGE.tpl' => 'message',
            'text/MAIL_SUBJECT.txt' => 'mail_subject',
            'templates/MAIL.tpl' => 'mail',
            'templates/MAIL_RAW.tpl' => 'mail_raw',
            'templates/GLOBAL_HTML_WRAP.tpl' => 'global_html_wrap',
            'templates/JAVASCRIPT_NEED.tpl' => 'global_html_wrap',
            'templates/JAVASCRIPT_NEED_FULL.tpl' => 'javascript_need_full',
            'templates/CSS_NEED_FULL.tpl' => 'css_need_full',
            'templates/CSS_NEED_INLINE.tpl' => 'css_need_inline',
            'templates/CSS_NEED.tpl' => 'global_html_wrap',
            'templates/RSS_HEADER.tpl' => 'global_html_wrap',
            'templates/GLOBAL_HELPER_PANEL.tpl' => 'global_html_wrap',
            'templates/SCREEN_TITLE.tpl' => 'screen_title',
            'templates/HEADER_CLASSIC.tpl' => 'header_classic',
            'templates/HEADER_MODERN.tpl' => 'header_modern',
            'templates/HEADER_SIDE.tpl' => 'header_side',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__breadcrumb() : object
    {
        $out = new Tempcode();
        $out->attach(do_lorem_template('BREADCRUMB_LINK_WRAP', ['URL' => placeholder_url(), 'TOOLTIP' => '', 'LABEL' => lorem_phrase()]));
        $out->attach(do_lorem_template('BREADCRUMB_SEPARATOR', []));
        $out->attach(do_lorem_template('BREADCRUMB_LONE_WRAP', ['LABEL' => lorem_phrase()]));
        return lorem_globalise($out, null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__closed_site() : object
    {
        return lorem_globalise(do_lorem_template('CLOSED_SITE', [
            'CLOSED' => lorem_phrase(),
            'LOGIN_URL' => placeholder_url(),
            'JOIN_URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__message() : object
    {
        return lorem_globalise(do_lorem_template('MESSAGE', [
            'TYPE' => placeholder_img_code('icons/status'),
            'MESSAGE' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__mail_subject() : object
    {
        return lorem_globalise(do_lorem_template('MAIL_SUBJECT', [
            'SUBJECT_LINE' => lorem_word(),
        ], null, false, null, '.txt', 'text'), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__mail() : object
    {
        return $this->do_mail_template_preview('MAIL');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__mail_raw() : object
    {
        return $this->do_mail_template_preview('MAIL_RAW');
    }

    /**
     * Get an e-mail template preview.
     *
     * @param  ID_TEXT $tpl Template name
     * @return Tempcode Preview
     */
    protected function do_mail_template_preview(string $tpl) : object
    {
        $message_raw = comcode_to_tempcode(do_lang('NEW_COMMENT_BODY', lorem_phrase(), lorem_phrase(), [lorem_phrase(), lorem_paragraph(), static_evaluate_tempcode(placeholder_url()), lorem_phrase(), placeholder_codename(), lorem_phrase()]));

        require_css('email');

        $map = [
            'LOGOURL' => placeholder_image_url(),
            'LOGOMAP' => '',
            'LANG' => fallback_lang(),
            'TITLE' => lorem_phrase(),
            'CONTENT' => $message_raw,
            'CSS' => '{CSS}',
        ];

        $_tpl = do_lorem_template($tpl, $map);

        $css = css_tempcode(true, false, $_tpl->evaluate());
        $_css = $css->evaluate();

        return make_string_tempcode(str_replace('{CSS}', $_css, $_tpl->evaluate()));
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__global_html_wrap() : object
    {
        return do_lorem_template('GLOBAL_HTML_WRAP', [
            'MIDDLE' => lorem_paragraph_html(),
        ]);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__screen_title() : object
    {
        $awards = null;
        if (addon_installed('awards')) {
            require_lang('awards');
            $awards = [
                [
                    'AWARD_TYPE' => lorem_screen_title(),
                    'AWARD_TIMESTAMP' => placeholder_date(),
                ]
            ];
        }
        return lorem_globalise(do_lorem_template('SCREEN_TITLE', [
            'TITLE' => lorem_phrase(),
            'AWARDS' => $awards,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__header_classic() : object
    {
        return do_lorem_template('HEADER_CLASSIC', []);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__header_modern() : object
    {
        return do_lorem_template('HEADER_MODERN', []);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__header_side() : object
    {
        return do_lorem_template('HEADER_SIDE', []);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__javascript_need_full() : object
    {
        return lorem_globalise(do_lorem_template('JAVASCRIPT_NEED_FULL', [
            'URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__css_need_full() : object
    {
        return lorem_globalise(do_lorem_template('CSS_NEED_FULL', [
            'URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__css_need_inline() : object
    {
        return lorem_globalise(do_lorem_template('CSS_NEED_INLINE', [
            'CODE' => '',
        ]), null, '', true);
    }
}
