<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    ecommerce
 */

/**
 * Hook class.
 */
class Hook_addon_registry_ecommerce
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        $chmod = [];
        if ($runtime) {
            $chmod = array_merge(
                $chmod,
                [
                    'uploads/ecommerce/*',
                ]
            );
        }
        $chmod = array_merge(
            $chmod,
            [
                'uploads/ecommerce',
            ]
        );
        return $chmod;
    }

    /**
     * Get the version of Composr this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return null;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'eCommerce';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'eCommerce infrastructure, with support for digital purchase and usergroup subscriptions. Provides a number of virtual products to your members in exchange for money or points. Accounting functionality.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_ecommerce',
            'tut_points',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/adminzone/audit/ecommerce/ecommerce.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_ecommerce.php',
            'adminzone/pages/modules/admin_ecommerce_reports.php',
            'adminzone/pages/modules/admin_invoices.php',
            'data/ecommerce.php',
            'lang/EN/ecommerce.ini',
            'site/pages/modules/invoices.php',
            'site/pages/modules/purchase.php',
            'site/pages/modules/subscriptions.php',
            'sources/currency.php',
            'sources/ecommerce.php',
            'sources/ecommerce2.php',
            'sources/ecommerce_permission_products.php',
            'sources/ecommerce_reports.php',
            'sources/ecommerce_shipping.php',
            'sources/ecommerce_subscriptions.php',
            'sources/ecommerce_tax.php',
            'sources/hooks/blocks/main_staff_checklist/ecommerce_sales.php',
            'sources/hooks/modules/admin_import_types/ecommerce.php',
            'sources/hooks/modules/admin_stats/subscriptions.php',
            'sources/hooks/modules/admin_stats/transactions.php',
            'sources/hooks/modules/members/ecommerce.php',
            'sources/hooks/systems/actionlog/ecommerce.php',
            'sources/hooks/systems/addon_registry/ecommerce.php',
            'sources/hooks/systems/cns_cpf_filter/ecommerce.php',
            'sources/hooks/systems/commandr_fs/usergroup_subscriptions.php',
            'sources/hooks/systems/commandr_fs_extended_config/ecom_prods_custom.php',
            'sources/hooks/systems/commandr_fs_extended_config/ecom_prods_permissions.php',
            'sources/hooks/systems/commandr_fs_extended_config/ecom_prods_prices.php',
            'sources/hooks/systems/commandr_fs_extended_member/invoices.php',
            'sources/hooks/systems/commandr_fs_extended_member/subscriptions.php',
            'sources/hooks/systems/config/average_gamble_multiplier.php',
            'sources/hooks/systems/config/banner_hit_price.php',
            'sources/hooks/systems/config/banner_hit_price_points.php',
            'sources/hooks/systems/config/banner_hit_tax_code.php',
            'sources/hooks/systems/config/banner_imp_price.php',
            'sources/hooks/systems/config/banner_imp_price_points.php',
            'sources/hooks/systems/config/banner_imp_tax_code.php',
            'sources/hooks/systems/config/banner_setup_price.php',
            'sources/hooks/systems/config/banner_setup_price_points.php',
            'sources/hooks/systems/config/banner_setup_tax_code.php',
            'sources/hooks/systems/config/business_city.php',
            'sources/hooks/systems/config/business_country.php',
            'sources/hooks/systems/config/business_county.php',
            'sources/hooks/systems/config/business_name.php',
            'sources/hooks/systems/config/business_post_code.php',
            'sources/hooks/systems/config/business_state.php',
            'sources/hooks/systems/config/business_street_address.php',
            'sources/hooks/systems/config/credit_card_cleanup_days.php',
            'sources/hooks/systems/config/currency.php',
            'sources/hooks/systems/config/currency_api_key.php',
            'sources/hooks/systems/config/currency_auto.php',
            'sources/hooks/systems/config/days_to_keep__ecommerce_log.php',
            'sources/hooks/systems/config/download_cat_buy_max_emailed_count.php',
            'sources/hooks/systems/config/download_cat_buy_max_emailed_size.php',
            'sources/hooks/systems/config/ecom_price_honour_time.php',
            'sources/hooks/systems/config/ecommerce_test_mode.php',
            'sources/hooks/systems/config/enable_taxjar_address_validation.php',
            'sources/hooks/systems/config/forw_url.php',
            'sources/hooks/systems/config/highlight_name_price.php',
            'sources/hooks/systems/config/highlight_name_price_points.php',
            'sources/hooks/systems/config/highlight_name_tax_code.php',
            'sources/hooks/systems/config/initial_banner_hits.php',
            'sources/hooks/systems/config/initial_quota.php',
            'sources/hooks/systems/config/is_on_banner_buy.php',
            'sources/hooks/systems/config/is_on_forw_buy.php',
            'sources/hooks/systems/config/is_on_gambling_buy.php',
            'sources/hooks/systems/config/is_on_highlight_name_buy.php',
            'sources/hooks/systems/config/is_on_pop3_buy.php',
            'sources/hooks/systems/config/is_on_topic_pin_buy.php',
            'sources/hooks/systems/config/mail_server.php',
            'sources/hooks/systems/config/manual_subscription_expiry_notice.php',
            'sources/hooks/systems/config/max_ip_addresses_per_subscriber.php',
            'sources/hooks/systems/config/max_quota.php',
            'sources/hooks/systems/config/maximum_gamble_amount.php',
            'sources/hooks/systems/config/maximum_gamble_multiplier.php',
            'sources/hooks/systems/config/minimum_gamble_amount.php',
            'sources/hooks/systems/config/payment_gateway.php',
            'sources/hooks/systems/config/payment_gateway_digest.php',
            'sources/hooks/systems/config/payment_gateway_password.php',
            'sources/hooks/systems/config/payment_gateway_special_identifier.php',
            'sources/hooks/systems/config/payment_gateway_username.php',
            'sources/hooks/systems/config/payment_gateway_vpn_password.php',
            'sources/hooks/systems/config/payment_gateway_vpn_username.php',
            'sources/hooks/systems/config/payment_memos.php',
            'sources/hooks/systems/config/pd_email.php',
            'sources/hooks/systems/config/pd_number.php',
            'sources/hooks/systems/config/pop_url.php',
            'sources/hooks/systems/config/quota_price.php',
            'sources/hooks/systems/config/quota_price_points.php',
            'sources/hooks/systems/config/quota_tax_code.php',
            'sources/hooks/systems/config/quota_url.php',
            'sources/hooks/systems/config/shipping_cost_base.php',
            'sources/hooks/systems/config/shipping_cost_factor.php',
            'sources/hooks/systems/config/shipping_density.php',
            'sources/hooks/systems/config/shipping_distance_units.php',
            'sources/hooks/systems/config/shipping_shippo_api_live.php',
            'sources/hooks/systems/config/shipping_shippo_api_test.php',
            'sources/hooks/systems/config/shipping_tax_code.php',
            'sources/hooks/systems/config/shipping_weight_units.php',
            'sources/hooks/systems/config/store_credit_card_numbers.php',
            'sources/hooks/systems/config/tax_api_service.php',
            'sources/hooks/systems/config/tax_country_regexp.php',
            'sources/hooks/systems/config/tax_detailed.php',
            'sources/hooks/systems/config/tax_number.php',
            'sources/hooks/systems/config/tax_state_regexp.php',
            'sources/hooks/systems/config/tax_system.php',
            'sources/hooks/systems/config/taxcloud_api_id.php',
            'sources/hooks/systems/config/taxcloud_api_key.php',
            'sources/hooks/systems/config/taxjar_api_key.php',
            'sources/hooks/systems/config/taxjar_api_sandbox.php',
            'sources/hooks/systems/config/topic_pin_price.php',
            'sources/hooks/systems/config/topic_pin_price_points.php',
            'sources/hooks/systems/config/topic_pin_tax_code.php',
            'sources/hooks/systems/config/transaction_flat_fee.php',
            'sources/hooks/systems/config/transaction_percentage_fee.php',
            'sources/hooks/systems/config/use_local_payment.php',
            'sources/hooks/systems/config_categories/ecommerce.php',
            'sources/hooks/systems/config_categories/ecommerce_products.php',
            'sources/hooks/systems/cron/credit_card_cleanup.php',
            'sources/hooks/systems/cron/ip_address_sharing.php',
            'sources/hooks/systems/cron/manual_subscription_notification.php',
            'sources/hooks/systems/cron/subscription_mails.php',
            'sources/hooks/systems/cron/topic_pin.php',
            'sources/hooks/systems/ecommerce/.htaccess',
            'sources/hooks/systems/ecommerce/custom.php',
            'sources/hooks/systems/ecommerce/email.php',
            'sources/hooks/systems/ecommerce/gambling.php',
            'sources/hooks/systems/ecommerce/highlight_name.php',
            'sources/hooks/systems/ecommerce/index.html',
            'sources/hooks/systems/ecommerce/interest.php',
            'sources/hooks/systems/ecommerce/other.php',
            'sources/hooks/systems/ecommerce/permission.php',
            'sources/hooks/systems/ecommerce/tax.php',
            'sources/hooks/systems/ecommerce/topic_pin.php',
            'sources/hooks/systems/ecommerce/usergroup.php',
            'sources/hooks/systems/ecommerce/wage.php',
            'sources/hooks/systems/ecommerce/work.php',
            'sources/hooks/systems/ecommerce_tax/.htaccess',
            'sources/hooks/systems/ecommerce_tax/composr.php',
            'sources/hooks/systems/ecommerce_tax/index.html',
            'sources/hooks/systems/ecommerce_tax/taxcloud.php',
            'sources/hooks/systems/ecommerce_tax/taxjar.php',
            'sources/hooks/systems/fields/tax_code.php',
            'sources/hooks/systems/health_checks/ecommerce.php',
            'sources/hooks/systems/logs/ecommerce.php',
            'sources/hooks/systems/notifications/ecommerce.php',
            'sources/hooks/systems/notifications/ecommerce_staff.php',
            'sources/hooks/systems/page_groupings/ecommerce.php',
            'sources/hooks/systems/payment_gateway/.htaccess',
            'sources/hooks/systems/payment_gateway/authorize.php',
            'sources/hooks/systems/payment_gateway/ccbill.php',
            'sources/hooks/systems/payment_gateway/index.html',
            'sources/hooks/systems/payment_gateway/paypal.php',
            'sources/hooks/systems/privacy/ecommerce.php',
            'sources/hooks/systems/realtime_rain/ecommerce.php',
            'sources/hooks/systems/resource_meta_aware/usergroup_subscription.php',
            'sources/hooks/systems/symbols/CURRENCY.php',
            'sources/hooks/systems/symbols/CURRENCY_SYMBOL.php',
            'sources/hooks/systems/symbols/CURRENCY_USER.php',
            'sources/hooks/systems/tasks/export_ecom_sales.php',
            'sources/hooks/systems/tasks/export_ecom_subscriptions.php',
            'sources/hooks/systems/tasks/export_ecom_transactions.php',
            'sources/hooks/systems/trusted_sites/ecommerce.php',
            'sources_custom/hooks/systems/ecommerce/.htaccess',
            'sources_custom/hooks/systems/ecommerce/index.html',
            'sources_custom/hooks/systems/payment_gateway/.htaccess',
            'sources_custom/hooks/systems/payment_gateway/index.html',
            'themes/default/css/ecommerce.css',
            'themes/default/images/ecommerce/checkmark.svg',
            'themes/default/images/ecommerce/index.html',
            'themes/default/images/icons/menu/adminzone/audit/ecommerce/cash_flow.svg',
            'themes/default/images/icons/menu/adminzone/audit/ecommerce/create_invoice.svg',
            'themes/default/images/icons/menu/adminzone/audit/ecommerce/ecommerce.svg',
            'themes/default/images/icons/menu/adminzone/audit/ecommerce/index.html',
            'themes/default/images/icons/menu/adminzone/audit/ecommerce/invoices.svg',
            'themes/default/images/icons/menu/adminzone/audit/ecommerce/profit_loss.svg',
            'themes/default/images/icons/menu/adminzone/audit/ecommerce/sales.svg',
            'themes/default/images/icons/menu/adminzone/audit/ecommerce/subscriptions.svg',
            'themes/default/images/icons/menu/adminzone/audit/ecommerce/transactions_log.svg',
            'themes/default/images/icons/menu/adminzone/setup/ecommerce_products.svg',
            'themes/default/images/icons/menu/rich_content/ecommerce/index.html',
            'themes/default/images/icons/menu/rich_content/ecommerce/purchase.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/audit/ecommerce/cash_flow.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/audit/ecommerce/create_invoice.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/audit/ecommerce/ecommerce.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/audit/ecommerce/index.html',
            'themes/default/images/icons_monochrome/menu/adminzone/audit/ecommerce/invoices.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/audit/ecommerce/profit_loss.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/audit/ecommerce/sales.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/audit/ecommerce/subscriptions.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/audit/ecommerce/transactions_log.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/setup/ecommerce_products.svg',
            'themes/default/images/icons_monochrome/menu/rich_content/ecommerce/index.html',
            'themes/default/images/icons_monochrome/menu/rich_content/ecommerce/purchase.svg',
            'themes/default/javascript/ecommerce.js',
            'themes/default/templates/CNS_MEMBER_ECOMMERCE_REPORTS.tpl',
            'themes/default/templates/CURRENCY.tpl',
            'themes/default/templates/ECOM_CASH_FLOW_SCREEN.tpl',
            'themes/default/templates/ECOM_INVOICES_SCREEN.tpl',
            'themes/default/templates/ECOM_INVOICE_OR_RECEIPT.tpl',
            'themes/default/templates/ECOM_LOGOS_AUTHORIZE.tpl',
            'themes/default/templates/ECOM_MEMBER_SUBSCRIPTION_STATUS.tpl',
            'themes/default/templates/ECOM_PAYMENT_PROCESSOR_LINKS_AUTHORIZE.tpl',
            'themes/default/templates/ECOM_PRODUCTS_PRICES_FORM_WRAP.tpl',
            'themes/default/templates/ECOM_PRODUCT_PRICE_SCREEN.tpl',
            'themes/default/templates/ECOM_PURCHASE_SCREEN.tpl',
            'themes/default/templates/ECOM_PURCHASE_STAGE_CHOOSE.tpl',
            'themes/default/templates/ECOM_PURCHASE_STAGE_DETAILS.tpl',
            'themes/default/templates/ECOM_PURCHASE_STAGE_FINISH.tpl',
            'themes/default/templates/ECOM_PURCHASE_STAGE_GUEST.tpl',
            'themes/default/templates/ECOM_PURCHASE_STAGE_MESSAGE.tpl',
            'themes/default/templates/ECOM_PURCHASE_STAGE_PAY.tpl',
            'themes/default/templates/ECOM_PURCHASE_STAGE_TERMS.tpl',
            'themes/default/templates/ECOM_PURCHASE_STAGE_TRANSACT.tpl',
            'themes/default/templates/ECOM_SALES_SCREEN.tpl',
            'themes/default/templates/ECOM_SUBSCRIPTIONS_SCREEN.tpl',
            'themes/default/templates/ECOM_SUBSCRIPTION_BUTTON_VIA_AUTHORIZE.tpl',
            'themes/default/templates/ECOM_SUBSCRIPTION_BUTTON_VIA_CCBILL.tpl',
            'themes/default/templates/ECOM_SUBSCRIPTION_BUTTON_VIA_PAYPAL.tpl',
            'themes/default/templates/ECOM_SUBSCRIPTION_CANCEL_BUTTON_VIA_AUTHORIZE.tpl',
            'themes/default/templates/ECOM_SUBSCRIPTION_CANCEL_BUTTON_VIA_CCBILL.tpl',
            'themes/default/templates/ECOM_SUBSCRIPTION_CANCEL_BUTTON_VIA_PAYPAL.tpl',
            'themes/default/templates/ECOM_TRANSACTION_BUTTON_VIA_AUTHORIZE.tpl',
            'themes/default/templates/ECOM_TRANSACTION_BUTTON_VIA_CCBILL.tpl',
            'themes/default/templates/ECOM_TRANSACTION_BUTTON_VIA_PAYPAL.tpl',
            'themes/default/templates/ECOM_TRANSACTION_LOGS_MANUAL_TRIGGER.tpl',
            'themes/default/text/ECOM_INVOICE_FULFILLED_MAIL.txt',
            'themes/default/text/ECOM_INVOICE_MAIL.txt',
            'themes/default/text/ECOM_PAYMENT_RECEIVED_MAIL.txt',
            'themes/default/text/ECOM_PAYMENT_SENT_MAIL.txt',
            'themes/default/text/ECOM_PRODUCT_FORWARDER_MAIL.txt',
            'themes/default/text/ECOM_PRODUCT_POP3_MAIL.txt',
            'themes/default/text/ECOM_PRODUCT_QUOTA_MAIL.txt',
            'uploads/ecommerce/.htaccess',
            'uploads/ecommerce/index.html',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/CURRENCY.tpl' => 'currency',
            'templates/ECOM_TRANSACTION_LOGS_MANUAL_TRIGGER.tpl' => 'ecom_subscriptions_screen',
            'templates/ECOM_CASH_FLOW_SCREEN.tpl' => 'administrative__ecom_cash_flow_screen',
            'templates/ECOM_INVOICE_OR_RECEIPT.tpl' => 'ecom_invoice_or_receipt_screen',
            'templates/ECOM_PURCHASE_SCREEN.tpl' => 'purchase_screen',
            'templates/ECOM_LOGOS_AUTHORIZE.tpl' => 'ecom_logos_authorize',
            'templates/ECOM_PAYMENT_PROCESSOR_LINKS_AUTHORIZE.tpl' => 'ecom_payment_processor_links_authorize',
            'templates/ECOM_TRANSACTION_BUTTON_VIA_PAYPAL.tpl' => 'ecom_transaction_button_via_paypal',
            'templates/ECOM_SUBSCRIPTION_BUTTON_VIA_PAYPAL.tpl' => 'ecom_subscription_button_via_paypal',
            'templates/ECOM_SUBSCRIPTION_CANCEL_BUTTON_VIA_PAYPAL.tpl' => 'ecom_subscription_cancel_button_via_paypal',
            'templates/ECOM_TRANSACTION_BUTTON_VIA_CCBILL.tpl' => 'ecom_transaction_button_via_ccbill',
            'templates/ECOM_SUBSCRIPTION_BUTTON_VIA_CCBILL.tpl' => 'ecom_subscription_button_via_ccbill',
            'templates/ECOM_SUBSCRIPTION_CANCEL_BUTTON_VIA_CCBILL.tpl' => 'ecom_subscription_cancel_button_via_ccbill',
            'templates/ECOM_TRANSACTION_BUTTON_VIA_AUTHORIZE.tpl' => 'ecom_transaction_button_via_authorize',
            'templates/ECOM_SUBSCRIPTION_BUTTON_VIA_AUTHORIZE.tpl' => 'ecom_subscription_button_via_authorize',
            'templates/ECOM_SUBSCRIPTION_CANCEL_BUTTON_VIA_AUTHORIZE.tpl' => 'ecom_subscription_cancel_button_via_authorize',
            'templates/ECOM_PURCHASE_STAGE_GUEST.tpl' => 'purchase_stage_guest',
            'templates/ECOM_PURCHASE_STAGE_CHOOSE.tpl' => 'purchase_stage_choose',
            'templates/ECOM_PURCHASE_STAGE_MESSAGE.tpl' => 'purchase_stage_message',
            'templates/ECOM_PURCHASE_STAGE_TERMS.tpl' => 'purchase_stage_terms',
            'templates/ECOM_PURCHASE_STAGE_DETAILS.tpl' => 'purchase_stage_details',
            'templates/ECOM_PURCHASE_STAGE_TRANSACT.tpl' => 'purchase_stage_transact',
            'templates/ECOM_PURCHASE_STAGE_PAY.tpl' => 'purchase_stage_pay',
            'templates/ECOM_PURCHASE_STAGE_FINISH.tpl' => 'purchase_stage_finish',
            'templates/ECOM_INVOICES_SCREEN.tpl' => 'ecom_invoices_screen',
            'templates/ECOM_SUBSCRIPTIONS_SCREEN.tpl' => 'ecom_subscriptions_screen',
            'templates/ECOM_MEMBER_SUBSCRIPTION_STATUS.tpl' => 'member_subscription_status_screen',
            'templates/ECOM_SALES_SCREEN.tpl' => 'administrative__ecom_sales_screen',
            'templates/ECOM_PRODUCTS_PRICES_FORM_WRAP.tpl' => 'administrative__ecom_products_price_screen',
            'templates/ECOM_PRODUCT_PRICE_SCREEN.tpl' => 'administrative__ecom_products_price_screen',
            'templates/CNS_MEMBER_ECOMMERCE_REPORTS.tpl' => 'administrative__cns_member_profile_ecommerce_reports_screen',
            'text/ECOM_PRODUCT_FORWARDER_MAIL.txt' => 'ecom_product_forwarder_mail',
            'text/ECOM_PRODUCT_POP3_MAIL.txt' => 'ecom_product_pop3_mail',
            'text/ECOM_PRODUCT_QUOTA_MAIL.txt' => 'ecom_product_quota_mail',
            'text/ECOM_PAYMENT_RECEIVED_MAIL.txt' => 'ecom_payment_received_mail',
            'text/ECOM_PAYMENT_SENT_MAIL.txt' => 'ecom_payment_sent_mail',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__currency() : object
    {
        return lorem_globalise(do_lorem_template('CURRENCY', [
            'AMOUNT' => placeholder_number(),
            'NEW_AMOUNT' => placeholder_number(),
            'FROM_CURRENCY' => 'GBP',
            'TO_CURRENCY' => 'GBP',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_logos_authorize() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_LOGOS_AUTHORIZE', [
            'CUSTOMER_ID' => placeholder_codename(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_payment_processor_links_authorize() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_PAYMENT_PROCESSOR_LINKS_AUTHORIZE', []), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_transaction_button_via_paypal() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_TRANSACTION_BUTTON_VIA_PAYPAL', [
            'TYPE_CODE' => lorem_word(),
            'ITEM_NAME' => lorem_phrase(),
            'PURCHASE_ID' => placeholder_codename(),
            'TRANS_EXPECTING_ID' => placeholder_numeric_id(),
            'PRICE' => placeholder_number(),
            'TAX' => placeholder_number(),
            'SHIPPING_COST' => placeholder_number(),
            'AMOUNT' => placeholder_number(),
            'CURRENCY' => 'GBP',
            'PAYMENT_ADDRESS' => lorem_word(),
            'FORM_URL' => placeholder_url(),
            'MEMBER_ADDRESS' => placeholder_array(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_subscription_button_via_paypal() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_SUBSCRIPTION_BUTTON_VIA_PAYPAL', [
            'TYPE_CODE' => lorem_phrase(),
            'ITEM_NAME' => lorem_phrase(),
            'PURCHASE_ID' => placeholder_codename(),
            'TRANS_EXPECTING_ID' => placeholder_numeric_id(),
            'LENGTH' => placeholder_number(),
            'LENGTH_UNITS' => lorem_phrase(),
            'PRICE' => placeholder_number(),
            'TAX' => placeholder_number(),
            'AMOUNT' => placeholder_number(),
            'CURRENCY' => 'GBP',
            'PAYMENT_ADDRESS' => lorem_word(),
            'FORM_URL' => placeholder_url(),
            'MEMBER_ADDRESS' => placeholder_array(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_subscription_cancel_button_via_paypal() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_SUBSCRIPTION_CANCEL_BUTTON_VIA_PAYPAL', [
            'PURCHASE_ID' => placeholder_codename(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_transaction_button_via_ccbill() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_TRANSACTION_BUTTON_VIA_CCBILL', [
            'TYPE_CODE' => lorem_word(),
            'ITEM_NAME' => lorem_phrase(),
            'PURCHASE_ID' => placeholder_codename(),
            'TRANS_EXPECTING_ID' => placeholder_numeric_id(),
            'PRICE' => placeholder_number(),
            'TAX' => placeholder_number(),
            'SHIPPING_COST' => placeholder_number(),
            'AMOUNT' => placeholder_number(),
            'CURRENCY' => 'GBP',
            'PAYMENT_ADDRESS' => lorem_paragraph(),
            'FORM_URL' => placeholder_url(),
            'MEMBER_ADDRESS' => placeholder_array(),
            'ACCOUNT_NUM' => placeholder_codename(),
            'SUBACCOUNT_NUM' => placeholder_codename(),
            'FORM_NAME' => placeholder_codename(),
            'FORM_PERIOD' => placeholder_codename(),
            'DIGEST' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_subscription_button_via_ccbill() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_SUBSCRIPTION_BUTTON_VIA_CCBILL', [
            'TYPE_CODE' => lorem_phrase(),
            'ITEM_NAME' => lorem_phrase(),
            'PURCHASE_ID' => placeholder_codename(),
            'TRANS_EXPECTING_ID' => placeholder_numeric_id(),
            'LENGTH' => placeholder_number(),
            'LENGTH_UNITS' => 'm',
            'PRICE' => placeholder_number(),
            'TAX' => placeholder_number(),
            'AMOUNT' => placeholder_number(),
            'CURRENCY' => 'GBP',
            'PAYMENT_ADDRESS' => lorem_paragraph(),
            'FORM_URL' => placeholder_url(),
            'MEMBER_ADDRESS' => placeholder_array(),
            'ACCOUNT_NUM' => placeholder_codename(),
            'SUBACCOUNT_NUM' => placeholder_codename(),
            'FORM_NAME' => placeholder_codename(),
            'FORM_PERIOD' => placeholder_codename(),
            'DIGEST' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_subscription_cancel_button_via_ccbill() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_SUBSCRIPTION_CANCEL_BUTTON_VIA_CCBILL', [
            'PURCHASE_ID' => placeholder_codename(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_transaction_button_via_authorize() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_TRANSACTION_BUTTON_VIA_AUTHORIZE', [
            'TYPE_CODE' => lorem_word(),
            'ITEM_NAME' => lorem_phrase(),
            'PURCHASE_ID' => placeholder_codename(),
            'TRANS_EXPECTING_ID' => placeholder_numeric_id(),
            'FORM_URL' => placeholder_url(),
            'MEMBER_ADDRESS' => placeholder_array(),
            'SEQUENCE' => placeholder_number(),
            'TIMESTAMP' => placeholder_date_raw(),
            'FINGERPRINT' => lorem_word(),
            'LOGIN_ID' => lorem_word(),
            'PRICE' => placeholder_number(),
            'TAX' => placeholder_number(),
            'SHIPPING_COST' => placeholder_number(),
            'AMOUNT' => placeholder_number(),
            'IS_TEST' => false,
            'CUST_ID' => placeholder_codename(),
            'CURRENCY' => 'GBP',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_subscription_button_via_authorize() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_SUBSCRIPTION_BUTTON_VIA_AUTHORIZE', [
            'TYPE_CODE' => lorem_phrase(),
            'ITEM_NAME' => lorem_phrase(),
            'PURCHASE_ID' => placeholder_codename(),
            'TRANS_EXPECTING_ID' => placeholder_numeric_id(),
            'FORM_URL' => placeholder_url(),
            'MEMBER_ADDRESS' => placeholder_array(),
            'SEQUENCE' => placeholder_number(),
            'TIMESTAMP' => placeholder_date_raw(),
            'FINGERPRINT' => lorem_word(),
            'LOGIN_ID' => lorem_word(),
            'PRICE' => placeholder_number(),
            'TAX' => placeholder_number(),
            'AMOUNT' => placeholder_number(),
            'IS_TEST' => false,
            'CUST_ID' => placeholder_codename(),
            'CURRENCY' => 'GBP',
            'LENGTH' => placeholder_number(),
            'LENGTH_UNITS' => 'm',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_subscription_cancel_button_via_authorize() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_SUBSCRIPTION_CANCEL_BUTTON_VIA_AUTHORIZE', [
            'CANCEL_URL' => placeholder_url(),
            'PURCHASE_ID' => placeholder_codename(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__purchase_screen() : object
    {
        $fields = do_lorem_template('ECOM_PURCHASE_STAGE_TRANSACT', [
            'TITLE' => lorem_screen_title(),
            'TYPE_CODE' => lorem_word(),
            'PURCHASE_ID' => placeholder_codename(),
            'FIELDS' => placeholder_fields(),
            'HIDDEN' => '',
            'LOGOS' => placeholder_image(),
            'PAYMENT_PROCESSOR_LINKS' => placeholder_link(),
            'ITEM_NAME' => lorem_phrase(),
        ]);

        return lorem_globalise(do_lorem_template('ECOM_PURCHASE_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'CONTENT' => $fields,
            'URL' => placeholder_url(),
            'SUBMIT_NAME' => lorem_phrase(),
            'ICON' => 'buttons/proceed',
            'GET' => false,
            'JS_FUNCTION_CALLS' => [],
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__purchase_stage_guest() : object
    {
        require_lang('cns');

        return lorem_globalise(do_lorem_template('ECOM_PURCHASE_STAGE_GUEST', [
            'TEXT' => lorem_sentence_html(),
            'FORM' => placeholder_form(),
            'HIDDEN' => '',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__purchase_stage_choose() : object
    {
        require_code('currency');

        $products = [];
        for ($i = 0; $i <= 5; $i++) {
            $map = [
                'ITEM_NAME' => lorem_phrase(),
                'DESCRIPTION' => lorem_paragraph(),
                'URL' => placeholder_url(),
                'IMAGE_URL' => placeholder_image_url(),
                'CAN_PURCHASE' => true,
                'IS_CATEGORY' => false,
                'NUM_PRODUCTS_IN_CATEGORY' => '',
                'NUM_PRODUCTS_IN_CATEGORY_AVAILABLE' => '',
            ];

            switch ($i) {
                case 0: // unavailable
                    $_discounted_price = do_lang('NA');
                    $written_price = do_lang_tempcode('NA_EM');

                    $map += [
                        'WRITTEN_PRICE' => $written_price,

                        'FULL_PRICE' => '0.00 GBP',
                        'DISCOUNTED_PRICE' => $_discounted_price,

                        '_FULL_PRICE' => '0.00',
                        '_DISCOUNTED_PRICE' => '',
                        '_CURRENCY' => 'GBP',
                    ];
                    break;

                case 1:
                    $_full_price = currency_convert(0.00, 'GBP', 'GBP', CURRENCY_DISPLAY_TEMPLATED);
                    $_discounted_price = do_lang('NA');
                    $written_price = do_lang_tempcode('ECOMMERCE_PRODUCT_PRICING_FOR_FREE');

                    $map += [
                        'WRITTEN_PRICE' => $written_price,

                        'FULL_PRICE' => $_full_price,
                        'DISCOUNTED_PRICE' => $_discounted_price,

                        '_FULL_PRICE' => '0.00',
                        '_DISCOUNTED_PRICE' => '',
                        '_CURRENCY' => 'GBP',
                        '_PRICE_POINTS' => null,
                        '_DISCOUNT_POINTS__NUM_POINTS' => null,
                        '_DISCOUNT_POINTS__PRICE_REDUCTION' => null,
                    ];
                    break;

                case 2:
                    $_full_price = currency_convert(4.56, 'GBP', 'GBP', CURRENCY_DISPLAY_TEMPLATED);
                    $_discounted_price = currency_convert(3.45, 'GBP', 'GBP', CURRENCY_DISPLAY_TEMPLATED);
                    $written_price = do_lang_tempcode('ECOMMERCE_PRODUCT_PRICING_WITH_DISCOUNT', $_discounted_price, $_full_price, [escape_html(integer_format(5, 0))]);

                    $map += [
                        'WRITTEN_PRICE' => $written_price,

                        'FULL_PRICE' => $_full_price,
                        'DISCOUNTED_PRICE' => $_discounted_price,

                        '_FULL_PRICE' => '4.56',
                        '_DISCOUNTED_PRICE' => '3.45',
                        '_CURRENCY' => 'GBP',
                        '_PRICE_POINTS' => null,
                        '_DISCOUNT_POINTS__NUM_POINTS' => '5',
                        '_DISCOUNT_POINTS__PRICE_REDUCTION' => '1.11',
                    ];
                    break;

                case 3:
                    $_full_price = currency_convert(1.23, 'GBP', 'GBP', CURRENCY_DISPLAY_TEMPLATED);
                    $_discounted_price = currency_convert(0.00, 'GBP', 'GBP', CURRENCY_DISPLAY_TEMPLATED);
                    $written_price = do_lang_tempcode('ECOMMERCE_PRODUCT_PRICING_FOR_FREE_WITH_POINTS', $_discounted_price, $_full_price, [escape_html(integer_format(5, 0))]);

                    $map += [
                        'WRITTEN_PRICE' => $written_price,

                        'FULL_PRICE' => $_full_price,
                        'DISCOUNTED_PRICE' => $_discounted_price,

                        '_FULL_PRICE' => '1.23',
                        '_DISCOUNTED_PRICE' => '0.00',
                        '_CURRENCY' => 'GBP',
                        '_PRICE_POINTS' => '5',
                        '_DISCOUNT_POINTS__NUM_POINTS' => null,
                        '_DISCOUNT_POINTS__PRICE_REDUCTION' => null,
                    ];
                    break;

                case 4:
                    $_full_price = currency_convert(1.23, 'GBP', 'GBP', CURRENCY_DISPLAY_TEMPLATED);
                    $_discounted_price = do_lang('NA');
                    $written_price = do_lang_tempcode('ECOMMERCE_PRODUCT_PRICING_FULL_PRICE', $_full_price);

                    $map += [
                        'WRITTEN_PRICE' => $written_price,

                        'FULL_PRICE' => $_full_price,
                        'DISCOUNTED_PRICE' => $_discounted_price,

                        '_FULL_PRICE' => '1.23',
                        '_DISCOUNTED_PRICE' => '',
                        '_CURRENCY' => 'GBP',
                        '_PRICE_POINTS' => null,
                        '_DISCOUNT_POINTS__NUM_POINTS' => null,
                        '_DISCOUNT_POINTS__PRICE_REDUCTION' => null,
                    ];
                    break;

                case 5: // category
                    $map['IS_CATEGORY'] = false;
                    $map['NUM_PRODUCTS_IN_CATEGORY'] = '1';
                    $map['NUM_PRODUCTS_IN_CATEGORY_AVAILABLE'] = '1';
                    break;
            }

            $products[] = $map;
        }

        return lorem_globalise(do_lorem_template('ECOM_PURCHASE_STAGE_CHOOSE', [
            'TITLE' => lorem_screen_title(),
            'PRODUCTS' => $products,

            'CATEGORY' => null,
            'MUST_SUPPORT_MONEY' => false,
            'MUST_SUPPORT_POINTS' => false,

            'POINTS_INVOLVED' => true,
            'MONEY_INVOLVED' => true,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__purchase_stage_message() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_PURCHASE_STAGE_MESSAGE', [
            'TITLE' => lorem_screen_title(),
            'ITEM_NAME' => lorem_phrase(),
            'TYPE_CODE' => lorem_word(),
            'TEXT' => lorem_sentence_html(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__purchase_stage_terms() : object
    {
        require_lang('installer');
        return lorem_globalise(do_lorem_template('ECOM_PURCHASE_STAGE_TERMS', [
            'TITLE' => lorem_screen_title(),
            'ITEM_NAME' => lorem_phrase(),
            'TYPE_CODE' => lorem_word(),
            'URL' => placeholder_url(),
            'TERMS' => lorem_paragraph(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__purchase_stage_details() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_PURCHASE_STAGE_DETAILS', [
            'TITLE' => lorem_screen_title(),
            'ITEM_NAME' => lorem_phrase(),
            'TYPE_CODE' => lorem_word(),
            'TEXT' => lorem_sentence_html(),
            'FIELDS' => placeholder_fields(),
            'HIDDEN' => '',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__purchase_stage_transact() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_PURCHASE_STAGE_TRANSACT', [
            'TITLE' => lorem_screen_title(),
            'ITEM_NAME' => lorem_phrase(),
            'TYPE_CODE' => lorem_word(),
            'PURCHASE_ID' => placeholder_codename(),
            'FIELDS' => placeholder_fields(),
            'HIDDEN' => '',
            'LOGOS' => placeholder_image(),
            'PAYMENT_PROCESSOR_LINKS' => placeholder_link(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__purchase_stage_pay() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_PURCHASE_STAGE_PAY', [
            'TITLE' => lorem_screen_title(),
            'TRANSACTION_BUTTON' => placeholder_button(),
            'CURRENCY' => 'GBP',
            'ITEM_NAME' => lorem_phrase(),
            'TYPE_CODE' => lorem_word(),
            'PURCHASE_ID' => placeholder_codename(),
            'LENGTH' => '3',
            'LENGTH_UNITS' => '$',
            'PRICE' => '123.45',
            'LOGOS' => placeholder_image(),
            'PAYMENT_PROCESSOR_LINKS' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__purchase_stage_finish() : object
    {
        return lorem_globalise(do_lorem_template('ECOM_PURCHASE_STAGE_FINISH', [
            'TITLE' => lorem_screen_title(),
            'TYPE_CODE' => lorem_word(),
            'MESSAGE' => lorem_phrase(),
            'SUCCESS' => true,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_invoices_screen() : object
    {
        $invoices = [];
        foreach (placeholder_array() as $k => $v) {
            $invoices[] = [
                'TRANSACTION_BUTTON' => placeholder_button(),
                'INVOICE_TITLE' => lorem_phrase(),
                'INVOICE_ID' => placeholder_numeric_id(),
                'PRICE' => placeholder_number(),
                'TAX' => placeholder_number(),
                'CURRENCY' => 'GBP',
                'DATE' => placeholder_date(),
                'STATE' => lorem_word(),
                'FULFILLABLE' => true,
                'PAYABLE' => lorem_word(),
                'NOTE' => lorem_phrase(),
                'TYPE_CODE' => lorem_word(),
            ];
        }

        return lorem_globalise(do_lorem_template('ECOM_INVOICES_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'INVOICES' => $invoices,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_subscriptions_screen() : object
    {
        $button = do_lorem_template('ECOM_TRANSACTION_LOGS_MANUAL_TRIGGER', [
            'STATUS' => lorem_phrase(),
            'TRIGGER_URL' => placeholder_url(),
        ]);

        $subscriptions = [];
        foreach (placeholder_array() as $k => $v) {
            $subscriptions[] = [
                'ITEM_NAME' => lorem_phrase(),
                'SUBSCRIPTION_ID' => placeholder_codename(),
                'PER' => lorem_word(),
                'LENGTH' => placeholder_number(),
                'LENGTH_UNITS' => 'm',
                'PRICE' => placeholder_number(),
                'TAX' => placeholder_number(),
                'TOTAL' => placeholder_number(),
                'CURRENCY' => 'GBP',
                '_START_TIME' => placeholder_date_raw(),
                '_TERM_START_TIME' => placeholder_date_raw(),
                '_TERM_END_TIME' => placeholder_date_raw(),
                '_EXPIRY_TIME' => placeholder_date_raw(),
                'START_TIME' => placeholder_date(),
                'TERM_START_TIME' => placeholder_date(),
                'TERM_END_TIME' => placeholder_date(),
                'EXPIRY_TIME' => placeholder_date(),
                'STATE' => lorem_word(),
                '_STATE' => lorem_word(),
                'TYPE_CODE' => lorem_word(),
                'CANCEL_BUTTON' => $button,
                'USERGROUP_SUBSCRIPTION_TITLE' => lorem_phrase(),
                'USERGROUP_SUBSCRIPTION_DESCRIPTION' => lorem_paragraph_html(),
                'USERGROUP_SUBSCRIPTION_ID' => placeholder_numeric_id(),
                'USERGROUP_ID' => placeholder_numeric_id(),
                'USERGROUP_NAME' => lorem_phrase(),
                '_PAYMENT_GATEWAY' => 'manual',
                'PAYMENT_GATEWAY' => lorem_word(),
            ];
        }

        return lorem_globalise(do_lorem_template('ECOM_SUBSCRIPTIONS_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'SUBSCRIPTIONS' => $subscriptions,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__member_subscription_status_screen() : object
    {
        $button = do_lorem_template('ECOM_TRANSACTION_LOGS_MANUAL_TRIGGER', [
            'STATUS' => lorem_phrase(),
            'TRIGGER_URL' => placeholder_url(),
        ]);

        $subscriptions = [];
        foreach (placeholder_array() as $k => $v) {
            $subscriptions[] = [
                'ITEM_NAME' => lorem_phrase(),
                'SUBSCRIPTION_ID' => placeholder_codename(),
                'PER' => lorem_word(),
                'LENGTH' => placeholder_number(),
                'LENGTH_UNITS' => 'm',
                'PRICE' => placeholder_number(),
                'TAX' => placeholder_number(),
                'TOTAL' => placeholder_number(),
                'CURRENCY' => 'GBP',
                '_START_TIME' => placeholder_date_raw(),
                '_TERM_START_TIME' => placeholder_date_raw(),
                '_TERM_END_TIME' => placeholder_date_raw(),
                '_EXPIRY_TIME' => placeholder_date_raw(),
                'START_TIME' => placeholder_date(),
                'TERM_START_TIME' => placeholder_date(),
                'TERM_END_TIME' => placeholder_date(),
                'EXPIRY_TIME' => placeholder_date(),
                'STATE' => lorem_word(),
                '_STATE' => lorem_word(),
                'TYPE_CODE' => lorem_word(),
                'CANCEL_BUTTON' => $button,
                'USERGROUP_SUBSCRIPTION_TITLE' => lorem_phrase(),
                'USERGROUP_SUBSCRIPTION_DESCRIPTION' => lorem_paragraph_html(),
                'USERGROUP_SUBSCRIPTION_ID' => placeholder_numeric_id(),
                'USERGROUP_ID' => placeholder_numeric_id(),
                'USERGROUP_NAME' => lorem_phrase(),
                '_PAYMENT_GATEWAY' => 'manual',
                'PAYMENT_GATEWAY' => lorem_word(),
            ];
        }

        return lorem_globalise(do_lorem_template('ECOM_MEMBER_SUBSCRIPTION_STATUS', [
            'MEMBER_ID' => placeholder_numeric_id(),
            'SUBSCRIPTIONS' => $subscriptions,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__ecom_sales_screen() : object
    {
        $cells = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $cells->attach(do_lorem_template('RESULTS_TABLE_FIELD', ['VALUE' => lorem_word()], null, false, 'RESULTS_TABLE_FIELD'));
        }
        $header_row = do_lorem_template('RESULTS_TABLE_ENTRY', ['VALUES' => $cells], null, false, 'RESULTS_TABLE_ENTRY');

        $out = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $cells = new Tempcode();
            foreach (placeholder_array() as $_k => $_v) {
                $cells->attach(do_lorem_template('COLUMNED_TABLE_ROW_CELL', ['VALUE' => $_v]));
            }

            $out->attach(do_lorem_template('COLUMNED_TABLE_ROW', ['CELLS' => $cells]));
        }

        $content = do_lorem_template('COLUMNED_TABLE', ['HEADER_ROW' => $header_row, 'ROWS' => $out, 'NONRESPONSIVE' => false]);

        require_code('form_templates');

        push_field_encapsulation(FIELD_ENCAPSULATION_RAW);

        $filters_row_a = [
            [
                'PARAM' => 'placeholder_a',
                'LABEL' => lorem_word(),
                'FIELD' => do_lorem_template('FORM_SCREEN_INPUT_LINE', [
                    '_GUID' => '02789c9af25cbc971e86bfcc0ad322d5',
                    'PLACEHOLDER' => null,
                    'MAXLENGTH' => strval(16),
                    'TABINDEX' => placeholder_number(),
                    'REQUIRED' => false,
                    'NAME' => 'placeholder_a',
                    'DEFAULT' => '',
                    'TYPE' => 'text',
                    'PATTERN' => null,
                    'SIZE' => strval(16),
                    'AUTOCOMPLETE' => null,
                ]),
            ],
        ];
        $filters_row_b = [
            [
                'PARAM' => 'placeholder_b',
                'LABEL' => lorem_word(),
                'FIELD' => do_lorem_template('FORM_SCREEN_INPUT_DATE', [
                    '_GUID' => '5ace58dd0f540f70fb3bd440fb02a430',
                    'REQUIRED' => false,
                    'TABINDEX' => placeholder_number(),
                    'NAME' => 'placeholder_b',
                    'TYPE' => 'datetime',

                    'YEAR' => '',
                    'MONTH' => '',
                    'DAY' => '',
                    'HOUR' => '',
                    'MINUTE' => '',

                    'MIN_DATE_DAY' => '',
                    'MIN_DATE_MONTH' => '',
                    'MIN_DATE_YEAR' => '',
                    'MAX_DATE_DAY' => '',
                    'MAX_DATE_MONTH' => '',
                    'MAX_DATE_YEAR' => '',

                    'AUTOCOMPLETE' => null,
                    'READ_ONLY' => strval(0)
                ]),
            ],
        ];

        pop_field_encapsulation();

        return lorem_globalise(do_lorem_template('ECOM_SALES_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'CONTENT' => $content,
            'PAGINATION' => placeholder_pagination(),
            'FILTERS_ROW_A' => $filters_row_a,
            'FILTERS_ROW_B' => $filters_row_b,
            'URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__cns_member_profile_ecommerce_reports_screen() : object
    {
        return lorem_globalise(do_lorem_template('CNS_MEMBER_ECOMMERCE_REPORTS', [
            'CONTENT' => lorem_chunk_html(),
            'PAGINATION' => placeholder_pagination(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__ecom_cash_flow_screen() : object
    {
        $types = [];
        foreach (placeholder_array() as $v) {
            $types[] = [
                'TYPE' => lorem_word(),
                'AMOUNT' => placeholder_number(),
                'SPECIAL' => placeholder_number(),
            ];
        }
        return lorem_globalise(do_lorem_template('ECOM_CASH_FLOW_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'TYPES' => $types,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__ecom_products_price_screen() : object
    {
        // This is for getting the do_ajax_request() javascript function.

        $warning_details = do_lorem_template('WARNING_BOX', ['WARNING' => lorem_phrase()]);

        $add_forms = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $add_forms->attach(do_lorem_template('ECOM_PRODUCTS_PRICES_FORM_WRAP', ['TITLE' => lorem_phrase(), 'FORM' => placeholder_form()]));
        }

        return lorem_globalise(do_lorem_template('ECOM_PRODUCT_PRICE_SCREEN', [
            'PING_URL' => placeholder_url(),
            'WARNING_DETAILS' => $warning_details,
            'TITLE' => lorem_screen_title(),
            'EDIT_FORM' => placeholder_form(),
            'ADD_FORMS' => $add_forms,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_product_forwarder_mail() : object
    {
        return do_lorem_template('ECOM_PRODUCT_FORWARDER_MAIL', [
            'ENCODED_REASON' => lorem_phrase(),
            'EMAIL' => lorem_word(),
            'PREFIX' => lorem_phrase(),
            'SUFFIX' => lorem_phrase(),
            'FORW_URL' => placeholder_url(),
        ], null, false, null, '.txt', 'text');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_product_pop3_mail() : object
    {
        return do_lorem_template('ECOM_PRODUCT_POP3_MAIL', [
            'EMAIL' => lorem_word(),
            'ENCODED_REASON' => lorem_phrase(),
            'LOGIN' => lorem_phrase(),
            '_QUOTA' => placeholder_number(),
            'QUOTA' => placeholder_number(),
            'MAIL_SERVER' => lorem_phrase(),
            'PASSWORD' => lorem_phrase(),
            'PREFIX' => lorem_phrase(),
            'SUFFIX' => lorem_phrase(),
            'POP3_URL' => placeholder_url(),
        ], null, false, null, '.txt', 'text');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_product_quota_mail() : object
    {
        return do_lorem_template('ECOM_PRODUCT_QUOTA_MAIL', [
            'ENCODED_REASON' => lorem_phrase(),
            '_QUOTA' => placeholder_number(),
            'QUOTA' => placeholder_number(),
            'EMAIL' => lorem_word(),
            'QUOTA_URL' => placeholder_url(),
        ], null, false, null, '.txt', 'text');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_payment_received_mail() : object
    {
        return do_lorem_template('ECOM_PAYMENT_RECEIVED_MAIL', [
            'AUTOMATIC_SETUP' => false,
            'AMOUNT' => placeholder_number(),
            'CURRENCY' => 'USD',
            'ITEM_NAME' => lorem_phrase(),
            'TAX' => placeholder_number(),
            'CURRENCY_SYMBOL' => '$',
            'RECEIPT' => lorem_chunk_html(),
            'MEMO' => lorem_chunk(),
            'SHIPPED' => true,

            // Shipping
            'SHIPPING_FIRSTNAME' => lorem_phrase(),
            'SHIPPING_LASTNAME' => lorem_phrase(),
            'SHIPPING_STREET_ADDRESS' => lorem_phrase(),
            'SHIPPING_CITY' => lorem_phrase(),
            'SHIPPING_COUNTY' => lorem_phrase(),
            'SHIPPING_STATE' => lorem_phrase(),
            'SHIPPING_POST_CODE' => lorem_phrase(),
            'SHIPPING_COUNTRY' => lorem_phrase(),
            'SHIPPING_EMAIL' => lorem_phrase(),
            'SHIPPING_PHONE' => lorem_phrase(),
            'SHIPPING_FORMATTED_ADDRESS' => lorem_phrase(),
        ], null, false, null, '.txt', 'text');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_payment_sent_mail() : object
    {
        return do_lorem_template('ECOM_PAYMENT_SENT_MAIL', [
            'AUTOMATIC_SETUP' => false,
            'AMOUNT' => placeholder_number(),
            'CURRENCY' => 'USD',
            'ITEM_NAME' => lorem_phrase(),
            'TAX' => placeholder_number(),
            'CURRENCY_SYMBOL' => '$',
            'RECEIPT' => lorem_chunk_html(),
            'MEMO' => lorem_chunk(),
            'SHIPPED' => true,

            // Shipping
            'SHIPPING_FIRSTNAME' => lorem_phrase(),
            'SHIPPING_LASTNAME' => lorem_phrase(),
            'SHIPPING_STREET_ADDRESS' => lorem_phrase(),
            'SHIPPING_CITY' => lorem_phrase(),
            'SHIPPING_COUNTY' => lorem_phrase(),
            'SHIPPING_STATE' => lorem_phrase(),
            'SHIPPING_POST_CODE' => lorem_phrase(),
            'SHIPPING_COUNTRY' => lorem_phrase(),
            'SHIPPING_EMAIL' => lorem_phrase(),
            'SHIPPING_PHONE' => lorem_phrase(),
            'SHIPPING_FORMATTED_ADDRESS' => lorem_phrase(),
        ], null, false, null, '.txt', 'text');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__ecom_invoice_or_receipt_screen() : object
    {
        $items = [];
        foreach (placeholder_array() as $k => $v) {
            $items[] = [
                'IDENTIFIER' => placeholder_number(),
                'TYPE_CODE' => $v,
                'ITEM_NAME' => lorem_phrase(),
                'QUANTITY' => placeholder_number(),
                'UNIT_PRICE' => placeholder_number(),
                'PRICE' => placeholder_number(),
                'TAX' => placeholder_number(),
                'TAX_RATE' => placeholder_number(),
            ];
        }

        return lorem_globalise(do_lorem_template('ECOM_INVOICE_OR_RECEIPT', [
            'INVOICE_ID' => placeholder_number(),
            'TXN_ID' => placeholder_number(),
            '_DATE' => placeholder_date_raw(),
            'DATE' => placeholder_date(),
            'TRANS_ADDRESS' => lorem_phrase() . "\n" . lorem_phrase() . "\n" . lorem_phrase(),
            'ITEMS' => $items,
            'STATUS' => lorem_phrase(),
            'CURRENCY' => 'GBP',
            'SUBTOTAL' => placeholder_number(),
            'TOTAL_TAX' => placeholder_number(),
            'TOTAL_AMOUNT' => placeholder_number(),
            'PURCHASE_ID' => placeholder_codename(),
        ]), null, '', true);
    }
}
