<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    filedump
 */

/**
 * Hook class.
 */
class Hook_addon_registry_filedump
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        $chmod = [];
        if ($runtime) {
            $chmod = array_merge(
                $chmod,
                [
                    'uploads/filedump/*',
                ]
            );
        }
        $chmod = array_merge(
            $chmod,
            [
                'uploads/filedump',
            ]
        );
        return $chmod;
    }

    /**
     * Get the version of Composr this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return null;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Admin Utilities';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'File/media library, for use in attachments or for general ad-hoc sharing.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_collaboration',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/cms/filedump.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'cms/pages/modules/filedump.php',
            'lang/EN/filedump.ini',
            'sources/filedump.php',
            'sources/hooks/blocks/side_stats/filedump.php',
            'sources/hooks/modules/admin_import_types/filedump.php',
            'sources/hooks/modules/search/filedump.php',
            'sources/hooks/systems/actionlog/filedump.php',
            'sources/hooks/systems/addon_registry/filedump.php',
            'sources/hooks/systems/ajax_tree/choose_filedump_file.php',
            'sources/hooks/systems/commandr_fs/filedump.php',
            'sources/hooks/systems/commandr_fs/home.php',
            'sources/hooks/systems/config/filedump_show_stats_count_total_files.php',
            'sources/hooks/systems/config/filedump_show_stats_count_total_space.php',
            'sources/hooks/systems/config/search_filedump.php',
            'sources/hooks/systems/notifications/filedump.php',
            'sources/hooks/systems/page_groupings/filedump.php',
            'sources/hooks/systems/privacy/filedump.php',
            'sources/hooks/systems/rss/filedump.php',
            'themes/default/css/filedump.css',
            'themes/default/images/icons/menu/cms/filedump.svg',
            'themes/default/images/icons_monochrome/menu/cms/filedump.svg',
            'themes/default/javascript/filedump.js',
            'themes/default/templates/FILEDUMP_EMBED_SCREEN.tpl',
            'themes/default/templates/FILEDUMP_FOOTER.tpl',
            'themes/default/templates/FILEDUMP_SCREEN.tpl',
            'themes/default/templates/FILEDUMP_SEARCH.tpl',
            'uploads/filedump/.htaccess',
            'uploads/filedump/index.html',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/FILEDUMP_SCREEN.tpl' => 'filedump_screen',
            'templates/FILEDUMP_FOOTER.tpl' => 'filedump_screen',
            'templates/FILEDUMP_SEARCH.tpl' => 'filedump_screen',
            'templates/FILEDUMP_EMBED_SCREEN.tpl' => 'filedump_embed_screen',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__filedump_screen() : object
    {
        require_css('forms');

        $thumbnails = [];
        $thumbnails[] = [
            'FILENAME' => lorem_word(),
            'SUBPATH' => placeholder_codename(),
            'IMAGE_URL' => placeholder_image_url(),
            'IS_IMAGE' => true,
            'URL' => placeholder_url(),
            'DESCRIPTION' => lorem_paragraph(),
            'ACTIONS' => lorem_paragraph(),
            '_SIZE' => placeholder_number(),
            'SIZE' => placeholder_number(),
            '_TIME' => placeholder_date_raw(),
            'DATE' => placeholder_date(),
            'WIDTH' => placeholder_number(),
            'HEIGHT' => placeholder_number(),
            'IS_DIRECTORY' => false,
            'CHOOSABLE' => false,
            'EMBED_URL' => placeholder_url(),
        ];

        return lorem_globalise(do_lorem_template('FILEDUMP_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'SUBPATH' => placeholder_codename(),
            'THUMBNAILS' => $thumbnails,
            'LISTING' => placeholder_table(),
            'UPLOAD_FORM' => placeholder_form(),
            'CREATE_FOLDER_FORM' => placeholder_form(),
            'TYPE_FILTER' => '',
            'SEARCH' => '',
            'SORT' => 'time ASC',
            'PAGINATION_LISTING' => placeholder_pagination(),
            'PAGINATION_THUMBNAILS' => placeholder_pagination(),
            'POST_URL' => placeholder_url(),
            'DIRECTORIES' => [lorem_word()],
            'OTHER_DIRECTORIES' => [lorem_word()],
            'FILTERED_DIRECTORIES' => [lorem_word()],
            'FILTERED_DIRECTORIES_MISSES' => [],
            'SOMETHING_EDITABLE' => true,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__filedump_embed_screen() : object
    {
        $image_sizes = [];
        foreach (placeholder_array() as $k => $v) {
            $image_sizes[] = [
                'LABEL' => $v,
                'SIZE_URL' => placeholder_image_url(),
                'SIZE_WIDTH' => placeholder_number() . strval($k),
                'SIZE_HEIGHT' => placeholder_number() . strval($k),
            ];
        }

        return lorem_globalise(do_lorem_template('FILEDUMP_EMBED_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'FORM' => placeholder_form(),
            'IMAGE_SIZES' => $image_sizes,
            'URL' => placeholder_image_url(),
            'EXISTING_COUNT' => placeholder_number(),
            'DESCRIPTION' => lorem_phrase(),
            'IMAGE' => true,
        ]), null, '', true);
    }
}
