<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    forum_blocks
 */

/**
 * Hook class.
 */
class Hook_addon_registry_forum_blocks
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of Composr this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return null;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Information Display';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Blocks to draw forum posts and topics into the main website.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_featured',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [
                'news_shared',
            ],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/admin/component.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'sources/blocks/bottom_forum_news.php',
            'sources/blocks/main_forum_news.php',
            'sources/blocks/main_forum_topics.php',
            'sources/blocks/side_forum_news.php',
            'sources/hooks/modules/admin_setupwizard/forum_blocks.php',
            'sources/hooks/systems/addon_registry/forum_blocks.php',
            'themes/default/templates/BLOCK_MAIN_FORUM_NEWS.tpl',
            'themes/default/templates/BLOCK_MAIN_FORUM_TOPICS.tpl',
            'themes/default/templates/BLOCK_SIDE_FORUM_NEWS.tpl',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/BLOCK_MAIN_FORUM_TOPICS.tpl' => 'block_main_forum_topics',
            'templates/BLOCK_SIDE_FORUM_NEWS.tpl' => 'block_side_forum_news',
            'templates/BLOCK_MAIN_FORUM_NEWS.tpl' => 'block_main_forum_news',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_forum_topics() : object
    {
        require_lang('cns');

        $topics = [];
        foreach (placeholder_array() as $k => $v) {
            $topics[] = [
                'POST' => lorem_paragraph(),
                'FORUM_ID' => null,
                'FORUM_NAME' => lorem_phrase(),
                'TOPIC_URL' => placeholder_url(),
                'TITLE' => lorem_phrase(),
                'DATE' => placeholder_date(),
                'DATE_RAW' => placeholder_date_raw(),
                'USERNAME' => lorem_word(),
                'MEMBER_ID' => null,
                '_NUM_POSTS' => placeholder_number(),
                'NUM_POSTS' => placeholder_number(),
            ];
        }

        return lorem_globalise(do_lorem_template('BLOCK_MAIN_FORUM_TOPICS', [
            'BLOCK_ID' => lorem_word(),
            'TITLE' => lorem_phrase(),
            'TOPICS' => $topics,
            'FORUM_NAME' => lorem_phrase(),
            'SUBMIT_URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_side_forum_news() : object
    {
        require_lang('news');
        require_lang('cns');

        $news = [];
        foreach (placeholder_array() as $k => $v) {
            $news[] = [
                'REPLIES' => placeholder_number(),
                'FIRSTTIME' => placeholder_date_raw(),
                'LASTTIME' => placeholder_date_raw(),
                'CLOSED' => false,
                'FIRSTUSERNAME' => lorem_word(),
                'LASTUSERNAME' => lorem_word(),
                'FIRSTMEMBERID' => placeholder_numeric_id(),
                'LASTMEMBERID' => placeholder_numeric_id(),
                '_DATE' => placeholder_date_raw(),
                'DATE' => placeholder_date(),
                'FULL_URL' => placeholder_url(),
                'NEWS_TITLE' => lorem_phrase_html(),
            ];
        }

        return lorem_globalise(do_lorem_template('BLOCK_SIDE_FORUM_NEWS', [
            'BLOCK_ID' => lorem_word(),
            'FORUM_NAME' => lorem_phrase(),
            'TITLE' => lorem_phrase_html(),
            'NEWS' => $news,
            'SUBMIT_URL' => placeholder_url(),
            'ARCHIVE_URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_forum_news() : object
    {
        require_lang('news');

        $out = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $out->attach(do_lorem_template('NEWS_BOX', [
                'TRUNCATE' => false,
                'BLOG' => false,
                'ID' => placeholder_random_id(),
                'FULL_URL' => placeholder_url(),
                'SUBMITTER' => placeholder_numeric_id(),
                'DATE' => placeholder_date(),
                'DATE_RAW' => placeholder_date_raw(),
                'NEWS_TITLE' => lorem_word(),
                'NEWS_TITLE_PLAIN' => lorem_word(),
                'CATEGORY' => '',
                '_CATEGORY' => placeholder_numeric_id(),
                'CATEGORY_URL' => placeholder_url(),
                'REP_IMAGE_URL' => placeholder_image_url(),
                'AUTHOR' => lorem_word(),
                'AUTHOR_URL' => placeholder_url(),
                'NEWS' => lorem_paragraph_html(),
                'GIVE_CONTEXT' => false,
                'FORUM_ID' => null,
                '_AUTHOR' => null,
                'AVATAR' => null,
                'COMMENTS' => null,
                'VIEW' => null,
                'COMMENT_COUNT' => null,
                'READ_MORE' => null,
                'FIRSTTIME' => placeholder_date(),
                'LASTTIME' => placeholder_date(),
                'CLOSED' => false,
                'FIRSTUSERNAME' => lorem_word(),
                'LASTUSERNAME' => lorem_word(),
                'FIRSTMEMBERID' => placeholder_numeric_id(),
                'LASTMEMBERID' => placeholder_numeric_id(),
                'BRIEF_CONTENT' => lorem_sentence_html(),
            ]));
        }

        return lorem_globalise(do_lorem_template('BLOCK_MAIN_FORUM_NEWS', [
            'BLOCK_ID' => lorem_word(),
            'TITLE' => lorem_phrase_html(),
            'FORUM_NAME' => lorem_phrase(),
            'CONTENT' => $out,
            'ARCHIVE_URL' => placeholder_url(),
            'SUBMIT_URL' => placeholder_url(),
            'RSS_URL' => placeholder_url(),
            'ATOM_URL' => placeholder_url(),
            'BRIEF_CONTENT' => lorem_sentence_html(),
        ]), null, '', true);
    }
}
