<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    import
 */

/**
 * Hook class.
 */
class Hook_addon_registry_import
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of Composr this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return null;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Admin Utilities';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Switch to Composr from other software. This addon provides the architecture for importing, and a number of prewritten importers.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_importer',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/admin/import.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_import.php',
            'lang/EN/import.ini',
            'sources/hooks/modules/admin_import/.htaccess',
            'sources/hooks/modules/admin_import/cms_merge.php',
            'sources/hooks/modules/admin_import/html_site.php',
            'sources/hooks/modules/admin_import/index.html',
            'sources/hooks/modules/admin_import/ipb2.php',
            'sources/hooks/modules/admin_import/mybb.php',
            'sources/hooks/modules/admin_import/phpbb3.php',
            'sources/hooks/modules/admin_import/smf2.php',
            'sources/hooks/modules/admin_import/vb3.php',
            'sources/hooks/modules/admin_import/wordpress.php',
            'sources/hooks/modules/admin_import_types/.htaccess',
            'sources/hooks/modules/admin_import_types/core.php',
            'sources/hooks/modules/admin_import_types/index.html',
            'sources/hooks/systems/actionlog/import.php',
            'sources/hooks/systems/addon_registry/import.php',
            'sources/hooks/systems/cns_auth/wordpress.php',
            'sources/hooks/systems/commandr_commands/continue_import.php',
            'sources/hooks/systems/page_groupings/import.php',
            'sources/import.php',
            'sources_custom/hooks/modules/admin_import/.htaccess',
            'sources_custom/hooks/modules/admin_import/index.html',
            'sources_custom/hooks/modules/admin_import_types/.htaccess',
            'sources_custom/hooks/modules/admin_import_types/index.html',
            'themes/default/css/importing.css',
            'themes/default/templates/IMPORT_ACTION_LINE.tpl',
            'themes/default/templates/IMPORT_ACTION_SCREEN.tpl',
            'themes/default/templates/IMPORT_MESSAGE.tpl',
            'themes/default/text/IMPORT_MKPORTAL_FCOMCODEPAGE.txt',
            'themes/default/text/IMPORT_PHPNUKE_FCOMCODEPAGE.txt',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/IMPORT_ACTION_LINE.tpl' => 'administrative__import_action_screen',
            'templates/IMPORT_ACTION_SCREEN.tpl' => 'administrative__import_action_screen',
            'templates/IMPORT_MESSAGE.tpl' => 'administrative__import_action_screen',
            'text/IMPORT_MKPORTAL_FCOMCODEPAGE.txt' => 'administrative__import_mkportal_fcomcodepage',
            'text/IMPORT_PHPNUKE_FCOMCODEPAGE.txt' => 'administrative__import_phpnuke_fcomcodepage',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__import_action_screen() : object
    {
        $extra = do_lorem_template('IMPORT_MESSAGE', [
            'MESSAGE' => lorem_phrase(),
        ]);
        $import_list = do_lorem_template('IMPORT_ACTION_LINE', [
            'CHECKED' => false,
            'DISABLED' => true,
            'NAME' => lorem_word(),
            'TEXT' => lorem_phrase(),
            'ADVANCED_URL' => placeholder_url(),
        ]);

        return lorem_globalise(do_lorem_template('IMPORT_ACTION_SCREEN', [
            'EXTRA' => $extra,
            'MESSAGE' => lorem_phrase(),
            'TITLE' => lorem_screen_title(),
            'FIELDS' => lorem_phrase(),
            'HIDDEN' => '',
            'IMPORTER' => lorem_phrase(),
            'IMPORT_LIST' => $import_list,
            'URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__import_mkportal_fcomcodepage() : object
    {
        return lorem_globalise(do_lorem_template('IMPORT_MKPORTAL_FCOMCODEPAGE', [
            'TITLE' => lorem_phrase(),
            'SUBTITLE' => lorem_phrase(),
            'PAGE_HEADER' => lorem_phrase(),
            'TEXT' => lorem_sentence_html(),
            'PAGE_FOOTER' => lorem_phrase(),
            'SIGNATURE' => lorem_phrase(),
        ], null, false, null, '.txt', 'text'), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__import_phpnuke_fcomcodepage() : object
    {
        return lorem_globalise(do_lorem_template('IMPORT_PHPNUKE_FCOMCODEPAGE', [
            'TITLE' => lorem_phrase(),
            'SUBTITLE' => lorem_phrase(),
            'PAGE_HEADER' => lorem_phrase(),
            'TEXT' => lorem_sentence_html(),
            'PAGE_FOOTER' => lorem_phrase(),
            'SIGNATURE' => lorem_phrase(),
        ], null, false, null, '.txt', 'text'), null, '', true);
    }
}
