<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    ldap
 */

/**
 * Hook class.
 */
class Hook_addon_registry_ldap
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of Composr this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return null;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return is_maintained('ldap') ? 'Third Party Integration' : 'Development';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Support for integrating Conversr with an LDAP server, so usergroup and members can be the same as those already on the network';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_ldap',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/adminzone/security/ldap.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_cns_ldap.php',
            'sources/cns_ldap.php',
            'sources/hooks/systems/actionlog/ldap.php',
            'sources/hooks/systems/addon_registry/ldap.php',
            'sources/hooks/systems/config/ldap_allow_joining.php',
            'sources/hooks/systems/config/ldap_base_dn.php',
            'sources/hooks/systems/config/ldap_bind_password.php',
            'sources/hooks/systems/config/ldap_bind_rdn.php',
            'sources/hooks/systems/config/ldap_group_class.php',
            'sources/hooks/systems/config/ldap_group_search_qualifier.php',
            'sources/hooks/systems/config/ldap_hostname.php',
            'sources/hooks/systems/config/ldap_is_enabled.php',
            'sources/hooks/systems/config/ldap_is_windows.php',
            'sources/hooks/systems/config/ldap_login_qualifier.php',
            'sources/hooks/systems/config/ldap_member_class.php',
            'sources/hooks/systems/config/ldap_member_property.php',
            'sources/hooks/systems/config/ldap_member_search_qualifier.php',
            'sources/hooks/systems/config/ldap_none_bind_logins.php',
            'sources/hooks/systems/config/ldap_version.php',
            'sources/hooks/systems/page_groupings/ldap.php',
            'themes/default/images/icons/menu/adminzone/security/ldap.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/security/ldap.svg',
            'themes/default/templates/CNS_LDAP_LIST_ENTRY.tpl',
            'themes/default/templates/CNS_LDAP_SYNC_SCREEN.tpl',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/CNS_LDAP_LIST_ENTRY.tpl' => 'administrative__cns_ldap_sync_screen',
            'templates/CNS_LDAP_SYNC_SCREEN.tpl' => 'administrative__cns_ldap_sync_screen',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__cns_ldap_sync_screen() : object
    {
        require_lang('cns');

        $members_delete = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $tpl = do_lorem_template('CNS_LDAP_LIST_ENTRY', [
                'NAME' => placeholder_random_id(),
                'NICE_NAME' => lorem_word(),
            ]);
            $members_delete->attach($tpl);
        }

        $groups_delete = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $tpl = do_lorem_template('CNS_LDAP_LIST_ENTRY', [
                'NAME' => placeholder_random_id(),
                'NICE_NAME' => lorem_word(),
            ]);
            $groups_delete->attach($tpl);
        }

        $groups_add = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $tpl = do_lorem_template('CNS_LDAP_LIST_ENTRY', [
                'NAME' => placeholder_random_id(),
                'NICE_NAME' => lorem_word(),
            ]);
            $groups_add->attach($tpl);
        }

        return lorem_globalise(do_lorem_template('CNS_LDAP_SYNC_SCREEN', [
            'URL' => placeholder_url(),
            'TITLE' => lorem_screen_title(),
            'MEMBERS_DELETE' => $members_delete,
            'GROUPS_DELETE' => $groups_delete,
            'GROUPS_ADD' => $groups_add,
        ]), null, '', true);
    }
}
