<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    setupwizard
 */

/**
 * Hook class.
 */
class Hook_addon_registry_setupwizard
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of Composr this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return null;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Admin Utilities';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Quick-start setup wizard.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_configuration',
            'tut_drinking',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
            'previously_in_addon' => ['core_setupwizard'],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/adminzone/setup/setupwizard.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_setupwizard.php',
            'sources/hooks/blocks/main_staff_checklist/default_content.php',
            'sources/hooks/modules/admin_setupwizard/.htaccess',
            'sources/hooks/modules/admin_setupwizard/core.php',
            'sources/hooks/modules/admin_setupwizard/index.html',
            'sources/hooks/modules/admin_setupwizard_installprofiles/.htaccess',
            'sources/hooks/modules/admin_setupwizard_installprofiles/community.php',
            'sources/hooks/modules/admin_setupwizard_installprofiles/index.html',
            'sources/hooks/modules/admin_setupwizard_installprofiles/infosite.php',
            'sources/hooks/modules/admin_setupwizard_installprofiles/minimalistic.php',
            'sources/hooks/systems/actionlog/setupwizard.php',
            'sources/hooks/systems/addon_registry/setupwizard.php',
            'sources/hooks/systems/page_groupings/setupwizard.php',
            'sources/hooks/systems/preview/setupwizard.php',
            'sources/hooks/systems/preview/setupwizard_blocks.php',
            'sources/setupwizard.php',
            'sources_custom/hooks/modules/admin_setupwizard/.htaccess',
            'sources_custom/hooks/modules/admin_setupwizard/index.html',
            'sources_custom/hooks/modules/admin_setupwizard_installprofiles/.htaccess',
            'sources_custom/hooks/modules/admin_setupwizard_installprofiles/index.html',
            'themes/default/css/setupwizard.css',
            'themes/default/images/icons/menu/adminzone/setup/setupwizard.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/setup/setupwizard.svg',
            'themes/default/javascript/setupwizard.js',
            'themes/default/templates/SETUPWIZARD_2.tpl',
            'themes/default/templates/SETUPWIZARD_7.tpl',
            'themes/default/templates/SETUPWIZARD_BLOCK_PREVIEW.tpl',
            'themes/default/templates/SETUPWIZARD_SCREEN.tpl',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/SETUPWIZARD_SCREEN.tpl' => 'administrative__setupwizard_2_screen',
            'templates/SETUPWIZARD_2.tpl' => 'administrative__setupwizard_2_screen',
            'templates/SETUPWIZARD_7.tpl' => 'administrative__setupwizard_7_screen',
            'templates/SETUPWIZARD_BLOCK_PREVIEW.tpl' => 'administrative__setupwizard_block_preview',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__setupwizard_block_preview() : object
    {
        return lorem_globalise(do_lorem_template('SETUPWIZARD_BLOCK_PREVIEW', [
            'LEFT' => lorem_paragraph_html(),
            'RIGHT' => lorem_paragraph_html(),
            'START' => lorem_paragraph_html(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__setupwizard_2_screen() : object
    {
        require_lang('config');

        $inner = do_lorem_template('SETUPWIZARD_2', [
            'SKIP_WEBSTANDARDS' => true,
            'URL' => placeholder_url(),
            'SUBMIT_ICON' => 'buttons/proceed',
            'SUBMIT_NAME' => lorem_word(),
        ]);

        return lorem_globalise(do_lorem_template('SETUPWIZARD_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'INNER' => $inner,
            'STEP' => '2',
            'NUM_STEPS_ENUMERABLE' => '10',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__setupwizard_7_screen() : object
    {
        require_lang('config');

        $inner = do_lorem_template('SETUPWIZARD_7', [
            'FORM' => placeholder_form(),
            'BALANCED' => lorem_phrase(),
            'LIBERAL' => lorem_phrase(),
            'CORPORATE' => lorem_phrase(),
        ]);

        return lorem_globalise(do_lorem_template('SETUPWIZARD_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'INNER' => $inner,
            'STEP' => '7',
            'NUM_STEPS_ENUMERABLE' => '10',
        ]), null, '', true);
    }
}
