<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    filedump
 */

/**
 * Hook class.
 */
class Hook_ajax_tree_choose_filedump_file
{
    /**
     * Run function for ajax-tree hooks. Generates XML for a tree list, which is interpreted by JavaScript and expanded on-demand (via new calls).
     *
     * @param  ?ID_TEXT $id The ID to do under (null: root)
     * @param  array $options Options being passed through
     * @param  ?ID_TEXT $default The ID to select by default (null: none)
     * @return string XML in the special category,entry format
     */
    public function run(?string $id, array $options, ?string $default = null) : string
    {
        if ($id === null) {
            $id = '';
        }

        require_code('files2');
        require_code('images');
        $full_path = get_custom_file_base() . '/uploads/filedump';
        if ($id != '') {
            $full_path .= '/' . $id;
        }

        $levels_to_expand = array_key_exists('levels_to_expand', $options) ? ($options['levels_to_expand']) : intval(get_value('levels_to_expand__' . substr(get_class($this), 5), null, true));
        $options['levels_to_expand'] = max(0, $levels_to_expand - 1);

        $images_only = array_key_exists('images_only', $options) ? intval($options['images_only']) : null;

        $folder = ((isset($options['folder'])) && ($options['folder'])); // We want to select folders, not files

        $out = '';

        $out .= '<options>' . xmlentities(json_encode($options)) . '</options>' . "\n";

        if ((has_actual_page_access(null, 'filedump')) && (file_exists($full_path))) {
            $files = get_directory_contents($full_path, '', IGNORE_ACCESS_CONTROLLERS, false);
            cms_mb_sort($files, SORT_NATURAL | SORT_FLAG_CASE);
            foreach ($files as $f) {
                $description = $GLOBALS['SITE_DB']->query_select_value_if_there('filedump', 'the_description', ['name' => basename($f), 'subpath' => $id . '/']);

                $entry_id = cms_rawurlrecode('uploads/filedump/' . (($id == '') ? '' : (str_replace('%2F', '/', rawurlencode($id)) . '/')) . str_replace('%2F', '/', rawurlencode($f)));

                if (is_dir($full_path . '/' . $f)) {
                    $has_children = (!empty(get_directory_contents($full_path . '/' . $f, '', IGNORE_ACCESS_CONTROLLERS, false)));

                    if ($has_children) {
                        $out .= '<category id="' . xmlentities((($id == '') ? '' : ($id . '/')) . $f) . '" title="' . xmlentities($f) . '" has_children="' . ($has_children ? 'true' : 'false') . '" selectable="' . ($folder ? 'true' : 'false') . '"></category>' . "\n";

                        if ($levels_to_expand > 0) {
                            $out .= '<expand>' . xmlentities((($id == '') ? '' : ($id . '/')) . $f) . '</expand>' . "\n";
                        }
                    }
                } elseif (!$folder) {
                    if (($images_only !== null) && (!is_image($f, $images_only, true))) {
                        continue;
                    }

                    if ((!isset($options['only_images'])) || (!$options['only_images']) || (is_image($f, IMAGE_CRITERIA_WEBSAFE, true))) {
                        if (($description === null) || (get_translated_text($description) == '')) {
                            $_description = '';
                            if (is_image($f, IMAGE_CRITERIA_WEBSAFE | IMAGE_CRITERIA_GD_READ, true)) {
                                $url = get_custom_base_url() . '/' . $entry_id;
                                $_description = trim(static_evaluate_tempcode(do_image_thumb($url, '', true, false, null, null, true)));
                            }
                        } else {
                            $_description = escape_html(get_translated_text($description));
                        }
                        $out .= '<entry id="' . xmlentities($entry_id) . '" title="' . xmlentities($f) . '" description_html="' . xmlentities($_description) . '" selectable="true"></entry>' . "\n";
                    }
                }
            }

            // Mark parent cats for pre-expansion
            if (!cms_empty_safe($default)) {
                $cat = '';
                foreach (explode('/', $default) as $_cat) {
                    if ($_cat != '') {
                        $cat .= '/';
                        $cat .= $_cat;
                    }
                    $out .= '<expand>' . xmlentities($cat) . '</expand>' . "\n";
                }
            }
        }

        return '<result>' . $out . '</result>';
    }

    /**
     * Generate a simple selection list for the ajax-tree hook. Returns a normal <select> style <option>-list, for fallback purposes.
     *
     * @param  ?ID_TEXT $id The ID to do under (null: root) - not always supported
     * @param  array $options Options being passed through
     * @param  ?ID_TEXT $it The ID to select by default (null: none)
     * @return Tempcode The nice list
     */
    public function simple(?string $id, array $options, ?string $it = null) : object
    {
        $out = new Tempcode();

        if (has_actual_page_access(null, 'filedump')) {
            require_code('filedump');
            $only_images = isset($options['only_images']) && ($options['only_images']);
            $out = nice_get_filedump_files($it, $only_images, $id);
        }

        return $out;
    }
}
