<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    newsletter
 */

require_code('resource_fs');

/**
 * Hook class.
 */
class Hook_commandr_fs_newsletter_subscribers extends Resource_fs_base
{
    public $file_resource_type = 'newsletter_subscriber';

    /**
     * Standard Commandr-fs function for seeing how many resources are. Useful for determining whether to do a full rebuild.
     *
     * @param  ID_TEXT $resource_type The resource type
     * @return integer How many resources there are
     */
    public function get_resources_count(string $resource_type) : int
    {
        return $GLOBALS['SITE_DB']->query_select_value('newsletter_subscribers', 'COUNT(*)');
    }

    /**
     * Standard Commandr-fs function for searching for a resource by label.
     *
     * @param  ID_TEXT $resource_type The resource type
     * @param  LONG_TEXT $label The resource label
     * @return array A list of resource IDs
     */
    public function find_resource_by_label(string $resource_type, string $label) : array
    {
        $_ret = $GLOBALS['SITE_DB']->query_select('newsletter_subscribers', ['id'], ['email' => $label]);
        $ret = [];
        foreach ($_ret as $r) {
            $ret[] = strval($r['id']);
        }
        return $ret;
    }

    /**
     * Whether the filesystem hook is active.
     *
     * @return boolean Whether it is
     */
    public function is_active() : bool
    {
        return addon_installed('newsletter');
    }

    /**
     * Standard Commandr-fs date fetch function for resource-fs hooks. Defined when getting an edit date is not easy.
     *
     * @param  array $row Resource row (not full, but does contain the ID)
     * @param  ID_TEXT $category Parent category (blank: root / not applicable)
     * @return ?TIME The edit date or add date, whichever is higher (null: could not find one)
     */
    protected function _get_file_edit_date(array $row, string $category = '') : ?int
    {
        //return $GLOBALS['SITE_DB']->query_select_value_if_there('newsletter_subscribers', 'MAX(join_time)');
        return null; // The above can't handle unsubscribes
    }

    /**
     * Standard Commandr-fs add function for resource-fs hooks. Adds some resource with the given label and properties.
     *
     * @param  LONG_TEXT $filename Filename OR Resource label
     * @param  string $path The path (blank: root / not applicable)
     * @param  array $properties Properties (may be empty, properties given are open to interpretation by the hook but generally correspond to database fields)
     * @param  ?ID_TEXT $force_type Resource type to try to force (null: do not force)
     * @return ~ID_TEXT The resource ID (false: error, could not create via these properties / here)
     */
    public function file_add(string $filename, string $path, array $properties, ?string $force_type = null)
    {
        list($properties, $label) = $this->_file_magic_filter($filename, $path, $properties, $this->file_resource_type);

        require_code('newsletter');
        require_code('newsletter2');

        $email = $label;
        $join_time = $this->_default_property_time($properties, 'join_time');
        $code_confirm = $this->_default_property_int_null($properties, 'code_confirm');
        if ($code_confirm === null) {
            $code_confirm = 0;
        }
        $password = $this->_default_property_str($properties, 'password');
        $salt = $this->_default_property_str($properties, 'salt');
        $language = $this->_default_property_str($properties, 'language');
        if ($language == '') {
            $language = get_site_default_lang();
        }
        $forename = $this->_default_property_str($properties, 'forename');
        $surname = $this->_default_property_str($properties, 'surname');

        $id = add_newsletter_subscriber($email, $join_time, $code_confirm, $password, $salt, $language, $forename, $surname);

        $this->_resource_save_extend($this->file_resource_type, strval($id), $filename, $label, $properties);

        return strval($id);
    }

    /**
     * Standard Commandr-fs load function for resource-fs hooks. Finds the properties for some resource.
     *
     * @param  SHORT_TEXT $filename Filename
     * @param  string $path The path (blank: root / not applicable). It may be a wildcarded path, as the path is used for content-type identification only. Filenames are globally unique across a hook; you can calculate the path using ->search.
     * @return ~array Details of the resource (false: error)
     */
    public function file_load(string $filename, string $path)
    {
        list($resource_type, $resource_id) = $this->file_convert_filename_to_id($filename);

        $rows = $GLOBALS['SITE_DB']->query_select('newsletter_subscribers', ['*'], ['id' => intval($resource_id)], '', 1);
        if (!array_key_exists(0, $rows)) {
            return false;
        }
        $row = $rows[0];

        $properties = [
            'label' => $row['email'],
            'join_time' => $row['join_time'],
            'code_confirm' => $row['code_confirm'],
            'password' => $row['the_password'],
            'salt' => $row['pass_salt'],
            'language' => $row['language'],
            'forename' => $row['n_forename'],
            'surname' => $row['n_surname'],
        ];
        $this->_resource_load_extend($resource_type, $resource_id, $properties, $filename, $path);
        return $properties;
    }

    /**
     * Standard Commandr-fs edit function for resource-fs hooks. Edits the resource to the given properties.
     *
     * @param  ID_TEXT $filename The filename
     * @param  string $path The path (blank: root / not applicable)
     * @param  array $properties Properties (may be empty, properties given are open to interpretation by the hook but generally correspond to database fields)
     * @param  boolean $explicit_move Whether we are definitely moving (as opposed to possible having it in multiple positions)
     * @return ~ID_TEXT The resource ID (false: error, could not create via these properties / here)
     */
    public function file_edit(string $filename, string $path, array $properties, bool $explicit_move = false)
    {
        list($resource_type, $resource_id) = $this->file_convert_filename_to_id($filename);
        list($properties,) = $this->_file_magic_filter($filename, $path, $properties, $this->file_resource_type);

        if ($resource_id === null) {
            return false;
        }

        require_code('newsletter');
        require_code('newsletter2');

        $email = $this->_default_property_str($properties, 'label');
        $join_time = $this->_default_property_time($properties, 'join_time');
        $code_confirm = $this->_default_property_int_null($properties, 'code_confirm');
        if ($code_confirm === null) {
            $code_confirm = 0;
        }
        $password = $this->_default_property_str($properties, 'password');
        $salt = $this->_default_property_str($properties, 'salt');
        $language = $this->_default_property_str($properties, 'language');
        if ($language == '') {
            $language = get_site_default_lang();
        }
        $forename = $this->_default_property_str($properties, 'forename');
        $surname = $this->_default_property_str($properties, 'surname');

        edit_newsletter_subscriber(intval($resource_id), $email, $join_time, $code_confirm, $password, $salt, $language, $forename, $surname);

        $this->_resource_save_extend($this->file_resource_type, $resource_id, $filename, $email, $properties);

        return $resource_id;
    }

    /**
     * Standard Commandr-fs delete function for resource-fs hooks. Deletes the resource.
     *
     * @param  ID_TEXT $filename The filename
     * @param  string $path The path (blank: root / not applicable)
     * @return boolean Success status
     */
    public function file_delete(string $filename, string $path) : bool
    {
        list($resource_type, $resource_id) = $this->file_convert_filename_to_id($filename);

        if ($resource_id === null) {
            return false;
        }

        require_code('newsletter');

        delete_newsletter_subscriber(intval($resource_id));

        return true;
    }
}
