<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    quizzes
 */

/**
 * Hook class.
 */
class Hook_content_meta_aware_quiz extends Hook_CMA
{
    /**
     * Get content type details.
     *
     * @param  ?ID_TEXT $zone The zone to link through to (null: autodetect)
     * @param  boolean $get_extended_data Populate additional data that is somewhat costly to compute (add_url, archive_url)
     * @return ?array Map of content-type info (null: disabled)
     */
    public function info(?string $zone = null, bool $get_extended_data = false) : ?array
    {
        if (!addon_installed('quizzes')) {
            return null;
        }

        return [
            'support_custom_fields' => true,

            'content_type_label' => 'quiz:QUIZ',
            'content_type_universal_label' => 'Quiz',
            'content_type_label_override' => 'CALL: generate_quiz_content_type_label',
            'content_type_universal_label_override' => 'CALL: generate_quiz_content_type_universal_label',

            'db' => $GLOBALS['SITE_DB'],
            'table' => 'quizzes',
            'id_field' => 'id',
            'id_field_numeric' => true,
            'parent_category_field' => null,
            'parent_category_meta_aware_type' => null,
            'is_category' => true,
            'is_entry' => true,
            'category_field' => 'q_type', // For category permissions
            'permission_module' => 'quiz', // For category permissions
            'parent_spec__table_name' => null,
            'parent_spec__parent_name' => null,
            'parent_spec__field_name' => null,
            'category_is_string' => true,

            'title_field' => 'q_name',
            'title_field_dereference' => true,
            'description_field' => 'q_start_text',
            'description_field_dereference' => true,
            'description_field_supports_comcode' => true,
            'image_field' => null,
            'image_field_is_theme_image' => false,
            'alternate_icon_theme_image' => 'icons/menu/rich_content/quiz',

            'view_page_link_pattern' => '_SEARCH:quiz:do:_WILD',
            'edit_page_link_pattern' => '_SEARCH:cms_quiz:_edit:_WILD',
            'view_category_page_link_pattern' => null,
            'add_url' => ($get_extended_data && has_submit_permission('high', get_member(), get_ip_address(), 'cms_quiz')) ? (get_module_zone('cms_quiz') . ':cms_quiz:add') : null,
            'archive_url' => $get_extended_data ? ((($zone !== null) ? $zone : get_module_zone('quiz')) . ':quiz') : null,

            'support_url_monikers' => true,

            'views_field' => null,
            'order_field' => null,
            'submitter_field' => 'q_submitter',
            'author_field' => null,
            'add_time_field' => 'q_add_date',
            'edit_time_field' => null,
            'date_field' => 'q_add_date',
            'validated_field' => 'q_validated',

            'seo_type_code' => null,

            'feedback_type_code' => null,

            'search_hook' => 'quiz',
            'rss_hook' => null,
            'attachment_hook' => null,
            'notification_hook' => null,
            'sitemap_hook' => 'quiz',

            'addon_name' => 'quizzes',

            'cms_page' => 'cms_quiz',
            'module' => 'quiz',

            'commandr_filesystem_hook' => 'quizzes',
            'commandr_filesystem__is_folder' => false,

            'support_revisions' => false,

            'support_privacy' => false,

            'support_content_reviews' => true,

            'support_spam_heuristics' => null,

            'actionlog_regexp' => '\w+_QUIZ',

            'default_prominence_weight' => PROMINENCE_WEIGHT_MEDIUM,
            'default_prominence_flags' => 0,
        ];
    }

    /**
     * Get headings of special relevant data this content type supports.
     *
     * @return array A map of heading codenames to Tempcode labels
     */
    public function get_special_keymap_headings() : array
    {
        $headings = [];

        $headings['type'] = do_lang_tempcode('TYPE');

        return $headings;
    }

    /**
     * Get special relevant data this content type supports.
     *
     * @param  array $row Database row
     * @return array A map of heading codenames to Tempcode values
     */
    public function get_special_keymap(array $row) : array
    {
        $keymap = [];

        $keymap['type'] = do_lang_tempcode($row['q_type']);

        return $keymap;
    }

    /**
     * Render a content box for a content row.
     *
     * @param  array $row The database row for the content
     * @param  ID_TEXT $zone The zone to display in
     * @param  boolean $give_context Whether to include context (i.e. say WHAT this is, not just show the actual content)
     * @param  boolean $include_breadcrumbs Whether to include breadcrumbs (if there are any)
     * @param  ?ID_TEXT $root Virtual root to use (null: none)
     * @param  boolean $attach_to_url_filter Whether to copy through any filter parameters in the URL, under the basis that they are associated with what this box is browsing
     * @param  ID_TEXT $guid Overridden GUID to send to templates (blank: none)
     * @return Tempcode Results
     */
    public function render_box(array $row, string $zone, bool $give_context = true, bool $include_breadcrumbs = true, ?string $root = null, bool $attach_to_url_filter = false, string $guid = '') : object
    {
        require_code('quiz');

        return render_quiz_box($row, $zone, $give_context, $guid);
    }
}

/**
 * Find an entry content-type language string label.
 *
 * @param  array $row Database row of entry
 * @return Tempcode Label
 */
function generate_quiz_content_type_label(array $row) : object
{
    if (!array_key_exists('q_type', $row)) {
        return do_lang_tempcode('quiz:QUIZ');
    }
    return do_lang_tempcode('quiz:' . $row['q_type']);
}

/**
 * Find an entry content-type universal label (doesn't depend on language pack).
 *
 * @param  array $row Database row of entry
 * @return string Label
 */
function generate_quiz_content_type_universal_label(array $row) : string
{
    $type = 'Quiz';
    if (!array_key_exists('q_type', $row)) {
        return $type;
    }
    switch ($row['q_type']) {
        case 'COMPETITION':
            $type = 'Competition';
            break;

        case 'SURVEY':
            $type = 'Survey';
            break;

        case 'TEST':
            $type = 'Test';
            break;
    }
    return $type;
}
