<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    galleries
 */

/**
 * Hook class.
 */
class Hook_content_meta_aware_video extends Hook_CMA
{
    /**
     * Get content type details.
     *
     * @param  ?ID_TEXT $zone The zone to link through to (null: autodetect)
     * @param  boolean $get_extended_data Populate additional data that is somewhat costly to compute (add_url, archive_url)
     * @return ?array Map of content-type info (null: disabled)
     */
    public function info(?string $zone = null, bool $get_extended_data = false) : ?array
    {
        if (!addon_installed('galleries')) {
            return null;
        }

        return [
            'support_custom_fields' => true,

            'content_type_label' => ((function_exists('get_option')) && (get_option('allow_audio_videos') == '2')) ? 'galleries:MULTIMEDIA_FILE' : 'global:VIDEO',
            'content_type_universal_label' => ((function_exists('get_option')) && (get_option('allow_audio_videos') == '2')) ? 'Multimedia file' : 'Video',

            'db' => $GLOBALS['SITE_DB'],
            'extra_where_sql' => 'cat NOT LIKE \'' . db_encode_like('download\_%') . '\'',
            'table' => 'videos',
            'id_field' => 'id',
            'id_field_numeric' => true,
            'parent_category_field' => 'cat',
            'parent_category_meta_aware_type' => 'gallery',
            'is_category' => false,
            'is_entry' => true,
            'category_field' => 'cat', // For category permissions
            'permission_module' => 'galleries', // For category permissions
            'parent_spec__table_name' => 'galleries',
            'parent_spec__parent_name' => 'parent_id',
            'parent_spec__field_name' => 'name',
            'category_is_string' => true,

            'title_field' => 'title',
            'title_field_dereference' => true,
            'description_field' => 'the_description',
            'description_field_dereference' => true,
            'description_field_supports_comcode' => true,
            'image_field' => 'thumb_url',
            'image_field_is_theme_image' => false,
            'alternate_icon_theme_image' => null,

            'video_field' => ['url', 'video_width', 'video_height', 'submitter'],
            'video_generator' => 'generate_video_entry_video_details',

            'view_page_link_pattern' => '_SEARCH:galleries:video:_WILD',
            'edit_page_link_pattern' => '_SEARCH:cms_galleries:_edit_other:_WILD',
            'view_category_page_link_pattern' => '_SEARCH:galleries:browse:_WILD',
            'add_url' => ($get_extended_data && has_submit_permission('mid', get_member(), get_ip_address(), 'cms_galleries')) ? (get_module_zone('cms_galleries') . ':cms_galleries:add_other') : null,
            'archive_url' => $get_extended_data ? ((($zone !== null) ? $zone : get_module_zone('galleries')) . ':galleries') : null,

            'support_url_monikers' => true,

            'views_field' => 'video_views',
            'order_field' => null,
            'submitter_field' => 'submitter',
            'author_field' => null,
            'add_time_field' => 'add_date',
            'edit_time_field' => 'edit_date',
            'date_field' => 'add_date',
            'validated_field' => 'validated',

            'seo_type_code' => 'video',

            'feedback_type_code' => 'videos',

            'search_hook' => 'videos',
            'rss_hook' => 'galleries',
            'attachment_hook' => null,
            'notification_hook' => 'gallery_entry',
            'sitemap_hook' => 'video',

            'addon_name' => 'galleries',

            'cms_page' => 'cms_galleries',
            'module' => 'galleries',

            'commandr_filesystem_hook' => 'galleries',
            'commandr_filesystem__is_folder' => false,

            'support_revisions' => false,

            'support_privacy' => true,

            'support_content_reviews' => true,

            'support_spam_heuristics' => 'the_description',

            'actionlog_regexp' => '\w+_VIDEO',

            'default_prominence_weight' => PROMINENCE_WEIGHT_NONE,
            'default_prominence_flags' => 0,
        ];
    }

    /**
     * Get content title of a content row.
     *
     * @param  array $row The database row for the content
     * @param  integer $render_type A FIELD_RENDER_* constant
     * @param  boolean $falled_back_to_id Whether this has had to fall back to an ID due to missing title (returned by reference)
     * @param  boolean $resource_fs_style Whether to use the content API as resource-fs requires (may be slightly different)
     * @return mixed Content title (string or Tempcode, depending on $render_type)
     */
    public function get_title(array $row, int $render_type = 1, bool &$falled_back_to_id = false, bool $resource_fs_style = false)
    {
        $falled_back_to_id = false;
        $ret = parent::get_title($row, $render_type, $falled_back_to_id, $resource_fs_style);
        if ($falled_back_to_id) {
            require_lang('galleries');
            $fullname = $GLOBALS['SITE_DB']->query_select_value_if_there('galleries', 'fullname', ['name' => $row['cat']]);
            if ($fullname !== null) {
                $ret = do_lang('VIEW_VIDEO_IN', get_translated_text($fullname));
            }
        }
        return $ret;
    }

    /**
     * Render a content box for a content row.
     *
     * @param  array $row The database row for the content
     * @param  ID_TEXT $zone The zone to display in
     * @param  boolean $give_context Whether to include context (i.e. say WHAT this is, not just show the actual content)
     * @param  boolean $include_breadcrumbs Whether to include breadcrumbs (if there are any)
     * @param  ?ID_TEXT $root Virtual root to use (null: none)
     * @param  boolean $attach_to_url_filter Whether to copy through any filter parameters in the URL, under the basis that they are associated with what this box is browsing
     * @param  ID_TEXT $guid Overridden GUID to send to templates (blank: none)
     * @return Tempcode Results
     */
    public function render_box(array $row, string $zone, bool $give_context = true, bool $include_breadcrumbs = true, ?string $root = null, bool $attach_to_url_filter = false, string $guid = '') : object
    {
        require_code('galleries');

        return render_video_box($row, $zone, $give_context, $include_breadcrumbs, ($root === null) ? null : $root, $guid);
    }

    /**
     * Get the hook name of an AJAX tree selection list.
     *
     * @return ?string Hook name (null: none)
     */
    public function create_selection_tree_list() : ?string
    {
        return 'choose_video';
    }
}

/**
 * Find an entry video details.
 *
 * @param  array $row Database row of entry
 * @return array A tuple: Video URL, Video width, Video height, Video mime-type
 */
function generate_video_entry_video_details(array $row) : array
{
    $url = $row['url'];

    $extension = get_file_extension($url);
    require_code('mime_types');
    $mime_type = get_mime_type($extension, has_privilege($row['submitter'], 'comcode_dangerous'));

    $width = $row['video_width'];
    $height = $row['video_height'];

    return [$url, $width, $height, $mime_type];
}
