<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    welcome_emails
 */

/**
 * Hook class.
 */
class Hook_cron_cns_welcome_emails
{
    protected $member_sets_to_send_to;
    protected $time_now;

    protected const INITIAL_BACK_TIME = 24 * 60 * 60 * 7; // Don't send for really old members, 7 day initial window is reasonable

    /**
     * Get info from this hook.
     *
     * @param  ?TIME $last_run Last time run (null: never)
     * @param  ?boolean $calculate_num_queued Calculate the number of items queued, if possible (null: the hook may decide / low priority)
     * @return ?array Return a map of info about the hook (null: disabled)
     */
    public function info(?int $last_run, ?bool $calculate_num_queued) : ?array
    {
        if (!addon_installed('welcome_emails')) {
            return null;
        }

        if (get_forum_type() != 'cns') {
            return null;
        }

        // Calculate on low priority
        if ($calculate_num_queued === null) {
            $calculate_num_queued = true;
        }

        if ($calculate_num_queued) {
            require_code('cns_welcome_emails');

            $num_queued = 0;

            $this->time_now = time();

            if ($last_run === null) {
                $last_run = $this->time_now - self::INITIAL_BACK_TIME;
            }

            $this->member_sets_to_send_to = [];

            $mails = $GLOBALS['SITE_DB']->query_select('f_welcome_emails', ['*']);
            foreach ($mails as $mail) {
                $send_seconds_after_joining = $mail['w_send_time'] * 60 * 60;

                $members = [];

                $newsletter_style = false;

                // By newsletter
                if (($mail['w_newsletter'] !== null) && (addon_installed('newsletter'))) {
                    $newsletter_style = true;

                    // Think of it like this, m_join_time (members join time) must between $last_run and $this->time_now, but offset back by $send_seconds_after_joining
                    $where = ' WHERE join_time>' . strval($last_run - $send_seconds_after_joining) . ' AND join_time<=' . strval($this->time_now - $send_seconds_after_joining) . ' AND (the_level=3 OR the_level=4) AND newsletter_id=' . strval($mail['w_newsletter']);
                    $members = array_merge($members, $GLOBALS['SITE_DB']->query('SELECT s.email AS m_email_address,the_password,n_forename,n_surname,n.id,join_time AS m_join_time FROM ' . get_table_prefix() . 'newsletter_subscribe s JOIN ' . get_table_prefix() . 'newsletter_subscribers n ON n.email=s.email ' . $where . ' GROUP BY s.email'));
                } elseif (($mail['w_usergroup'] !== null) && (get_forum_type() == 'cns')) { // By usergroup
                    $where = ' WHERE join_time>' . strval($last_run - $send_seconds_after_joining) . ' AND join_time<=' . strval($this->time_now - $send_seconds_after_joining) . ' AND um.usergroup_id=' . strval($mail['w_usergroup']) . ' AND ' . db_string_not_equal_to('m_email_address', '');
                    $query = 'SELECT m.id as id, m.m_email_address AS m_email_address,m.m_username AS m_username,um.join_time AS m_join_time FROM ' . $GLOBALS['FORUM_DB']->get_table_prefix() . 'f_group_join_log as um JOIN ' . $GLOBALS['FORUM_DB']->get_table_prefix() . 'f_members as m ON m.id=um.member_id ' . $where;
                    $_members = $GLOBALS['FORUM_DB']->query($query);
                    foreach ($_members as $member) {
                        $ok = false;
                        switch ($mail['w_usergroup_type']) {
                            case '':
                                $ok = in_array($mail['w_usergroup'], $GLOBALS['FORUM_DRIVER']->get_members_groups($member['id'])); // If member still in the group
                                break;
                            case 'primary':
                                $ok = ($GLOBALS['FORUM_DRIVER']->get_member_row_field($member['id'], 'm_primary_group') == $mail['w_usergroup']); // If member still in the group
                                break;
                            case 'secondary':
                                $ok = ($GLOBALS['FORUM_DB']->query_select_value_if_there('f_group_members', 'gm_member_id', ['gm_group_id' => $mail['w_usergroup'], 'gm_member_id' => $member['id']]) !== null);
                                break;
                        }
                        if ($ok) {
                            $members[] = $member;
                        }
                    }
                } elseif (($mail['w_newsletter'] === null) && ($mail['w_usergroup'] === null)) { // By general membership
                    // Think of it like this, m_join_time (members join time) must between $last_run and $this->time_now, but offset back by $send_seconds_after_joining
                    $where = ' WHERE m_join_time>' . strval($last_run - $send_seconds_after_joining);
                    $where .= ' AND m_join_time<=' . strval($this->time_now - $send_seconds_after_joining);
                    $where .= ' AND ' . db_string_not_equal_to('m_email_address', '');
                    if (get_option('staff_email_receipt_configurability') != '0') {
                        $where .= ' AND m_allow_emails=1';
                    }
                    if ($send_seconds_after_joining != 0) {
                        $where .= ' AND ' . db_string_equal_to('m_validated_email_confirm_code', '');
                        if (addon_installed('validation')) {
                            $where .= ' AND m_validated=1';
                        }
                    }
                    $query = 'SELECT m_email_address,m_username,id,m_join_time FROM ' . $GLOBALS['FORUM_DB']->get_table_prefix() . 'f_members' . $where;
                    $members = array_merge($members, $GLOBALS['FORUM_DB']->query($query));
                }

                $this->member_sets_to_send_to[] = [$members, $newsletter_style, $mail];
                $num_queued += count($members);
            }
        } else {
            $num_queued = null;
        }

        return [
            'label' => 'Send welcome e-mails',
            'num_queued' => $num_queued,
            'minutes_between_runs' => 60,
        ];
    }

    /**
     * Run function for system scheduler hooks. Searches for things to do. ->info(..., true) must be called before this method.
     *
     * @param  ?TIME $last_run Last time run (null: never)
     */
    public function run(?int $last_run)
    {
        require_code('mail');
        require_code('cns_welcome_emails');

        foreach ($this->member_sets_to_send_to as $member_set) {
            list($members, $newsletter_style, $mail) = $member_set;

            foreach ($members as $member) {
                list($subject, $message, $is_html, $name) = cns_prepare_welcome_email($mail, $member);

                if (get_value('notification_safety_testing') === '1') {
                    $test = $GLOBALS['SITE_DB']->query_select_value_if_there('logged_mail_messages', 'm_date_and_time', ['m_subject' => $subject, 'm_to_email' => serialize([$member['m_email_address']])]);
                    if ($test !== null) {
                        if ($test > $member['m_join_time']) {
                            fatal_exit(do_lang('INTERNAL_ERROR') . ' [' . $member['m_email_address'] . ']');
                        }
                        // otherwise they probably just resubscribed and hence bumped their time
                    }
                }

                dispatch_mail($subject, $message, [$member['m_email_address']], $name, '', '', ['as_admin' => true, 'in_html' => $is_html]);
            }
        }
    }
}
