<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    ecommerce
 */

/**
 * Hook class.
 */
class Hook_cron_manual_subscription_notification
{
    /**
     * Get info from this hook.
     *
     * @param  ?TIME $last_run Last time run (null: never)
     * @param  ?boolean $calculate_num_queued Calculate the number of items queued, if possible (null: the hook may decide / low priority)
     * @return ?array Return a map of info about the hook (null: disabled)
     */
    public function info(?int $last_run, ?bool $calculate_num_queued) : ?array
    {
        if (!addon_installed('ecommerce')) {
            return null;
        }

        if (get_option('manual_subscription_expiry_notice') == '') {
            return null;
        }

        return [
            'label' => 'Send subscription expiry notifications (for manual subscriptions)',
            'num_queued' => null, // Too time-consuming to calculate
            'minutes_between_runs' => 60 * 24,
        ];
    }

    /**
     * Run function for system scheduler hooks. Searches for things to do. ->info(..., true) must be called before this method.
     *
     * @param  ?TIME $last_run Last time run (null: never)
     */
    public function run(?int $last_run)
    {
        /*
        Send staff notifications for expiring manual notifications.
        This might be used by the staff in order to get someone to send in a cheque, for example.
        */

        $manual_subscription_expiry_notice = intval(get_option('manual_subscription_expiry_notice'));

        require_lang('ecommerce');

        $max = 1000;
        $start = 0;
        do {
            $subscribers = $GLOBALS['SITE_DB']->query_select('ecom_subscriptions', ['DISTINCT s_member_id'], ['s_state' => 'active'], '', $max, $start);
            foreach ($subscribers as $subscriber) {
                $member_id = $subscriber['s_member_id'];

                require_code('ecommerce_subscriptions');
                $subscriptions = find_member_subscriptions($member_id);
                foreach ($subscriptions as $subscription) {
                    $expiry_time = $subscription['expiry_time'];
                    if (($expiry_time !== null) && (($expiry_time - time()) < ($manual_subscription_expiry_notice * 24 * 60 * 60)) && ($expiry_time >= time())) {
                        if ($last_run !== null) {
                            if (($expiry_time - $last_run) < ($manual_subscription_expiry_notice * 24 * 60 * 60)) {
                                continue; // Notification already sent!
                            }
                        }

                        if (($expiry_time - time()) < ($manual_subscription_expiry_notice * 24 * 60 * 60)) {
                            $expiry_date = get_timezoned_date_time($expiry_time, false);
                            $member_name = $GLOBALS['FORUM_DRIVER']->get_username($member_id, false, USERNAME_DEFAULT_NULL);
                            if ($member_name !== null) { // If not a deleted member
                                $member_profile_url = $GLOBALS['CNS_DRIVER']->member_profile_url($member_id, false);
                                $cancel_url = build_url(['page' => 'admin_ecommerce_reports', 'type' => 'cancel_subscription', 'subscription_id' => $subscription['subscription_id']], get_module_zone('admin_ecommerce'), [], false, false, true);

                                $item_name = $subscription['item_name'];

                                require_code('notifications');
                                $subject = do_lang('MANUAL_SUBSCRIPTION_NOTIFICATION_MAIL_SUBJECT', $member_name, $expiry_date, [$item_name]);
                                $mail = do_notification_lang('MANUAL_SUBSCRIPTION_NOTIFICATION_MAIL', comcode_escape($member_profile_url), comcode_escape($cancel_url->evaluate()), [strval($manual_subscription_expiry_notice), comcode_escape($member_name), comcode_escape($expiry_date), comcode_escape($item_name)]);

                                dispatch_notification('paid_subscription_messages', null, $subject, $mail);
                            }
                        }
                    }
                }
            }

            $start += $max;
        } while (!empty($subscribers));
    }
}
