<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core
 */

/**
 * Hook class.
 */
class Hook_cron_static_cache_cleanup
{
    /**
     * Get info from this hook.
     *
     * @param  ?TIME $last_run Last time run (null: never)
     * @param  ?boolean $calculate_num_queued Calculate the number of items queued, if possible (null: the hook may decide / low priority)
     * @return ?array Return a map of info about the hook (null: disabled)
     */
    public function info(?int $last_run, ?bool $calculate_num_queued) : ?array
    {
        return [
            'label' => 'Clear stale pages from the static cache',
            'num_queued' => null,
            'minutes_between_runs' => 60 * 24,
        ];
    }

    /**
     * Run function for system scheduler hooks. Searches for things to do. ->info(..., true) must be called before this method.
     *
     * @param  ?TIME $last_run Last time run (null: never)
     */
    public function run(?int $last_run)
    {
        // Infinite files can build up in the static cache, as there are infinite unused URL parameters bots can tack on.
        //  This is a disk space waste concern, and can trigger a Health Check error.
        //  So we clean out anything very expired so this stuff cannot build up (a little expired stays for fail-over mode).
        //  (Not caching foreign parameters is not possible, as they cannot feasibly be detected, tracking etc.)

        $path = get_custom_file_base() . '/caches/static';
        $dh = @opendir($path);
        while (($f = readdir($dh)) !== false) {
            if (preg_match('#\.(htm|br|gz|xml)$#', $f) != 0) {
                // Over a week old
                if (filemtime($path . '/' . $f) < time() - 24 * 60 * 60 * 7) {
                    @unlink($path . '/' . $f);
                }
            }
        }
        closedir($dh);
    }
}
