<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    actionlog
 */

/**
 * Hook class.
 */
class Hook_notification_actionlog extends Hook_notification__Staff
{
    /**
     * Find the initial setting that members have for a notification code (only applies to the member_could_potentially_enable members).
     *
     * @param  ID_TEXT $notification_code Notification code
     * @param  ?SHORT_TEXT $category The category within the notification code (null: none)
     * @param  MEMBER $member_id The member the notification would be for
     * @return integer Initial setting
     */
    public function get_initial_setting(string $notification_code, ?string $category, int $member_id) : int
    {
        return A_NA;
    }

    /**
     * Find whether a handled notification code supports categories.
     * (Content types, for example, will define notifications on specific categories, not just in general. The categories are interpreted by the hook and may be complex. E.g. it might be like a regexp match, or like FORUM:3 or TOPIC:100).
     *
     * @param  ID_TEXT $notification_code Notification code
     * @return boolean Whether it does
     */
    public function supports_categories(string $notification_code) : bool
    {
        return true;
    }

    /**
     * Standard function to create the standardised category tree.
     *
     * @param  ID_TEXT $notification_code Notification code
     * @param  ?ID_TEXT $id The ID of where we're looking under (null: N/A)
     * @return array Tree structure
     */
    public function create_category_tree(string $notification_code, ?string $id) : array
    {
        if (!addon_installed('actionlog')) {
            return [];
        }

        $page_links = [];

        require_all_lang();

        $types = $GLOBALS['SITE_DB']->query_select('actionlogs', ['DISTINCT the_type']);
        if (get_forum_type() == 'cns') {
            $types = array_merge($types, $GLOBALS['FORUM_DB']->query_select('f_moderator_logs', ['DISTINCT l_the_type AS the_type']));
        }
        foreach ($types as $type) {
            $lang = do_lang($type['the_type'], null, null, null, null, false);
            if ($lang === null) {
                continue;
            }
            $page_links[] = [
                'id' => $type['the_type'],
                'title' => $lang,
            ];
        }
        sort_maps_by($page_links, 'title', false, true);

        return $page_links;
    }

    /**
     * Get a list of all the notification codes this hook can handle.
     * (Addons can define hooks that handle whole sets of codes, so hooks are written so they can take wide authority).
     *
     * @return array List of codes (mapping between code names, and a pair: section and labelling for those codes)
     */
    public function list_handled_codes() : array
    {
        if (!addon_installed('actionlog')) {
            return [];
        }

        $list = [];
        $list['actionlog'] = [do_lang('STAFF'), do_lang('actionlog:NOTIFICATION_TYPE_actionlog')];
        return $list;
    }
}
