<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    cns_warnings
 */

/**
 * Hook class.
 */
class Hook_profiles_tabs_warnings
{
    /**
     * Find whether this hook is active.
     *
     * @param  MEMBER $member_id_of The ID of the member who is being viewed
     * @param  MEMBER $member_id_viewing The ID of the member who is doing the viewing
     * @return boolean Whether this hook is active
     */
    public function is_active(int $member_id_of, int $member_id_viewing) : bool
    {
        if (!addon_installed('cns_warnings')) {
            return false;
        }

        return (($member_id_of == $member_id_viewing) || (has_privilege($member_id_viewing, 'assume_any_member'))) && ($GLOBALS['FORUM_DB']->query_select_value('f_warnings', 'COUNT(*)', ['w_member_id' => $member_id_of, 'w_is_warning' => 1]) > 0);
    }

    /**
     * Render function for profile tab hooks.
     *
     * @param  MEMBER $member_id_of The ID of the member who is being viewed
     * @param  MEMBER $member_id_viewing The ID of the member who is doing the viewing
     * @param  boolean $leave_to_ajax_if_possible Whether to leave the tab contents null, if this hook supports it, so that AJAX can load it later
     * @return array A tuple: The tab title, the tab contents, the suggested tab order, the icon
     */
    public function render_tab(int $member_id_of, int $member_id_viewing, bool $leave_to_ajax_if_possible = false) : array
    {
        require_lang('cns_warnings');
        $title = do_lang_tempcode('WARNINGS');

        $order = 80;

        if ($leave_to_ajax_if_possible) {
            return [$title, null, $order, 'menu/social/warnings'];
        }

        require_lang('cns');
        require_css('cns');

        require_code('templates_columned_table');
        require_code('cns_topics');
        require_code('cns_groups');

        require_lang('fields');

        if (addon_installed('securitylogging')) {
            require_lang('submitban');
        }

        $header_row = columned_table_header_row([
            do_lang_tempcode('DATE'),
            do_lang_tempcode('BY'),
            do_lang_tempcode('PRIVATE_TOPIC'),
            do_lang_tempcode('ADDITIONAL_INFO')
        ]);

        $table_rows = new Tempcode();
        $rows = $GLOBALS['FORUM_DB']->query_select('f_warnings', ['*'], ['w_member_id' => $member_id_of, 'w_is_warning' => 1], 'ORDER BY w_time DESC');
        foreach ($rows as $row) {
            $row_contents = new Tempcode();
            // Basic warning info
            if ($row['w_explanation'] == '') {
                $explanation = do_lang('EXPLANATION') . ': ?';
            } else {
                $explanation = do_lang('EXPLANATION') . ': ' . str_replace("\n", ' ', $row['w_explanation']);
            }
            $row['w_explanation_orig'] = $row['w_explanation'];
            if (cms_mb_strlen($row['w_explanation']) > 61) {
                $explanation = do_lang('EXPLANATION') . ': ' . cms_mb_substr(str_replace("\n", ' ', $row['w_explanation']), 0, 61) . '...';
            }
            $row_contents->attach($explanation);

            // Warning private topic
            $private_topic = do_lang('NA');
            if ($row['w_topic_id'] !== null) {
                $topic_rows = $GLOBALS['FORUM_DB']->query_select('f_topics', ['*'], ['id' => $row['w_topic_id']], '', 1);
                if (array_key_exists(0, $topic_rows)) {
                    $topic_row = $topic_rows[0];
                    if (cns_may_access_topic($row['w_topic_id'])) {
                        $pt_url = build_url(['page' => 'topicview', 'id' => $row['w_topic_id']], get_module_zone('topicview'));
                        $private_topic = hyperlink($pt_url, do_lang('VIEW_TOPIC'), false, true);
                    }
                }
            }

            // Punitive actions
            $prows = $GLOBALS['FORUM_DB']->query_select('f_warnings_punitive', ['*'], ['p_warning_id' => $row['id']]);
            foreach ($prows as $prow) {
                $hook = get_hook_ob('systems', 'cns_warnings', $prow['p_hook'], 'Hook_cns_warnings_');
                if (method_exists($hook, 'generate_text') && ($hook->get_details() !== null)) {
                    $row_contents->attach('<br />' . $hook->generate_text($prow));
                }
            }

            $table_rows->attach(columned_table_row([
                get_timezoned_date_time($row['w_time']),
                $GLOBALS['FORUM_DRIVER']->member_profile_hyperlink($row['w_by']),
                $private_topic,
                $row_contents,
            ], false));
        }

        $warn_members_text = new Tempcode();
        if (has_privilege($member_id_viewing, 'warn_members')) {
            $full_warnings_url = build_url(['page' => 'warnings', 'type' => 'history', 'id' => $member_id_of], get_module_zone('warnings'));
            $warn_members_text = do_lang_tempcode('DESCRIPTION_CAN_WARN', hyperlink($full_warnings_url, do_lang_tempcode('REVIEW_WARNING_HISTORY'), false, false));
        }

        $table = do_template('COLUMNED_TABLE', ['_GUID' => '49a356e5af2576351360054344fc472b', 'HEADER_ROW' => $header_row, 'ROWS' => $table_rows, 'NONRESPONSIVE' => false]);

        $_content = do_template('COLUMNED_TABLE_SCREEN', [
            '_GUID' => '9dfa0fb6ea396d3b57cb447bc228a885',
            'TITLE' => '',
            'TEXT' => $warn_members_text,
            'TABLE' => $table,
            'SUBMIT_ICON' => null,
            'JS_FUNCTION_CALLS' => [],
        ]);

        $content = do_template('CNS_MEMBER_PROFILE_WARNINGS', ['_GUID' => 'fea98858f6bf89f1d9dc3ec995785a39', 'MEMBER_ID' => strval($member_id_of), 'WARNINGS' => $_content]);

        return [$title, $content, $order, 'menu/social/warnings'];
    }
}
