<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    galleries
 */

/**
 * Hook class.
 */
class Hook_rss_galleries
{
    /**
     * Run function for RSS hooks.
     *
     * @param  string $_filters A list of categories we accept from
     * @param  TIME $cutoff Cutoff time, before which we do not show results from
     * @param  string $prefix Prefix that represents the template set we use
     * @set RSS_ ATOM_
     * @param  string $date_string The standard format of date to use for the syndication type represented in the prefix
     * @param  integer $max The maximum number of entries to return, ordering by date
     * @return ?array A pair: The main syndication section, and a title (null: error)
     */
    public function run(string $_filters, int $cutoff, string $prefix, string $date_string, int $max) : ?array
    {
        if (!addon_installed('galleries')) {
            return null;
        }

        if (!has_actual_page_access(get_member(), 'galleries')) {
            return null;
        }

        $filters_1 = selectcode_to_sqlfragment($_filters, 'name', 'galleries', 'parent_id', 'name', 'name', false, false); // Note that the parameters are fiddled here so that category-set and record-set are the same, yet SQL is returned to deal in an entirely different record-set (entries' record-set)
        $filters = selectcode_to_sqlfragment($_filters, 'cat', 'galleries', 'parent_id', 'cat', 'name', false, false); // Note that the parameters are fiddled here so that category-set and record-set are the same, yet SQL is returned to deal in an entirely different record-set (entries' record-set)

        require_lang('galleries');

        $content = new Tempcode();
        $_galleries = [];
        if ($GLOBALS['SITE_DB']->query_value_if_there('SELECT COUNT(*) FROM ' . get_table_prefix() . 'galleries WHERE ' . $filters_1, false, true) < 3000) {
            $_galleries = $GLOBALS['SITE_DB']->query('SELECT fullname,name FROM ' . get_table_prefix() . 'galleries WHERE ' . $filters_1, null, 0, false, true);
            foreach ($_galleries as $i => $_gallery) {
                $_galleries[$i]['_title'] = get_translated_text($_gallery['fullname']);
            }
        }
        $galleries = collapse_2d_complexity('name', '_title', $_galleries);

        $extra_join = '';
        $extra_where = '';
        if (addon_installed('content_privacy')) {
            require_code('content_privacy');
            list($extra_join, $extra_where) = get_privacy_where_clause('video', 'r');
        }
        if (get_option('filter_regions') == '1') {
            require_code('locations');
            $extra_where .= sql_region_filter('video', 'r.id');
        }
        $query = 'SELECT r.*,\'video\' AS type,r.id AS r_id FROM ' . $GLOBALS['SITE_DB']->get_table_prefix() . 'videos r' . $extra_join . ' WHERE add_date>' . strval($cutoff) . ' AND ' . $filters . (((!has_privilege(get_member(), 'see_not_validated')) && (addon_installed('validation'))) ? ' AND validated=1 ' : '') . $extra_where . ' ORDER BY add_date DESC';
        $rows1 = $GLOBALS['SITE_DB']->query($query, $max, 0, false, false, ['title' => 'SHORT_TRANS', 'the_description' => 'LONG_TRANS__COMCODE']);

        $extra_join = '';
        $extra_where = '';
        if (addon_installed('content_privacy')) {
            require_code('content_privacy');
            list($extra_join, $extra_where) = get_privacy_where_clause('image', 'r');
        }
        if (get_option('filter_regions') == '1') {
            require_code('locations');
            $extra_where .= sql_region_filter('image', 'r.id');
        }
        $query = 'SELECT r.*,\'image\' AS type,r.id AS r_id FROM ' . $GLOBALS['SITE_DB']->get_table_prefix() . 'images r' . $extra_join . ' WHERE add_date>' . strval($cutoff) . ' AND ' . $filters . (((!has_privilege(get_member(), 'see_not_validated')) && (addon_installed('validation'))) ? ' AND validated=1 ' : '') . $extra_where . ' ORDER BY add_date DESC';
        $rows2 = browser_matches('itunes') ? [] : $GLOBALS['SITE_DB']->query($query, $max, 0, false, false, ['title' => 'SHORT_TRANS', 'the_description' => 'LONG_TRANS__COMCODE']);

        $rows = array_merge($rows1, $rows2);
        foreach ($rows as $row) {
            // Workaround that some DB backends don't allow multiple fields to have the same name, so we have to use r_id for our specifically selected field instead of id
            $row['id'] = $row['r_id'];
            unset($row['r_id']);

            if (has_category_access(get_member(), 'galleries', $row['cat'])) {
                $id = strval($row['id']);
                $author = $GLOBALS['FORUM_DRIVER']->get_username($row['submitter'], false, USERNAME_DEFAULT_BLANK);

                $news_date = date($date_string, $row['add_date']);
                $edit_date = ($row['edit_date'] === null) ? '' : date($date_string, $row['edit_date']);

                if (get_translated_text($row['title']) != '') {
                    $news_title = xmlentities(get_translated_text($row['title']));
                } else {
                    $news_title = xmlentities(do_lang('THIS_WITH_SIMPLE', (($row['type'] == 'video') ? do_lang('VIDEO') : do_lang('IMAGE')), $id));
                }
                $just_row = db_map_restrict($row, ['id', 'the_description']);
                $_summary = get_translated_tempcode($row['type'] . 's', $just_row, 'the_description');
                $summary = xmlentities($_summary->evaluate());
                $news = '';

                if (!array_key_exists($row['cat'], $galleries)) {
                    $_fullname = $GLOBALS['SITE_DB']->query_select_value_if_there('galleries', 'fullname', ['name' => $row['cat']]);
                    if ($_fullname === null) {
                        continue;
                    }
                    $galleries[$row['cat']] = get_translated_text($_fullname);
                }
                $category = $galleries[$row['cat']];
                $category_raw = $row['cat'];

                if ($row['type'] == 'video') {
                    $thumb_url = $row['thumb_url'];
                } else {
                    $thumb_url = $row['url'];
                }
                if (url_is_local($thumb_url)) {
                    $thumb_url = get_custom_base_url() . '/' . $thumb_url;
                }

                $view_url = build_url(['page' => 'galleries', 'type' => $row['type'], 'id' => $row['id']], get_module_zone('galleries'), [], false, false, true);

                if (($prefix == 'RSS_') && (get_option('is_on_comments') == '1') && ($row['allow_comments'] >= 1)) {
                    $if_comments = do_template('RSS_ENTRY_COMMENTS', ['_GUID' => '65dc0cec8c75f565c58c95fa1667aa1e', 'COMMENT_URL' => $view_url, 'ID' => $id], null, false, null, '.xml', 'xml');
                } else {
                    $if_comments = new Tempcode();
                }

                $enclosure_url = $row['url'];
                if ((url_is_local($enclosure_url)) || ($row['type'] == 'image')) {
                    list($enclosure_length, $enclosure_type) = get_enclosure_details($row['url'], $enclosure_url);
                } else {
                    $enclosure_url = null;
                    $enclosure_length = null;
                    $enclosure_type = null;
                }

                $meta = seo_meta_get_for($row['type'], $id);
                $keywords = trim($meta[0], ', ');

                $content->attach(do_template($prefix . 'ENTRY', [
                    'ENCLOSURE_URL' => $enclosure_url,
                    'ENCLOSURE_LENGTH' => $enclosure_length,
                    'ENCLOSURE_TYPE' => $enclosure_type,
                    'VIEW_URL' => $view_url,
                    'SUMMARY' => $summary,
                    'EDIT_DATE' => $edit_date,
                    'CONTENT_TYPE' => $row['type'],
                    'IF_COMMENTS' => $if_comments,
                    'TITLE' => $news_title,
                    'CATEGORY_RAW' => $category_raw,
                    'CATEGORY' => $category,
                    'AUTHOR' => $author,
                    'ID' => $id,
                    'NEWS' => $news,
                    'THUMB_URL' => $thumb_url,
                    'DATE' => $news_date,
                    'DURATION' => array_key_exists('video_length', $row) ? (strval(intval(floor(floatval($row['video_length'])) / 60.0)) . ':' . strval($row['video_length'] % 60)) : null,
                    'KEYWORDS' => $keywords,
                ], null, false, null, '.xml', 'xml'));
            }
        }

        require_lang('galleries');
        return [$content, do_lang('GALLERIES')];
    }
}
