<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    newsletter
 */

/**
 * Hook class.
 */
class Hook_task_export_newsletter_subscribers
{
    /**
     * Run the task hook.
     *
     * @param  LANGUAGE_NAME $lang Language subscribers should be using (if we're running a multi-language site)
     * @param  string $key Newsletter identification key
     * @param  ?string $file_type The file type to export with (null: default)
     * @return ?array A tuple of at least 2: Return mime-type, content (either Tempcode, or a string, or a filename and file-path pair to a temporary file), map of HTTP headers if transferring immediately, map of ini_set commands if transferring immediately (null: show standard success message)
     */
    public function run(string $lang, string $key, ?string $file_type = null) : ?array
    {
        if (!addon_installed('newsletter')) {
            return null;
        }

        require_lang('cns');

        require_code('newsletter');

        $send_details = [
            $key => true,
        ];

        require_code('files_spreadsheets_write');
        if ($file_type === null) {
            $file_type = spreadsheet_write_default();
        }
        $filename = 'newsletter_subscribers_' . $key . '.' . $file_type;
        $outfile_path = null;
        $sheet_writer = spreadsheet_open_write($outfile_path, $filename, CMS_Spreadsheet_Writer::ALGORITHM_RAW);

        $sheet_writer->write_row([
            do_lang('EMAIL_ADDRESS'),
            do_lang('FORENAME'),
            do_lang('SURNAME'),
            do_lang('NAME'),
            do_lang('NEWSLETTER_SEND_ID'),
            do_lang('PASSWORD_HASH'),
            do_lang('LANGUAGE'),
            do_lang('SALT'),
            do_lang('CONFIRM_CODE'),
            do_lang('JOIN_DATE'),
        ]);

        $max = 100;
        $max_rows = null;
        $start = 0;
        $count = 0;
        do {
            $_subscribers = newsletter_who_send_to($send_details, $lang, $start, $max, [], false);
            if ($count == 0) {
                list($subscribers, $totals) = $_subscribers;
                $max_rows = array_sum($totals);
            } else {
                list($subscribers) = $_subscribers;
            }

            foreach ($subscribers as $email_address => $subscriber_map) {
                task_log($this, 'Processing subscriber row', $count, $max_rows);

                $sheet_writer->write_row([
                    $email_address,
                    $subscriber_map['forename'],
                    $subscriber_map['surname'],
                    $subscriber_map['name'],
                    $subscriber_map['send_id'],
                    $subscriber_map['hash'],
                    $subscriber_map['language'],
                    $subscriber_map['salt'],
                    $subscriber_map['code_confirm'],
                    ($subscriber_map['join_time'] === null) ? '' : date('Y-m-d h:i:s', $subscriber_map['join_time']),
                ]);

                $count++;
            }

            $start += $max;
        } while (array_key_exists(0, $subscribers));

        $sheet_writer->close();

        $headers = [];
        $headers['Content-Type'] = $sheet_writer->get_mime_type();
        $headers['Content-Disposition'] = 'attachment; filename="' . escape_header($filename) . '"';

        $ini_set = [];
        $ini_set['ocproducts.xss_detect'] = '0';

        return [$sheet_writer->get_mime_type(), [$filename, $outfile_path], $headers, $ini_set];
    }
}
