<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core
 */

/**
 * Returns a list of pairs, for which permissions are false by default for ordinary usergroups.
 *
 * @return array List of pairs
 */
function get_false_permissions() : array
{
    return [
        ['_COMCODE', 'allow_html'],
        ['_COMCODE', 'comcode_dangerous'],
        ['_COMCODE', 'comcode_nuisance'],
        ['_COMCODE', 'use_very_dangerous_comcode'],
        ['STAFF_ACTIONS', 'access_closed_site'],
        ['STAFF_ACTIONS', 'bypass_bandwidth_restriction'],
        ['STAFF_ACTIONS', 'see_stack_trace'],
        ['STAFF_ACTIONS', 'view_profiling_modes'],
        ['STAFF_ACTIONS', 'access_overrun_site'],
        ['SUBMISSION', 'feature'],
        ['SUBMISSION', 'bypass_validation_highrange_content'],
        ['SUBMISSION', 'bypass_validation_midrange_content'],
        ['SUBMISSION', 'edit_highrange_content'],
        ['SUBMISSION', 'edit_midrange_content'],
        ['SUBMISSION', 'edit_lowrange_content'],
        ['SUBMISSION', 'edit_own_highrange_content'],
        ['SUBMISSION', 'edit_own_midrange_content'],
        ['SUBMISSION', 'delete_highrange_content'],
        ['SUBMISSION', 'delete_midrange_content'],
        ['SUBMISSION', 'delete_lowrange_content'],
        ['SUBMISSION', 'delete_own_highrange_content'],
        ['SUBMISSION', 'delete_own_midrange_content'],
        ['SUBMISSION', 'delete_own_lowrange_content'],
        ['SUBMISSION', 'can_submit_to_others_categories'],
        ['SUBMISSION', 'search_engine_links'],
        ['SUBMISSION', 'submit_cat_highrange_content'],
        ['SUBMISSION', 'submit_cat_midrange_content'],
        ['SUBMISSION', 'submit_cat_lowrange_content'],
        ['SUBMISSION', 'edit_cat_highrange_content'],
        ['SUBMISSION', 'edit_cat_midrange_content'],
        ['SUBMISSION', 'edit_cat_lowrange_content'],
        ['SUBMISSION', 'delete_cat_highrange_content'],
        ['SUBMISSION', 'delete_cat_midrange_content'],
        ['SUBMISSION', 'delete_cat_lowrange_content'],
        ['SUBMISSION', 'edit_own_cat_highrange_content'],
        ['SUBMISSION', 'edit_own_cat_midrange_content'],
        ['SUBMISSION', 'edit_own_cat_lowrange_content'],
        ['SUBMISSION', 'delete_own_cat_highrange_content'],
        ['SUBMISSION', 'delete_own_cat_midrange_content'],
        ['SUBMISSION', 'delete_own_cat_lowrange_content'],
        ['SUBMISSION', 'mass_import'],
        ['SUBMISSION', 'scheduled_publication_times'],
        ['SUBMISSION', 'mass_delete_from_ip'],
        ['SUBMISSION', 'exceed_filesize_limit'],
        ['SUBMISSION', 'draw_to_server'],
        ['GENERAL_SETTINGS', 'open_virtual_roots'],
        ['GENERAL_SETTINGS', 'sees_javascript_error_alerts'],
        ['GENERAL_SETTINGS', 'see_software_docs'],
        ['GENERAL_SETTINGS', 'see_not_validated'],
        ['GENERAL_SETTINGS', 'may_enable_staff_notifications'],
        ['GENERAL_SETTINGS', 'bypass_flood_control'],
        ['GENERAL_SETTINGS', 'remove_page_split'],
        ['GENERAL_SETTINGS', 'bypass_wordfilter'],
        ['SUBMISSION', 'perform_keyword_check'],
        ['SUBMISSION', 'have_personal_category'],
    ];
}

/**
 * Returns a list of pairs, for which permissions are true by default for ordinary usergroups.
 *
 * @return array List of pairs
 */
function get_true_permissions() : array
{
    return [
        ['SUBMISSION', 'edit_own_lowrange_content'],
        ['SUBMISSION', 'submit_highrange_content'],
        ['SUBMISSION', 'submit_midrange_content'],
        ['SUBMISSION', 'submit_lowrange_content'],
        ['SUBMISSION', 'bypass_validation_lowrange_content'],
        ['_FEEDBACK', 'rate'],
        ['_FEEDBACK', 'comment'],
        ['VOTE', 'vote_in_polls'],
        ['GENERAL_SETTINGS', 'jump_to_not_validated'],
        ['_COMCODE', 'reuse_others_attachments'],
        ['GENERAL_SETTINGS', 'see_php_errors'],
    ];
}

/**
 * Check if a privilege exists.
 *
 * @param  ID_TEXT $name The name of the option
 * @return boolean Whether it exists
 */
function privilege_exists(string $name) : bool
{
    $test = $GLOBALS['SITE_DB']->query_select_value_if_there('privilege_list', 'the_name', ['the_name' => $name]);
    return $test !== null;
}

/**
 * Add a privilege, and apply it to every usergroup.
 *
 * @param  ID_TEXT $section The section the privilege is filled under
 * @param  ID_TEXT $name The codename for the privilege
 * @param  boolean $default Whether this privilege is granted to all usergroups by default
 * @param  boolean $not_even_mods Whether this privilege is not granted to supermoderators by default (something very sensitive); only applies if $default is true
 * @param  boolean $not_for_probation An exception for if $default is true, don't assign the privilege to the probation group
 * @param  boolean $insert Whether to insert into the database (if this is set to false you need to read the return parameters and do it yourself - which is used for efficient bulk inserts across many privileges at once)
 * @return array A pair: The map parameter for the group_privileges table, the map parameter for the privilege_list table
 */
function add_privilege(string $section, string $name, bool $default = false, bool $not_even_mods = false, bool $not_for_probation = false, bool $insert = true) : array
{
    if (get_forum_type() == 'cns') {
        require_code('cns_groups');
        $probation_group = get_probation_group();
    } else {
        $probation_group = null;
    }

    // We do bulk inserts, for performance reasons
    $ins_privilege = [];
    $ins_group_id = [];
    $ins_the_page = [];
    $ins_module_the_name = [];
    $ins_category_name = [];
    $ins_the_value = [];

    $usergroups = $GLOBALS['FORUM_DRIVER']->get_usergroup_list(false, true);
    $admin_groups = array_merge($GLOBALS['FORUM_DRIVER']->get_super_admin_groups(), $GLOBALS['FORUM_DRIVER']->get_moderator_groups());
    foreach (array_keys($usergroups) as $id) {
        if ((($default) && ((!$not_for_probation) || ($id !== $probation_group))) || ((in_array($id, $admin_groups)) && (!$not_even_mods))) {
            $ins_privilege[] = $name;
            $ins_group_id[] = $id;
            $ins_the_page[] = '';
            $ins_module_the_name[] = '';
            $ins_category_name[] = '';
            $ins_the_value[] = 1;
        }
    }

    $group_privileges_insert = [
        'privilege' => $ins_privilege,
        'group_id' => $ins_group_id,
        'the_page' => $ins_the_page,
        'module_the_name' => $ins_module_the_name,
        'category_name' => $ins_category_name,
        'the_value' => $ins_the_value,
    ];

    $privilege_list_insert = [
        'p_section' => [$section],
        'the_name' => [$name],
        'the_default' => [$default ? 1 : 0],
    ];

    if ($insert) {
        $GLOBALS['SITE_DB']->query_insert('group_privileges', $group_privileges_insert);
        $GLOBALS['SITE_DB']->query_insert('privilege_list', $privilege_list_insert);
    }

    return [$group_privileges_insert, $privilege_list_insert];
}

/**
 * Sets the privilege of a usergroup.
 *
 * @param  GROUP $group_id The usergroup having the permission set
 * @param  ID_TEXT $permission The codename of the permission
 * @param  boolean $value Whether the usergroup has the permission
 * @param  ?ID_TEXT $page The page being checked (null: current page)
 * @param  ?ID_TEXT $permission_module The permission module (null: none required)
 * @param  ?ID_TEXT $category_name The category-name/value for the permission (null: none required)
 */
function set_privilege(int $group_id, string $permission, bool $value, ?string $page = null, ?string $permission_module = null, ?string $category_name = null)
{
    if ($page === null) {
        $page = '';
    }
    if ($permission_module === null) {
        $permission_module = '';
    }
    if ($category_name === null) {
        $category_name = '';
    }

    $db = $GLOBALS[((($permission_module == 'forums') || ($permission_module == 'topics')) && (get_forum_type() == 'cns')) ? 'FORUM_DB' : 'SITE_DB'];

    $db->query_delete('group_privileges', ['privilege' => $permission, 'group_id' => $group_id, 'the_page' => $page, 'module_the_name' => $permission_module, 'category_name' => $category_name], '', 1);
    $db->query_insert('group_privileges', ['privilege' => $permission, 'group_id' => $group_id, 'the_page' => $page, 'module_the_name' => $permission_module, 'category_name' => $category_name, 'the_value' => $value ? 1 : 0]);

    global $PRIVILEGE_CACHE;
    $PRIVILEGE_CACHE = [];
}

/**
 * Rename a privilege.
 *
 * @param  ID_TEXT $old The old name
 * @param  ID_TEXT $new The new name
 */
function rename_privilege(string $old, string $new)
{
    $GLOBALS['SITE_DB']->query_update('privilege_list', ['the_name' => $new], ['the_name' => $old], '', 1);
    $GLOBALS['SITE_DB']->query_update('group_privileges', ['privilege' => $new], ['privilege' => $old], '', 1);
    $GLOBALS['SITE_DB']->query_update('member_privileges', ['privilege' => $new], ['privilege' => $old], '', 1);
}

/**
 * Delete a privilege.
 *
 * @param  mixed $privileges The codename of the privilege(s) to delete
 */
function delete_privilege($privileges)
{
    if (!is_array($privileges)) {
        $privileges = [$privileges];
    }

    $privilege_list_query = 'DELETE FROM ' . get_table_prefix() . 'privilege_list WHERE 1=0';
    $group_privileges_query = 'DELETE FROM ' . get_table_prefix() . 'group_privileges WHERE 1=0';

    foreach ($privileges as $privilege) {
        $privilege_list_query .= ' OR ' . db_string_equal_to('the_name', $privilege);
        $group_privileges_query .= ' OR ' . db_string_equal_to('privilege', $privilege);
    }

    $GLOBALS['SITE_DB']->query($privilege_list_query);
    $GLOBALS['SITE_DB']->query($group_privileges_query);
}
