<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core
 */

/*EXTRA FUNCTIONS: xcache\_.+*/

/**
 * Cache driver class.
 */
class Persistent_caching_xcache
{
    public $objects_list = null;

    /**
     * Instruction to load up the objects list.
     *
     * @return array The list of objects
     */
    public function &load_objects_list() : array
    {
        if ($this->objects_list === null) {
            $this->objects_list = xcache_get(get_file_base() . 'PERSISTENT_CACHE_OBJECTS');
            if ($this->objects_list === null) {
                $this->objects_list = [];
            }
        }
        return $this->objects_list;
    }

    /**
     * Get data from the persistent cache.
     *
     * @param  string $key Key
     * @param  ?TIME $min_cache_date Minimum timestamp that entries from the cache may hold (null: don't care)
     * @return ?mixed The data (null: not found / null entry)
     */
    public function get(string $key, ?int $min_cache_date = null)
    {
        $data = xcache_get($key);
        if ($data === false) {
            return null;
        }
        if (($min_cache_date !== null) && ($data[0] < $min_cache_date)) {
            return null;
        }
        return $data[1];
    }

    /**
     * Put data into the persistent cache.
     *
     * @param  string $key Key
     * @param  mixed $data The data
     * @param  integer $flags Various flags (parameter not used)
     * @param  ?integer $expire_secs The expiration time in seconds (null: no expiry)
     */
    public function set(string $key, $data, int $flags = 0, ?int $expire_secs = null)
    {
        // Update list of persistent-objects
        $this->load_objects_list();
        if (!array_key_exists($key, $this->objects_list)) {
            $this->objects_list[$key] = true;
            xcache_set(get_file_base() . 'PERSISTENT_CACHE_OBJECTS', $this->objects_list);
        }

        xcache_set($key, [time(), $data], $expire_secs);
    }

    /**
     * Delete data from the persistent cache.
     *
     * @param  string $key Key
     */
    public function delete(string $key)
    {
        // Update list of persistent-objects
        $this->load_objects_list();
        unset($this->objects_list[$key]);
        //xcache_set(get_file_base() . 'PERSISTENT_CACHE_OBJECTS', $this->objects_list); Wasteful

        xcache_unset($key);
    }

    /**
     * Remove all data from the persistent cache.
     */
    public function flush()
    {
        // Update list of persistent-objects
        $this->objects_list = [];
        xcache_set(get_file_base() . 'PERSISTENT_CACHE_OBJECTS', $this->objects_list);

        xcache_unset_by_prefix('');
    }
}
