<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    setupwizard
 */

/**
 * Standard code module initialisation function.
 *
 * @ignore
 */
function init__setupwizard()
{
    if (!defined('BLOCK_POSITION_MAIN')) {
        define('BLOCK_POSITION_MAIN', 1);
        define('BLOCK_POSITION_CELL', 2);
        define('BLOCK_POSITION_PANEL', 4);
    }
}

/**
 * Install test content.
 */
function install_test_content()
{
    require_code('lorem');

    if (!running_script('install')) {
        uninstall_test_content();
    }

    push_query_limiting(false);
    set_mass_import_mode(true);

    $hooks = find_all_hook_obs('systems', 'addon_registry', 'Hook_addon_registry_');
    foreach ($hooks as $ob) {
        if (method_exists($ob, 'install_test_content')) {
            $ob->install_test_content();
        }
    }

    pop_query_limiting();
    set_mass_import_mode(false);

    delete_cache_entry('main_staff_checklist');
    delete_cache_entry('menu');
}

/**
 * Uninstall test content.
 */
function uninstall_test_content()
{
    require_code('lorem');

    push_query_limiting(false);
    set_mass_import_mode(true);

    $hooks = find_all_hook_obs('systems', 'addon_registry', 'Hook_addon_registry_');
    foreach ($hooks as $ob) {
        if (method_exists($ob, 'uninstall_test_content')) {
            $ob->uninstall_test_content();
        }
    }

    pop_query_limiting();
    set_mass_import_mode(false);

    delete_cache_entry('main_staff_checklist');
    delete_cache_entry('menu');
}

/**
 * Get Comcode for the pages in the zone.
 *
 * @param  array $installprofileblocks List of blocks in the install profile
 * @param  ?array $block_options Block options in the install profile (null: none)
 * @param  boolean $collapse_zones Whether we have zone collapsing on
 * @param  ID_TEXT $installprofile ID of the install profile
 * @return array Structure of pages
 *
 * @ignore
 */
function _get_zone_pages(array $installprofileblocks, ?array $block_options, bool $collapse_zones, string $installprofile) : array
{
    $page_structure = [];

    // Find blocks used, from environment
    $zone_blocks = [];
    $zone_blocks['site'] = [];
    foreach ($_POST as $key => $value) {
        if (substr($key, 0, 6) == 'block_') {
            $block = substr($key, 6);
            if (substr($block, 0, 5) == 'SITE_') {
                $zone_blocks['site'][substr($block, 5)] = $value;
            }
        }
    }
    $zone_blocks[''] = $zone_blocks['site'];

    // Order them according to profile
    foreach ($installprofileblocks as $set) {
        foreach ($set as $block) {
            if (isset($zone_blocks['site'][$block])) {
                $value = $zone_blocks['site'][$block];
                unset($zone_blocks['site'][$block]);
                $zone_blocks['site'] += [$block => $value];
            }
        }
    }

    // Work out all Comcode
    foreach ($zone_blocks as $zone => $blocks) {
        if ($collapse_zones) {
            if ($zone == '') {
                continue;
            }
            if ($zone == 'site') {
                $zone = '';
            }
        }

        $page_structure[$zone] = [];

        // Generally work out what Comcode goes where
        $cells = '';
        $main = '';
        $left = '';
        $right = '';
        $cell_count = 0;
        foreach ($blocks as $block => $val) {
            $params = '';
            if ($block_options !== null) {
                if (array_key_exists($block, $block_options)) {
                    foreach ($block_options[$block] as $block_option => $block_option_value) {
                        $params .= ' ' . $block_option . '="' . comcode_escape($block_option_value) . '"';
                    }
                }
            }
            $block_comcode = "[block" . $params . "]" . $block . "[/block]";
            if ($val == 'PANEL_LEFT') {
                $left .= $block_comcode . "\n";
            }
            if ($val == 'PANEL_RIGHT') {
                $right .= $block_comcode . "\n";
            }
            if ($val == 'YES') {
                $main .= $block_comcode . "\n\n";
            }
            if ($val == 'YES_CELL') {
                if ($cell_count % 2 == 0) {
                    $cells .= '[surround="fp-col-blocks-wrap"]';
                }
                $cells .= "\t" . '[surround="fp-col-block ' . (($cell_count % 2 == 0) ? 'left' : 'right') . '"]' . "\n\t\t" . $block_comcode . "\n\t" . '[/surround]' . "\n";
                if ($cell_count % 2 == 1) {
                    $cells .= "[/surround]\n\n";
                }
                $cell_count++;
            }
        }
        if ($cells != '') { // Odd number of cells chosen, close off at odd point
            if ($cell_count % 2 == 1) {
                $cells .= "[/surround]\n\n";
            }
        }

        // Front page
        $comcode = '';
        $comcode .= "[semihtml]\n[title=\"1\"]Welcome to {\$SITE_NAME*}[/title]\n\n[block=\"3\"]main_greeting[/block]\n\n";
        if ($cells != '') {
            $comcode .= $cells;
        }
        $main .= "\n\n[block]main_comcode_page_children[/block]\n[/semihtml]";
        $comcode .= $main;
        $page_structure[$zone][DEFAULT_ZONE_PAGE_NAME] = $comcode;

        // Left panel
        $comcode = '';
        require_lang('menus');
        if ($installprofile == '') {
            if (($zone == '') && (!$collapse_zones)) {
                $comcode .= cms_file_get_contents_safe(get_file_base() . '/pages/comcode/' . fallback_lang() . '/panel_left.txt', FILE_READ_UNIXIFIED_TEXT | FILE_READ_BOM);
            } else {
                $comcode .= cms_file_get_contents_safe(get_file_base() . '/site/pages/comcode/' . fallback_lang() . '/panel_left.txt', FILE_READ_UNIXIFIED_TEXT | FILE_READ_BOM);
            }
        }
        $comcode .= $left;
        $page_structure[$zone]['left'] = $comcode;

        // Right panel
        $comcode = '';
        $comcode .= $right;
        if (post_param_integer('include_cms_advert', 0) == 1) {
            $comcode .= '[center][url="' . get_brand_base_url() . '/?from=logo"][img="Powered by Composr"]' . get_brand_base_url() . '/uploads/website_specific/composr.app/logos/a.png[/img][/url][/center]';
        }
        $page_structure[$zone]['right'] = $comcode;
    }

    return $page_structure;
}
