<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_themeing
 */

/**
 * Standard code module initialisation function.
 *
 * @ignore
 */
function init__themes2()
{
    global $THEME_SEED_CACHE;
    $THEME_SEED_CACHE = [];
}

/**
 * Find the seed of a theme.
 *
 * @param  ?ID_TEXT $theme The theme name (null: current theme)
 * @return ID_TEXT The seed colour
 */
function find_theme_seed(?string $theme = null) : string
{
    if ($theme === null) {
        $theme = $GLOBALS['FORUM_DRIVER']->get_theme();
    }

    global $THEME_SEED_CACHE;
    if (isset($THEME_SEED_CACHE[$theme])) {
        return $THEME_SEED_CACHE[$theme];
    }

    $test = get_theme_option('themewizard_built_with_seed', '');
    if ($test != '') {
        $THEME_SEED_CACHE[$theme] = $test;
    } else {
        $css_path = get_custom_file_base() . '/themes/' . $theme . '/css_custom/_colours.css';
        if (!is_file($css_path)) {
            $css_path = get_file_base() . '/themes/default/css/_colours.css';
        }
        $css_file_contents = cms_file_get_contents_safe($css_path, FILE_READ_LOCK | FILE_READ_UNIXIFIED_TEXT);
        $matches = [];
        if (preg_match('#\{\$THEMEWIZARD_COLOR,\#(.{6}),seed,.*\}#', $css_file_contents, $matches) != 0) {
            $THEME_SEED_CACHE[$theme] = $matches[1];
        } else {
            /*if ($no_easy_anchor)
            {
                   We could put some auto-detection code here; possibly a future improvement but not needed currently.
            } else {*/
            if ($theme == 'default') {
                fatal_exit(do_lang_tempcode('INTERNAL_ERROR'));
            }
            $THEME_SEED_CACHE[$theme] = find_theme_seed('default');
            //}
        }
    }

    return $THEME_SEED_CACHE[$theme];
}

/**
 * Try and find some CDNs to use.
 *
 * @return string List of CDNs
 */
function autoprobe_cdns() : string
{
    require_code('files');

    $base_url = get_base_url();
    $parsed = cms_parse_url_safe($base_url);
    $domain_name = $parsed['host'];

    if (is_local_machine()) {
        set_value('cdn', '');
        return '';
    }

    $try = [
        'cdn' . '.' . $domain_name,
        (substr($domain_name, 0, 4) == 'www.') ? preg_replace('#^www\.#', '', $domain_name) : ('www' . '.' . $domain_name),
        'ftp' . '.' . $domain_name,
        'mail' . '.' . $domain_name,
        /*'smtp' . '.' . $domain_name,  Let's be at least somewhat reasonable ;-)
        'imap' . '.' . $domain_name,
        'pop' . '.' . $domain_name,
        'webmail' . '.' . $domain_name,*/
    ];

    $detected_cdns = '';
    $expected = cms_file_get_contents_safe(get_file_base() . '/themes/default/images/icons/editor/comcode.svg', FILE_READ_LOCK | FILE_READ_BOM);
    foreach ($try as $t) {
        if ($t === null) {
            continue;
        }

        if (preg_match('#^' . preg_quote($t, '#') . '($|\.|/|:)#', $domain_name) == 0) { // Don't use it if it is in the base URL
            $test_url = (tacit_https() ? 'https://' : 'http://') . $t . $parsed['path'] . '/themes/default/images/icons/editor/comcode.svg';

            $test_result = http_get_contents($test_url, ['convert_to_internal_encoding' => true, 'trigger_error' => false, 'timeout' => 0.25]);

            if (($test_result !== null) && ($test_result == $expected)) {
                if ($detected_cdns != '') {
                    $detected_cdns .= ',';
                }
                $detected_cdns .= $t;
            }
        }
    }

    set_value('cdn', $detected_cdns);
    return $detected_cdns;
}

/**
 * Get an array listing all the themes present.
 *
 * @param  boolean $needs_themewizard_support Whether to only return Theme Wizard supporting themes
 * @return array A map of all themes (name=>title) OR if requested a map of theme name to full theme details
 */
function find_all_themes(bool $needs_themewizard_support = false) : array
{
    if ($GLOBALS['IN_MINIKERNEL_VERSION']) {
        return ['default' => do_lang('DEFAULT')];
    }

    require_code('files');

    $themes = [];
    $_dir = @opendir(get_file_base() . '/themes/');
    if ($_dir !== false) {
        while (false !== ($file = readdir($_dir))) {
            if ((strpos($file, '.') === false) && (is_dir(get_file_base() . '/themes/' . $file))) {
                $themes[$file] = null;
            }
        }
        closedir($_dir);
    }
    if (get_custom_file_base() != get_file_base()) {
        $_dir = @opendir(get_custom_file_base() . '/themes/');
        if ($_dir !== false) {
            while (false !== ($file = readdir($_dir))) {
                if ((strpos($file, '.') === false) && (is_dir(get_file_base() . '/themes/' . $file))) {
                    $themes[$file] = null;
                }
            }
            closedir($_dir);
        }
    }

    foreach (array_keys($themes) as $file) {
        if (get_theme_option('enable_themewizard', null, $file) == '1') {
            $themes[$file] = get_theme_option('title', null, $file);
        } else {
            unset($themes[$file]);
        }
    }

    // Sort
    cms_mb_asort($themes, SORT_NATURAL | SORT_FLAG_CASE);

    // Default theme should go first
    if (isset($themes['default'])) {
        $temp = $themes['default'];
        $temp_2 = $themes;
        $themes = ['default' => $temp] + $temp_2;
    }

    // Admin theme should go last
    if (isset($themes['admin'])) {
        $temp = $themes['admin'];
        unset($themes['admin']);
        $themes['admin'] = $temp;
    }

    return $themes;
}

/**
 * Get a UI list for choosing a theme.
 *
 * @param  ?ID_TEXT $theme The theme to select by default (null: no specific default)
 * @param  boolean $include_rely Whether to include the 'rely on forums' entry
 * @param  boolean $show_everything Whether to forget about permissions for this list, and also show additional detail
 * @param  ID_TEXT $default_message_string The language string to use for the default answer
 * @param  boolean $administrative_only Whether to only show administrative themes
 * @param  boolean $themewizard_only Whether to only show Theme Wizard supporting themes
 * @return Tempcode The list
 */
function create_selection_list_themes(?string $theme = null, bool $include_rely = true, bool $show_everything = false, string $default_message_string = 'RELY_FORUMS', bool $administrative_only = false, bool $themewizard_only = false) : object
{
    if ($include_rely) {
        $entries = form_input_list_entry('-1', false, do_lang_tempcode($default_message_string));
    } else {
        $entries = new Tempcode();
    }
    $themes = find_all_themes();
    foreach ($themes as $_theme => $title) {
        if (($administrative_only) && (get_theme_option('capability_administrative', null, $_theme) == '0')) {
            continue;
        }
        if (($themewizard_only) && (get_theme_option('enable_themewizard', null, $_theme) == '0')) {
            continue;
        }

        if (($show_everything) || (has_category_access(get_member(), 'theme', $_theme))) {
            $selected = ($theme == $_theme);
            $_title = $title;
            if ($show_everything) {
                $has_guest_access = ($_theme == 'default') || (has_category_access($GLOBALS['FORUM_DRIVER']->get_guest_id(), 'theme', $_theme));
                if ($has_guest_access) {
                    $_title = do_lang('themes:THEME_IN_LIST_NON_RESTRICTED', $_title, $_theme);
                } else {
                    $_title = do_lang('themes:THEME_IN_LIST_RESTRICTED', $_title, $_theme);
                }
            }
            $entries->attach(form_input_list_entry($_theme, $selected, $_title));
        }
    }
    if ($entries->is_empty()) {
        $entries->attach(form_input_list_entry('default', false, $themes['default']));
    }
    return $entries;
}

/**
 * Get all the templates for a theme.
 *
 * @param  ID_TEXT $theme The theme to search for
 * @param  string $directory Subdirectory type to look in
 * @set templates javascript xml text css
 * @param  ?string $suffix File type suffix of template file (e.g. .tpl) (null: calculate for the $directory norm)
 * @set .tpl .js .xml .txt .css
 * @param  boolean $this_theme_only Just for this theme
 * @return array A map of the files (file=>path)
 */
function get_template_files_list(string $theme, string $directory, ?string $suffix = null, bool $this_theme_only = false) : array
{
    if ($suffix === null) {
        switch ($directory) {
            case 'templates':
                $suffix = '.tpl';
                break;

            case 'javascript':
                $suffix = '.js';
                break;

            case 'xml':
                $suffix = '.xml';
                break;

            case 'text':
                $suffix = '.txt';
                break;

            case 'css':
                $suffix = '.css';
                break;
        }
    }

    $out = [];
    if (($theme == 'default') || (!$this_theme_only)) {
        $out = array_merge($out, _get_template_files_list(get_file_base(), 'default/' . $directory, $suffix));
        $out = array_merge($out, _get_template_files_list(get_custom_file_base(), 'default/' . $directory . '_custom', $suffix));
    }
    if ($theme != 'default') {
        $out = array_merge($out, _get_template_files_list(get_custom_file_base(), $theme . '/' . $directory, $suffix));
        $out = array_merge($out, _get_template_files_list(get_custom_file_base(), $theme . '/' . $directory . '_custom', $suffix));
    }
    ksort($out);

    return $out;
}

/**
 * Get all the template files / revisions for a template file, in a certain directory.
 *
 * @param  PATH $base_dir The path to search relative to
 * @param  PATH $subdir The subdirectory to search
 * @param  string $suffix File type suffix of template file (e.g. .tpl)
 * @set .tpl .js .xml .txt .css
 * @return array A map of the files (file=>path)
 *
 * @ignore
 */
function _get_template_files_list(string $base_dir, string $subdir, string $suffix) : array
{
    $_dir = @opendir($base_dir . '/themes/' . $subdir);
    if ($_dir !== false) {
        // Find all the themes
        $files_list = [];
        while (false !== ($file = readdir($_dir))) {
            if (cms_strtolower_ascii(substr($file, -strlen($suffix))) == $suffix) {
                $files_list[$file] = $base_dir . '/themes/' . $subdir . '/' . $file;
            }
        }
        closedir($_dir);
        return $files_list;
    }
    return [];
}

/**
 * Find where a template is.
 *
 * @param  ID_TEXT $file The file
 * @param  ID_TEXT $subdir The theme subdirectory we're working against
 * @param  ID_TEXT $theme The theme
 * @return ?PATH The path (null: not found)
 */
function find_template_path(string $file, string $subdir, string $theme) : ?string
{
    static $cache = [];
    if (isset($cache[$file][$subdir][$theme])) {
        return $cache[$file][$subdir][$theme];
    }

    $suffix = '.' . get_file_extension($file);
    $_file = basename($file, $suffix);
    list($searched_theme, $searched_directory, $searched_suffix) = find_template_place($_file, $theme, $suffix, $subdir);
    if ($searched_theme === null) {
        return null;
    }
    $template_path = get_custom_file_base() . '/themes/' . $searched_theme . $searched_directory . $_file . $suffix;
    if (!is_file($template_path)) {
        $template_path = get_file_base() . '/themes/' . $searched_theme . $searched_directory . $_file . $suffix;
        if (!is_file($template_path)) {
            $template_path = null;
        }
    }

    $cache[$file][$subdir][$theme] = $template_path;

    return $template_path;
}

/**
 * Find GUIDs used for a template, for use with the template editor.
 *
 * @param  ID_TEXT $file The template (including subdirectory and suffix)
 * @param  ?ID_TEXT $active_guid GUID currently in use, to highlight; e.g. opening up the editor on a template instance within a screen meta-tree (null: none)
 * @return array List of GUID details (template-ready)
 */
function find_template_guids(string $file, ?string $active_guid = null) : array
{
    $suffix = '.' . get_file_extension($file);
    $clean_file = basename($file, $suffix);

    $guids = [];
    $_guids = @unserialize(@cms_file_get_contents_safe(get_file_base() . '/data/guids.bin', FILE_READ_LOCK));
    if (($_guids !== false) && (array_key_exists($clean_file, $_guids))) {
        foreach ($_guids[$clean_file] as $_guid) {
            $guids[] = [
                'GUID_FILENAME' => $_guid[0],
                'GUID_LINE' => strval($_guid[1]),
                'GUID_GUID' => $_guid[2],
                'GUID_IS_LIVE' => ($_guid[2] === $active_guid),
            ];
        }
    }
    return $guids;
}

/**
 * Find parameters used by a template.
 *
 * @param  ID_TEXT $file The template (including subdirectory and suffix)
 * @return array List of template parameters
 */
function find_template_parameters(string $file) : array
{
    $parameters = [];

    require_code('lorem');
    $all_previews = find_all_previews__by_template();
    if (array_key_exists($file, $all_previews)) {
        list($hook, $function) = $all_previews[$file];
        global $LOREM_AVOID_GLOBALISE;
        $LOREM_AVOID_GLOBALISE = true;
        $temp = render_screen_preview($hook, $function, $file);
    }

    global $KNOWN_TEMPLATE_PARAMETERS;
    $_file = basename($file, '.' . get_file_extension($file));
    if (isset($KNOWN_TEMPLATE_PARAMETERS[$_file])) {
        return $KNOWN_TEMPLATE_PARAMETERS[$_file];
    }

    // No screen preview, so look inside the template instead...

    $template_path = find_template_path(basename($file), dirname($file), 'default');
    if ($template_path === null) {
        return $parameters;
    }

    $matches = [];
    $cnt = preg_match_all('#\{([\w]\w*)[\*;%\#]?\}#', cms_file_get_contents_safe($template_path, FILE_READ_LOCK), $matches);
    $p_done = [];
    for ($j = 0; $j < $cnt; $j++) {
        $parameters[] = $matches[1][$j];
    }

    return array_unique($parameters);
}

/**
 * A theme image has been passed through by POST, either as a file (a new theme image), or as a reference to an existing one. Get the image code from the POST data.
 * Note that post_param_image is a more comprehensive function, if you accept other types of inputs too.
 *
 * @param  ID_TEXT $type The type of theme image
 * @param  boolean $required Whether a code is required
 * @param  ID_TEXT $field_file Form field for uploading
 * @param  ID_TEXT $field_choose Form field for choosing
 * @param  ?object $db Database connector (null: site database)
 * @param  ?PATH $upload_to Where to upload the theme images to (null: something sensible)
 * @return ID_TEXT The (possibly randomised) theme image code
 */
function post_param_theme_img_code(string $type, bool $required = false, string $field_file = 'file', string $field_choose = 'theme_img_code', ?object $db = null, ?string $upload_to = null) : string
{
    if ($db === null) {
        $db = $GLOBALS['SITE_DB'];
    }

    if ($upload_to === null) {
        $upload_to = 'themes/default/images_custom/' . $type;
        @mkdir(get_custom_file_base() . '/' . $upload_to, 0777);
        if (file_exists(get_custom_file_base() . '/' . $upload_to)) {
            fix_permissions(get_custom_file_base() . '/' . $upload_to);
        } else {
            $upload_to = 'themes/default/images_custom';
        }
    }

    require_code('uploads');
    is_plupload(true);
    if (((array_key_exists($field_file, $_FILES)) && ((is_plupload()) || (is_uploaded_file($_FILES[$field_file]['tmp_name']))))) {
        $urls = get_url('', $field_file, $upload_to, OBFUSCATE_NEVER, CMS_UPLOAD_IMAGE, false);

        $theme_img_code = $type . '/' . basename($urls[2], '.' . get_file_extension($urls[2]));
        if ($GLOBALS['SITE_DB']->query_select_value_if_there('theme_images', 'id', ['id' => $theme_img_code]) !== null) {
            $theme_img_code = $type . '/' . uniqid('', true);
        }

        $db->query_insert('theme_images', ['id' => $theme_img_code, 'theme' => 'default', 'url' => $urls[0], 'lang' => get_site_default_lang()]);

        Self_learning_cache::erase_smart_cache();
    } else {
        $theme_img_code = post_param_string($field_choose, '');

        if ($theme_img_code == '') {
            if (!$required) {
                return '';
            }
            warn_exit(do_lang_tempcode('IMPROPERLY_FILLED_IN_UPLOAD'));
        }
    }
    return $theme_img_code;
}

/**
 * Recursively find theme images under the specified details. Does not find custom theme images, as it doesn't check the DB.
 *
 * @param  ID_TEXT $theme The theme
 * @param  string $subdir The subdirectory to search under
 * @param  array $langs A map (lang=>true) of the languages in the system, so the codes may be filtered out of the image codes in our result list
 * @return array A map, theme-image-code=>URL
 */
function find_images_do_dir(string $theme, string $subdir, array $langs) : array
{
    require_code('images');

    $full = (($theme == 'default' || $theme == 'admin') ? get_file_base() : get_custom_file_base()) . '/themes/' . filter_naughty($theme) . '/' . filter_naughty($subdir);
    $out = [];

    $_dir = @opendir($full);
    if ($_dir !== false) {
        while (false !== ($file = readdir($_dir))) {
            if ($file[0] != '.') {
                if (is_dir($full . $file)) {
                    $out = array_merge($out, find_images_do_dir($theme, $subdir . $file . '/', $langs));
                } elseif (is_image($file, IMAGE_CRITERIA_WEBSAFE, true)) {
                    $_file = explode('.', $file);
                    $_subdir = $subdir;
                    foreach (array_keys($langs) as $lang) {
                        $_subdir = str_replace('/' . $lang . '/', '/', $_subdir);
                    }
                    $_subdir = preg_replace('#(^|/)images(_custom)?/#', '', $_subdir);
                    $out[$_subdir . $_file[0]] = cms_rawurlrecode('themes/' . rawurlencode($theme) . '/' . $subdir . rawurlencode($file));
                }
            }
        }

        closedir($_dir);
    }

    return $out;
}

/**
 * Get all the image IDs (both already known, and those uncached) of a certain type (i.e. under a subdirectory).
 *
 * @param  ID_TEXT $type The type of image (e.g. 'cns_emoticons')
 * @param  boolean $recurse Whether to search recursively; i.e. in subdirectories of the type subdirectory
 * @param  ?object $db The database connector to work over (null: site db)
 * @param  ?ID_TEXT $theme The theme to search in, in addition to the default theme (null: no other)
 * @param  boolean $dirs_only Whether to only return directories (advanced option, rarely used)
 * @param  boolean $db_only Whether to only return from the database (advanced option, rarely used)
 * @param  array $skip The list of files/directories to skip
 * @param  boolean $include_missing Whether to include missing images in the list
 * @return array The list of image IDs
 */
function get_all_image_ids_type(string $type, bool $recurse = false, ?object $db = null, ?string $theme = null, bool $dirs_only = false, bool $db_only = false, array $skip = [], bool $include_missing = false) : array
{
    if ($db === null) {
        $db = $GLOBALS['SITE_DB'];
    }

    static $cache = [];
    $cache_sig = serialize([$type, $recurse, $theme, $dirs_only, $db_only, $skip]);
    if ($db === $GLOBALS['SITE_DB']) {
        if (isset($cache[$cache_sig])) {
            return $cache[$cache_sig];
        }
    }

    require_code('images');
    require_code('files');

    global $THEME_IMAGES_CACHE;

    if ($theme === null) {
        $theme = 'default';
    }

    if ((substr($type, 0, 4) == 'cns_') && (file_exists(get_file_base() . '/themes/default/images/avatars/index.html'))) { // Allow debranding of theme img dirs
        $type = substr($type, 4);
    }

    if (substr($type, -1) == '/') {
        $type = substr($type, 0, strlen($type) - 1);
    }

    $ids = [];

    if ((!$db_only) && ((!$db->is_forum_db()) || ($dirs_only) || (!is_on_multi_site_network()))) {
        _get_all_image_ids_type($ids, get_file_base() . '/themes/default/images/' . (($type == '') ? '' : ($type . '/')), $type, $recurse, $dirs_only, $skip);
        _get_all_image_ids_type($ids, get_file_base() . '/themes/default/images/' . get_site_default_lang() . '/' . (($type == '') ? '' : ($type . '/')), $type, $recurse, $dirs_only, $skip);
        if ($theme != 'default') {
            _get_all_image_ids_type($ids, get_custom_file_base() . '/themes/' . $theme . '/images/' . (($type == '') ? '' : ($type . '/')), $type, $recurse, $dirs_only, $skip);
            _get_all_image_ids_type($ids, get_custom_file_base() . '/themes/' . $theme . '/images/' . get_site_default_lang() . '/' . (($type == '') ? '' : ($type . '/')), $type, $recurse, $dirs_only, $skip);
        }
        _get_all_image_ids_type($ids, get_file_base() . '/themes/default/images_custom/' . (($type == '') ? '' : ($type . '/')), $type, $recurse, $dirs_only, $skip);
        _get_all_image_ids_type($ids, get_file_base() . '/themes/default/images_custom/' . get_site_default_lang() . '/' . (($type == '') ? '' : ($type . '/')), $type, $recurse, $dirs_only, $skip);
        if ($theme != 'default') {
            _get_all_image_ids_type($ids, get_custom_file_base() . '/themes/' . $theme . '/images_custom/' . (($type == '') ? '' : ($type . '/')), $type, $recurse, $dirs_only, $skip);
            _get_all_image_ids_type($ids, get_custom_file_base() . '/themes/' . $theme . '/images_custom/' . get_site_default_lang() . '/' . (($type == '') ? '' : ($type . '/')), $type, $recurse, $dirs_only, $skip);
        }
    }

    if (!$dirs_only) {
        $expected_slashes = substr_count($type, '/') + 1;

        $query = 'SELECT DISTINCT id,url,theme FROM ' . $db->get_table_prefix() . 'theme_images WHERE ';
        if (!$db_only) {
            $query .= 'url NOT LIKE \'' . db_encode_like('themes/default/images/%') . '\' AND ' . db_string_not_equal_to('url', 'themes/default/images/blank.gif') . ' AND ';
        }
        $query .= '(' . db_string_equal_to('theme', $theme) . ' OR ' . db_string_equal_to('theme', 'default') . ') AND id LIKE \'' . db_encode_like($type . '%') . '\' ORDER BY url';
        $rows = $db->query($query);
        foreach ($rows as $row) {
            if ($row['url'] == '') {
                continue;
            }

            foreach ($skip as $s) {
                if (preg_match('#(^|/)' . preg_quote($s, '#') . '(/|$)#', $row['url']) != 0) {
                    continue 2;
                }
            }

            if ((!$recurse) && (substr_count($row['url'], '/') > $expected_slashes)) {
                continue;
            }

            if ((url_is_local($row['url'])) && (!$include_missing) && (!file_exists(((substr($row['url'], 0, 15) == 'themes/default/') ? get_file_base() : get_custom_file_base()) . '/' . rawurldecode($row['url'])))) {
                continue;
            }
            if ($row['url'] != 'themes/default/images/blank.gif') { // We sometimes associate to blank.gif to essentially delete images so they can never be found again
                // Optimisation to avoid having to build the full theme image table for a new theme in one step (huge numbers of queries)
                if (!multi_lang()) {
                    if (($theme == $row['theme']) || (($row['theme'] == 'default') && (!isset($THEME_IMAGES_CACHE['site'][$row['id']])))) {
                        $THEME_IMAGES_CACHE['site'][$row['id']] = $row['url'];
                    }
                }

                $ids[] = $row['id'];
            } else {
                $key = array_search($row['id'], $ids);
                if (is_integer($key)) {
                    unset($ids[$key]);
                }
            }
        }
    }
    sort($ids);

    $ret = array_unique($ids);

    if ($db === $GLOBALS['SITE_DB']) {
        $cache[$cache_sig] = $ret;
    }

    return $ret;
}

/**
 * Get all the image IDs (both already known, and those uncached) of a certain type (i.e. under a subdirectory).
 *
 * @param  array $ids The list of image IDs found so far. This list will be appended as we proceed
 * @param  ID_TEXT $dir The specific theme image subdirectory we are currently looking under
 * @param  ID_TEXT $type The type of image (e.g. 'cns_emoticons')
 * @param  boolean $recurse Whether to search recursively; i.e. in subdirectories of the type subdirectory
 * @param  boolean $dirs_only Whether to only return directories (advanced option, rarely used)
 * @param  array $skip The list of files/directories to skip
 *
 * @ignore
 */
function _get_all_image_ids_type(array &$ids, string $dir, string $type, bool $recurse, bool $dirs_only, array $skip)
{
    $has_skip = (!empty($skip));

    $_dir = @opendir($dir);
    if ($_dir !== false) {
        while (false !== ($file = readdir($_dir))) {
            if ($file[0] == '.' || $file == 'index.html') {
                continue; // Optimisation, so no need for should_ignore_file call
            }
            if ($has_skip && in_array($file, $skip)) {
                continue;
            }

            $path = $dir . (($dir != '') ? '/' : '') . $file;
            $is_dir = is_dir($path);

            if ($is_dir) {
                if (($recurse) && (!should_ignore_file($file, IGNORE_ACCESS_CONTROLLERS)) && ((strlen($file) != 2) || (cms_strtoupper_ascii($file) != $file))) {
                    $type_path = $type . (($type != '') ? '/' : '');

                    if ($dirs_only) {
                        $ids[] = $type_path . $file;
                    }
                    _get_all_image_ids_type($ids, $path, $type_path . $file, true, $dirs_only, $skip);
                }
            } else {
                if (!$dirs_only) {
                    if ((preg_match('#^[' . URL_CONTENT_REGEXP . ']+\.(png|jpg|jpeg|gif|svg)$#', $file) != 0/*optimisation*/) || (!should_ignore_file($file, IGNORE_ACCESS_CONTROLLERS))) {
                        $type_path = $type . (($type != '') ? '/' : '');

                        $dot_pos = strrpos($file, '.');
                        if ($dot_pos === false) {
                            $dot_pos = strlen($file);
                        }
                        if (is_image($file, IMAGE_CRITERIA_NONE, true)) {
                            $ids[] = $type_path . substr($file, 0, $dot_pos);
                        }
                    }
                }
            }
        }
        closedir($_dir);
    }
}

/**
 * Get Tempcode for a radio list to choose an image from the image FILES in the theme.
 *
 * @param  string $selected_path The currently selected image path (blank for none)
 * @param  URLPATH $base_url The base URL to where we are searching for images
 * @param  PATH $base_path The base-path to where we are searching for images
 * @return Tempcode The generated Tempcode
 */
function combo_get_image_paths(string $selected_path, string $base_url, string $base_path) : object
{
    $out = new Tempcode();

    $paths = get_image_paths($base_url, $base_path);
    $i = 0;
    foreach ($paths as $pretty => $url) {
        $checked = (($url == $selected_path) || (($selected_path == '') && ($i == 0)));
        $out->attach(do_template('FORM_SCREEN_INPUT_RADIO_LIST_ENTRY_PICTURE', ['_GUID' => 'd2ff01291e5f0c0e4cf4ee5b6061593c', 'CHECKED' => $checked, 'NAME' => 'path', 'VALUE' => $url, 'URL' => $url, 'PRETTY' => $pretty]));
        $i++;
    }

    return $out;
}

/**
 * Search under a base path for image FILE URLs.
 *
 * @param  URLPATH $base_url The base URL to where we are searching for images
 * @param  PATH $base_path The base-path to where we are searching for images
 * @param  boolean $include_lang_dir_structure Whether to also include stuff under language-specific directories
 * @return array path->url map of found images
 */
function get_image_paths(string $base_url, string $base_path, bool $include_lang_dir_structure = false) : array
{
    $out = [];

    require_code('images');
    require_code('files');

    $handle = @opendir($base_path);
    if ($handle !== false) {
        while (false !== ($file = readdir($handle))) {
            if (!should_ignore_file($file, IGNORE_ACCESS_CONTROLLERS)) {
                $this_path = $base_path . $file;
                if (is_file($this_path)) {
                    if (is_image($file, IMAGE_CRITERIA_NONE, true)) {
                        $this_url = cms_rawurlrecode($base_url . rawurlencode($file));
                        $out[$this_path] = $this_url;
                    }
                } elseif (($include_lang_dir_structure) || (strlen($file) != 2) || (cms_strtoupper_ascii($file) != $file)) {
                    $out = array_merge($out, get_image_paths($base_url . $file . '/', $base_path . $file . '/'));
                }
            }
        }
        closedir($handle);
    }

    return $out;
}

/**
 * Get all the themes image codes. THIS DOES NOT SEARCH THE DB - DO NOT USE UNLESS IT'S ON A PURE PACKAGED THEME.
 *
 * @param  PATH $base_path The base-path to where we are searching for images
 * @param  PATH $search_under The path to search under, relative to the base-path. This is not the same as the base-path, as we are cropping paths to the base-path
 * @param  boolean $recurse Whether to search recursively from the given directory
 * @return array A list of image codes
 */
function get_all_image_codes(string $base_path, string $search_under, bool $recurse = true) : array
{
    $out = [];

    require_code('images');
    require_code('files');

    if (!file_exists($base_path . '/' . $search_under)) {
        return [];
    }
    $handle = @opendir($base_path . '/' . $search_under);
    if ($handle !== false) {
        while (false !== ($file = readdir($handle))) {
            if (!should_ignore_file($file, IGNORE_ACCESS_CONTROLLERS)) {
                $full_path = $base_path . '/' . $search_under . '/' . $file;
                if (is_file($full_path)) {
                    if (is_image($file, IMAGE_CRITERIA_NONE, true)) {
                        $dot_pos = strrpos($file, '.');
                        if ($dot_pos === false) {
                            $dot_pos = strlen($file);
                        }
                        $_file = substr($file, 0, $dot_pos);
                        $short_path = ($search_under == '') ? $_file : ($search_under . '/' . $_file);
                        $out[$short_path] = 1;
                    }
                } elseif ((strlen($file) != 2) || (cms_strtoupper_ascii($file) != $file)) {
                    if ($recurse) {
                        $out += get_all_image_codes($base_path, $search_under . '/' . $file);
                    }
                }
            }
        }
        closedir($handle);
    }

    return $out;
}

/**
 * Get Tempcode for a dropdown to choose a theme from the themes present.
 *
 * @param  ?ID_TEXT $it The currently selected image ID (null: none selected)
 * @param  ?string $filter An SQL where clause (including the WHERE), that filters the query somehow (null: none)
 * @param  boolean $do_id Whether to show IDs as the list entry captions, rather than URLs
 * @param  boolean $include_all Whether to include images not yet used (i.e not in theme_images map yet)
 * @param  string $under Only include images under this path. Including a trailing slash unless you specifically want to filter allowing filename stubs as well as URLs (blank: no limitation)
 * @return Tempcode Tempcode for a list selection of theme images
 */
function create_selection_list_theme_images(?string $it = null, ?string $filter = null, bool $do_id = false, bool $include_all = false, string $under = '') : object
{
    $out = new Tempcode();
    if (!$include_all) {
        $rows = $GLOBALS['SITE_DB']->query('SELECT id,url FROM ' . get_table_prefix() . 'theme_images WHERE ' . db_string_equal_to('theme', $GLOBALS['FORUM_DRIVER']->get_theme()) . ' ' . $filter . ' ORDER BY url');
        foreach ($rows as $myrow) {
            $id = $myrow['id'];

            if (substr($id, 0, strlen($under)) != $under) {
                continue;
            }

            $selected = ($id == $it);

            $out->attach(form_input_list_entry($id, $selected, ($do_id) ? $id : $myrow['url']));
        }
    } else {
        $rows = get_all_image_ids_type($under, true);
        foreach ($rows as $id) {
            if (substr($id, 0, strlen($under)) != $under) {
                continue;
            }

            $selected = ($id == $it);

            $out->attach(form_input_list_entry($id, $selected));
        }
    }

    return $out;
}

/**
 * Delete a theme image used for a resource that was added, but only if the theme image is now unused.
 *
 * @param  ?ID_TEXT $new The new theme image (null: no new one)
 * @param  ID_TEXT $old The old theme image we might be tidying up
 * @param  ID_TEXT $table Table to check against
 * @param  ID_TEXT $field Field in table
 * @param  ?object $db Database connector to check against (null: site database)
 */
function tidy_theme_img_code(?string $new, string $old, string $table, string $field, ?object $db = null)
{
    if ($new === $old) {
        return; // Still being used
    }

    $url = ($old == '') ? null : find_theme_image($old, true, true);
    if (empty($url)) {
        return;
    }

    if ((strpos($url, '/images_custom/') !== false) && ($GLOBALS['SITE_DB']->query_select_value('theme_images', 'COUNT(DISTINCT id)', ['url' => $url]) == 1)) {
        if ($db === null) {
            $db = $GLOBALS['SITE_DB'];
        }
        $count = $db->query_select_value($table, 'COUNT(*)', [$field => $old]);
        if ($count == 0) {
            @unlink(get_custom_file_base() . '/' . urldecode($url));
            sync_file(get_custom_file_base() . '/' . urldecode($url));
            $GLOBALS['SITE_DB']->query_delete('theme_images', ['id' => $old]);
        }
    }
}
